//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <src/util.hpp>
#include <src/uri++.hpp>
#include <src/file_utils.hpp>
#include <src/vfs.hpp>
#include <src/database.hpp>
#include <src/library.hpp>
#include <src/main.hpp>
#include <src/x_library.hpp>
#include <src/x_play.hpp>

#include <glibmm.h>

namespace Bmp
{
  namespace VFS
  {
      class PluginContainerFolder
          : public Bmp::VFS::PluginContainerBase
      {
        public:

          virtual bool
          can_process (const Glib::ustring& uri)
          {
            try
              {
                std::string path = Glib::filename_from_uri (uri);
                if (Glib::file_test (path, Glib::FILE_TEST_IS_DIR)) return true;
              }
            catch (...) {}
            return false;
          }

          virtual bool 
          handle_read	(Bmp::VFS::Handle&  handle,
                       UriList&           list)
          {
            try
              {
                Util::FileList flist;
                Util::collect_path (Glib::filename_from_uri (handle.get_uri()), flist);

                for (Util::FileList::const_iterator iter = flist.begin (), end = flist.end (); iter != end; ++iter)
                { 
                  if (play->is_audio_file ((*iter)))
                    {
                      list.push_back (Glib::filename_to_uri (*iter));
                    }
                }
                return true;
              }
            catch (...)
              {
                return false;
              }
          }

          virtual bool
          can_write ()
          {
            return false;
          }

          virtual bool
          handle_write  (Bmp::VFS::Handle&	  handle,
                         const UriList&  uri_list)
          {
            return false;
          }

          virtual Bmp::VFS::ExportData
          get_export_data ()
          {
            static Bmp::VFS::ExportData export_data ("Folder Container Plugin", ""); 
            return export_data;
          }
      };
  }
}
  
extern "C" Bmp::VFS::PluginContainerBase* plugin_create ()
{
  return new Bmp::VFS::PluginContainerFolder;
}

extern "C" void plugin_delete (Bmp::VFS::PluginContainerFolder* p)
{
  delete p;
}
