#include <gtkmm.h>
#include <glibmm/i18n.h>
#include <libglademm.h>

#include "ui_util.hpp"

#include "x_play.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

#include "main.hpp"
#include "export.hpp"

namespace Bmp
{
    // Ctor/Creation
      ExportDialog*
      ExportDialog::create ()
      {
        const std::string path = Glib::build_filename (DATA_DIR, "glade" G_DIR_SEPARATOR_S "export.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        ExportDialog *browser = 0;
        glade_xml->get_widget_derived ("export_dialog", browser);
        return browser;
      }

      ExportDialog::ExportDialog (BaseObjectType *cobject,
                                  const Glib::RefPtr<Gnome::Glade::Xml> &xml)
            : Gtk::Dialog (cobject),
              ref_xml (xml)
      {
        using namespace Gtk;
        using namespace Glib;
        using namespace std;

        ref_xml->get_widget ("view_plugins",
                  view_plugins);

        ref_xml->get_widget ("entry_filename",
                  entry_filename);

        ref_xml->get_widget ("button_browse",
                      button_browse);

        ref_xml->get_widget ("all",
                  all);

        ref_xml->get_widget ("filebrowser",
                  filebrowser);
    
        bmp_window_set_icon_list (GTK_WIDGET (this->gobj()), "player");


        Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
        view_plugins->append_column (_("Export As..."), *cell);
        Gtk::TreeViewColumn *column  = view_plugins->get_column (0);
        column->add_attribute (*cell, "text", 0);

        store_plugins = Gtk::ListStore::create (p_record);

        Bmp::VFS::ExportDataList containers;
        vfs->get_containers (containers);

        for (Bmp::VFS::ExportDataList::const_iterator container = containers.begin ();
             container != containers.end ();
             container++)
        {
          TreeModel::iterator i = store_plugins->append ();
          (*i)[p_record.desc] = container->description;
          (*i)[p_record.extension] = container->extension;
            }

        view_plugins->set_model (store_plugins);
        view_plugins->signal_cursor_changed().connect (sigc::mem_fun (*this, &Bmp::ExportDialog::cursor_changed));
        button_browse->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::ExportDialog::select_path));
        all->signal_toggled().connect (sigc::mem_fun (*this, &Bmp::ExportDialog::all_toggled));

        all_files = all->get_active ();

        TreeModel::iterator first = store_plugins->children().begin();  
        view_plugins->get_selection()->select (first);
        string extension = (*first)[p_record.extension];
        string listname  = build_filename (get_home_dir (), "Playlist." + extension);
        entry_filename->set_text (listname);
     }

      void
      ExportDialog::all_toggled ()
      {
        this->all_files = (all->get_active());
      }

      void
      ExportDialog::cursor_changed ()
      {
        using namespace Gtk;
        using namespace Glib;
        using namespace std;

        ustring name;

        ustring text = entry_filename->get_text ();

        string::size_type position = text.rfind (".");

        if (position != string::npos)
          {
            name = text.substr (0, position+1);
          }
        else
          {
            name = text;
          }

        TreeModel::iterator selected = view_plugins->get_selection()->get_selected (); 
        string extension = (*selected)[p_record.extension];
        ustring composited;
        if (position != string::npos)
          {
            composited = name + extension;
          } 
        else
          {
            composited = name + "." + extension;
          }
    
        entry_filename->set_text (composited);
        this->container = (*selected)[p_record.extension];
      }

      void
      ExportDialog::select_path ()
      {
        using namespace Glib;
        using namespace Gtk;
        using namespace std;

        ustring filename = entry_filename->get_text ();
        if (!filename.empty ())
        {
          string basename = Glib::path_get_basename (filename);
          string dirname = Glib::path_get_dirname (filename);
          
          filebrowser->set_current_folder (dirname);
          filebrowser->set_current_name (basename);
        }

        int response = filebrowser->run ();

        switch (response)
        {
            case Gtk::RESPONSE_CANCEL: break;
            case Gtk::RESPONSE_OK:
              {
            entry_filename->set_text (filebrowser->get_filename());     
            break;
              }
        }
        filebrowser->hide ();
      }
}
