//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <iostream>

#include <gtkmm.h>
#include <dbus/dbus-glib.h>
#include <dbus-gaim.h>

#include "gaim_dbus.hpp"

namespace Bmp
{
  namespace Gaim
  {
    DBUS::DBUS () : bus (0), gaimproxy (0)
    {
        GError *error = 0;

        bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error); 
        if (error)
          {
            throw Exception (error->message, Exception::NO_CONNECTION);
          }

        gaimproxy = dbus_g_proxy_new_for_name_owner (bus,
                                                      DBUS_SERVICE_GAIM,
                                                      DBUS_PATH_GAIM,
                                                      DBUS_INTERFACE_GAIM,
                                                      &error);

        if (error)
          {
            throw Exception (error->message, Exception::NO_CONNECTION);
          }
    }

    DBUS::~DBUS ()
    {
      g_object_unref (gaimproxy);
      dbus_g_connection_unref (bus);
    }

    DBUS::OnlineBuddies
    DBUS::get_online_buddies ()
    {
      GError *error = 0;
      OnlineBuddies buddies;
      int node;

      if (!dbus_g_proxy_call (gaimproxy, "GaimBlistGetRoot",
                              &error,
                              G_TYPE_INVALID,
                              G_TYPE_INT,
                              &node,
                              G_TYPE_INVALID))
        {
          throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
        }

      while (node)
        {
          GError *error = 0;
          Buddy buddy;

          int is_buddy = 0;
          if (!dbus_g_proxy_call (gaimproxy, "GaimBlistNodeIsBuddy",
                                &error,
                                G_TYPE_INT,
                                node,
                                G_TYPE_INVALID,
                                G_TYPE_INT,
                                &is_buddy,
                                G_TYPE_INVALID))
            {
              throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
            }

          if (is_buddy)
            {
                // Is online? 
                int online = 0;
                if (!dbus_g_proxy_call (gaimproxy, "GaimBuddyIsOnline",
                                      &error,
                                      G_TYPE_INT,
                                      node,
                                      G_TYPE_INVALID,
                                      G_TYPE_INT,
                                      &online,
                                      G_TYPE_INVALID))
                  {
                    throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                  }

                if (online)
                  {
                    char    *name           = 0,
                            *alias          = 0,
                            *protocol_name  = 0,
                            *protocol_id    = 0;

                    int      account        = 0,
                             connection     = 0,
                             icon           = 0,
                             contact        = 0;

                    // Contact 
                    if (!dbus_g_proxy_call (gaimproxy, "GaimBuddyGetContact",
                                          &error,
                                          G_TYPE_INT,
                                          node,
                                          G_TYPE_INVALID,
                                          G_TYPE_INT,
                                          &contact,
                                          G_TYPE_INVALID))
                      {
                        throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                      }


                    // Name 
                    if (!dbus_g_proxy_call (gaimproxy, "GaimBuddyGetName",
                                          &error,
                                          G_TYPE_INT,
                                          node,
                                          G_TYPE_INVALID,
                                          G_TYPE_STRING,
                                          &name,
                                          G_TYPE_INVALID))
                      {
                        throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                      }

                    // ContactAlias 
                    if (!dbus_g_proxy_call (gaimproxy, "GaimContactGetAlias",
                                          &error,
                                          G_TYPE_INT,
                                          contact,
                                          G_TYPE_INVALID,
                                          G_TYPE_STRING,
                                          &alias,
                                          G_TYPE_INVALID))
                      {
                        throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                      }

                    // Get buddy account
                    if (!dbus_g_proxy_call (gaimproxy, "GaimBuddyGetAccount",
                                          &error,
                                          G_TYPE_INT,
                                          node,
                                          G_TYPE_INVALID,
                                          G_TYPE_INT,
                                          &account,
                                          G_TYPE_INVALID))
                      {
                        throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                      }

                    // Get icon
                    if (!dbus_g_proxy_call (gaimproxy, "GaimBuddyGetIcon",
                                          &error,
                                          G_TYPE_INT,
                                          node,
                                          G_TYPE_INVALID,
                                          G_TYPE_INT,
                                          &icon,
                                          G_TYPE_INVALID))
                      {
                        throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                      }


                    // Get connection 
                    if (!dbus_g_proxy_call (gaimproxy, "GaimAccountGetConnection",
                                          &error,
                                          G_TYPE_INT,
                                          account,
                                          G_TYPE_INVALID,
                                          G_TYPE_INT,
                                          &connection,
                                          G_TYPE_INVALID))
                      {
                        throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                      }

                    // Get protocol id 
                    if (!dbus_g_proxy_call (gaimproxy, "GaimAccountGetProtocolId",
                                          &error,
                                          G_TYPE_INT,
                                          account,
                                          G_TYPE_INVALID,
                                          G_TYPE_STRING,
                                          &protocol_id,
                                          G_TYPE_INVALID))
                      {
                        throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                      }

                    // Get protocol name 
                    if (!dbus_g_proxy_call (gaimproxy, "GaimAccountGetProtocolName",
                                          &error,
                                          G_TYPE_INT,
                                          account,
                                          G_TYPE_INVALID,
                                          G_TYPE_STRING,
                                          &protocol_name,
                                          G_TYPE_INVALID))
                      {
                        throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
                      }

                    buddy.name          = name;
                    buddy.alias         = alias;
                    buddy.protocol_name = protocol_name;
                    buddy.protocol_id   = protocol_id;

                    buddy.account       = account;
                    buddy.connection    = connection;
                    buddy.contact       = contact;
                    buddy.icon          = icon;

                    buddies.push_back (buddy);
                  }
            }
  
          int current_node = node;
          if (!dbus_g_proxy_call (gaimproxy, "GaimBlistNodeNext",
                                  &error,
                                  G_TYPE_INT,
                                  current_node,
                                  G_TYPE_INT,
                                  int(0),
                                  G_TYPE_INVALID,
                                  G_TYPE_INT,
                                  &node,
                                  G_TYPE_INVALID))
            {
              throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
            }
        }

      return buddies;
    }

    void  
    DBUS::send_file (const Buddy& buddy, std::string const& filename)
    {
      GError *error = 0;
      if (!dbus_g_proxy_call (gaimproxy, "ServSendFile",
                                  &error,
                                  G_TYPE_INT,
                                  buddy.connection,
                                  G_TYPE_STRING,
                                  buddy.name.c_str(),
                                  G_TYPE_STRING,
                                  filename.c_str(),
                                  G_TYPE_INVALID))
        {
          throw Exception (error->message, Exception::GAIM_DBUS_ERROR);
        }
    }

  } // ns::Gaim
} // ns::Bmp
