//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_HAL_HPP
#define BMP_HAL_HPP

#include <glibmm.h>
#include <gtkmm.h>

#include <config.h>
#include <stdio.h>
#include <dirent.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <libhal.h>
#include <libhal-storage.h>
#include <goa/libgoa.h>


namespace Bmp
{
  namespace Library
  {

    class HalException : public std::exception
    {
      public:

        enum ExceptionID
        {
          HAL_NO_VOLUME_FOR_MOUNT_PATH,
          HAL_NO_VOLUME_FOR_URI,
          HAL_NO_MOUNT_PATH_FOR_UDI,
        };

        HalException (std::string const& what, ExceptionID id, Glib::ustring const& item)
            : what_ (what), id_ (id), item_ (item)
        {
          g_log (BMP_HAL_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "what (\"%s\") id (\"%d\"), item (\"%s\")", what.c_str(), id, item_.c_str());
        }

        virtual ~HalException () throw()
        {}

        virtual const char *
        what () const throw()
        {
          return what_.c_str();
        }

        const char * item () const
        {
           return item_.c_str();
        }

        ExceptionID id () const
        {
          return id_;
        }

      private:
        
        std::string             what_;
        ExceptionID             id_;
        Glib::ustring           item_;
    };



      struct HALVolume
      {
          std::string	    volume_udi;
          std::string	    device_udi;
          std::string	    mount_path;
          std::string	    device_serial;
          std::string	    volume_name;
          std::string	    device_file;
          bool		        is_disc;
          LibHalDriveBus  device_bus;
      };
      typedef std::map < std::string, HALVolume > HALVolumes;

      struct HALDevice
      {
          std::string     device_udi;
          std::string     device_name;
          std::string	    device_file;
          bool		        is_disc;
      };
      typedef std::vector  < HALDevice > HALDevices;
      typedef sigc::signal  <void, HALVolume > SignalVolume;

      class HAL
      {
        public:

          HAL ();
          ~HAL ();

          HALVolume
          get_volume_for_uri (Glib::ustring const& uri);

          std::string
          get_mount_path_for_volume_udi (std::string const& volume_udi);

          HALDevices
          get_optical_devices () const
          {
            return optical;
          }

          SignalVolume&
          signal_volume_removed ();

          SignalVolume&
          signal_volume_added   ();
          
          bool
          volume_is_present (std::string const& udi)
          {
            if (!initialized)
              {
                return true; 
              }
            return (volumes_by_udi.find (udi) != volumes_by_udi.end());
          }

          bool
          is_initialized () const
          {
            return initialized;
          }

        private:

          SignalVolume signal_volume_removed_;
          SignalVolume signal_volume_added_;

          DBusConnection     *dbus_connection;
          LibHalContext      *ctx;

          HALVolumes		volumes;
          HALVolumes		volumes_by_udi;
          HALDevices		optical;

          int			msg_domain_id;
          bool		initialized;

          dbus_bool_t
          hal_mainloop_integration (LibHalContext *ctx,
                  DBusError     *error);
          bool
          hal_init ();

          void
          scan_devices      (const char    *udi);

          static void
          device_added      (LibHalContext *ctx,
                 const char    *udi);

          static void
          device_removed    (LibHalContext *ctx,
                 const char    *udi);
          static void
          property_modified (LibHalContext    *ctx,
                 const char       *udi,
                 const char       *key,
                 dbus_bool_t       is_removed,
                 dbus_bool_t       is_added);

      };
  }
}

#endif
