//-*- Mode: C++; indent-tabs-mode: nil; -*-

/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <glib.h>
#include <glib-object.h>
#include <glibmm.h>
#include <gtkmm.h>

#include "interfaces.h"

/******** PlugMetadata Plugin IFace */
static void
bmp_plug_metadata_iface_base_init (gpointer g_class)
{
        static gboolean initialized = FALSE;

        if (!initialized) {
                /* create interface signals here. */
                initialized = TRUE;
        }
}

GType
bmp_plug_metadata_iface_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (BmpPlugMetadataInterface),
                        bmp_plug_metadata_iface_base_init,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        0,
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };
                type = g_type_register_static (G_TYPE_INTERFACE, "BmpPlugMetadataInterface", &info, GTypeFlags (0));
        }
        return type;
}

const gchar**
bmp_plug_metadata_get (BmpPlugMetadataInterface *self)
{
	g_return_val_if_fail (BMP_IS_PLUG_METADATA(self), NULL);

        return BMP_PLUG_METADATA_GET_IFACE (self)->get_metadata (self);
}


/******** Flow Plugin IFace */
static void
bmp_plugin_flow_iface_base_init (gpointer g_class)
{
        static gboolean initialized = FALSE;

        if (!initialized) {
                /* create interface signals here. */
                initialized = TRUE;
        }
}

GType
bmp_plugin_flow_iface_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (BmpPluginFlowInterface),
                        bmp_plugin_flow_iface_base_init,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        0,
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };
                type = g_type_register_static (G_TYPE_INTERFACE, "BmpPluginFlowInterface", &info, GTypeFlags (0));
        }
        return type;
}

gboolean
bmp_plugin_flow_item_prev (BmpPluginFlowInterface *self, GtkListStore *list, GtkTreeIter *iter)
{
	g_return_val_if_fail (BMP_IS_PLUGIN_FLOW(self), FALSE);

        return BMP_PLUGIN_FLOW_GET_IFACE (self)->item_prev (self, list, iter);
}

gboolean
bmp_plugin_flow_item_next (BmpPluginFlowInterface *self, GtkListStore *list, GtkTreeIter *iter)
{
	g_return_val_if_fail (BMP_IS_PLUGIN_FLOW(self), FALSE);

        return BMP_PLUGIN_FLOW_GET_IFACE (self)->item_next (self, list, iter);
}

void
bmp_plugin_flow_reset_state (BmpPluginFlowInterface *self)
{
	g_return_if_fail (BMP_IS_PLUGIN_FLOW(self));

        if  (BMP_PLUGIN_FLOW_GET_IFACE (self)->reset_state != NULL)
	     BMP_PLUGIN_FLOW_GET_IFACE (self)->reset_state (self);
}

/******** PlugProperty Plugin IFace */
static void
bmp_plug_property_iface_base_init (gpointer g_class)
{
        static gboolean initialized = FALSE;

        if (!initialized) {
                /* create interface signals here. */
                initialized = TRUE;
        }
}

GType
bmp_plug_property_iface_get_type (void)
{
        static GType type = 0;
        if (type == 0) {
                static const GTypeInfo info = {
                        sizeof (BmpPlugPropertyInterface),
                        bmp_plug_property_iface_base_init,   /* base_init */
                        NULL,   /* base_finalize */
                        NULL,   /* class_init */
                        NULL,   /* class_finalize */
                        NULL,   /* class_data */
                        0,
                        0,      /* n_preallocs */
                        NULL    /* instance_init */
                };
                type = g_type_register_static (G_TYPE_INTERFACE, "BmpPlugPropertyInterface", &info, GTypeFlags (0));
        }
        return type;
}

GList*
bmp_plug_property_get_props (BmpPlugPropertyInterface *self)
{
	g_return_val_if_fail (BMP_IS_PLUG_PROPERTY(self), FALSE);

        return BMP_PLUG_PROPERTY_GET_IFACE (self)->get_props (self);
}

gboolean
bmp_plug_property_set_prop (BmpPlugPropertyInterface *self,
			    GQuark		      property,
			    GValue		     *value)
{
	g_return_val_if_fail (BMP_IS_PLUG_PROPERTY(self), FALSE);

        return BMP_PLUG_PROPERTY_GET_IFACE (self)->set_prop (self, property, value);
}
