//-*- Mode: C++; indent-tabs-mode: nil; -*-

/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 *
 */

#ifndef LASTFM_CLASS_HPP
#define LASTFM_CLASS_HPP

#include <glibmm.h>
#include <system_control.hpp>
#include <mcs/mcs.h>

namespace Bmp
{
  namespace LastFM
  {
    struct Friend
    { 
      Glib::ustring	username;
      Glib::ustring	url;
      Glib::ustring	image;
    };
    typedef std::vector< Friend > Friends;

    class FriendParser
      : public Glib::Markup::Parser
    {
        public:
          FriendParser () : state (0) {}
          virtual ~FriendParser () {}

          Friends friends; 

        protected:

          void
          on_start_element  (Glib::Markup::ParseContext   &context,
                             Glib::ustring const          &element_name,
                             const AttributeMap           &attributes);
          void
          on_end_element    (Glib::Markup::ParseContext   &context,
                             Glib::ustring const          &element_name);
          void
          on_text	          (Glib::Markup::ParseContext   &context,
                             Glib::ustring const          &text);
          void
          on_passtrough     (Glib::Markup::ParseContext   &context,
                             Glib::ustring const          &text);
          void
          on_error	        (Glib::Markup::ParseContext   &context,
                             const Glib::MarkupError      &error);
        private:

            enum Element
            {
              E_NONE	= 0, 
              E_USER  = 1 << 0,
              E_URL 	= 1 << 1,
              E_IMAGE = 1 << 2,
            };

            Friend  current;
            int           state;
          };

    /** Client code for Last.FM Radio
     *
     */
    class Radio
      : public Glib::Object
    {
      public:
        Radio ();
        ~Radio ();

        void
        handshake (MCS_CB_DEFAULT_SIGNATURE);

      private:
        bool handshaked;
    };


    /** Client code for Last.FM
     *
     * Bmp::LastFM::Service is a client implementation of Audioscrobbler Protocol 1.1,
     * (http://www.audioscrobbler.com), used by Last.FM (http://last.fm), a public,
     * free service that tracks your listening habits and provides you with statistics
     * and friends. 
     *
     */
    class Service
      : public Glib::Object
    {
      public:

        Service ();
        ~Service ();

        //Signals
        typedef sigc::signal<void, unsigned int> SignalQueueSize;
        SignalQueueSize& signal_queue_size(); 

        typedef sigc::signal<void> SignalSubmitStart;
        SignalSubmitStart& signal_submit_start(); 

        typedef sigc::signal<void> SignalSubmitEnd;
        SignalSubmitEnd& signal_submit_end();

        void
        emit_queue_size ();

        void
        send_song_information (int position);

        void
        get_friends   (Glib::ustring const& username,
                       Friends& friends);

      private:

        SignalQueueSize signal_queue_size_; 
        SignalSubmitStart signal_submit_start_; 
        SignalSubmitEnd signal_submit_end_;

        struct QueueItem
        {
          std::string md5_pre;
          std::string md5_post;
        };

        struct HandshakeInfo
        {
            /** Whether the server signalized that we're up to date 
             */
            bool uptodate;

            /** URI for track submissions 
             */
            std::string uri;

            /** MD5 challenge string from the server 
             */
            std::string md5;

            /** Initial interval received from the server 
             */
            int interval;

            /** Used for FAILED reason and UPDATE url
             */
            std::string info;
        };

        enum HandshakeStatus
        {
          LASTFM_HANDSHAKE_UPTODATE,
          LASTFM_HANDSHAKE_UPDATE,
          LASTFM_HANDSHAKE_BADUSER,
          LASTFM_HANDSHAKE_FAILED,
          LASTFM_HANDSHAKE_NETWORK_ERROR,
          LASTFM_HANDSHAKE_STATE_UNDEFINED
        };

        enum SignalHandlers
        {
          H_LASTFM_ENABLE,
          H_LASTFM_ENABLE_GENERAL,
          N_HANDLERS
        };
           
        Glib::Mutex	          lock_green_light;
        Glib::Mutex	          lock_queue;
         
        bool		              green_light;
        sigc::connection      green_light_conn;

        HandshakeInfo	        info;
        std::string	          md5_response;
        bool		              handshaked;
        bool		              enabled;
        bool		              sent;

           
        std::list<QueueItem>  queue;
        sigc::connection      queue_conn;

        gulong		            handler[N_HANDLERS];
        int		                message_domain_id;

        HandshakeStatus
        handshake ();

        void
        run_handshake ();

        bool
        process_queue ();

        void
        restart_green_light_timeout (int seconds);

        bool
        set_green_light ();

        
        //MCS callbacks 
        void
        general_enable (std::string const&	    domain,
                        std::string const&	    key,
                        const Mcs::KeyVariant&	value);

        void
        enable_changed (std::string const&	    domain,
                        std::string const&	    key,
                        const Mcs::KeyVariant&	    value);
    };
  } // ns::LastFM
} // ns::Bmp

#endif // LASTFM_CLASS_HPP

