//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <glibmm.h>
#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <ne_basic.h>
#include <ne_compress.h>
#include <ne_socket.h>
#include <ne_utils.h>
#include <ne_props.h>
#include <ne_session.h>
#include <ne_request.h>
#include <ne_uri.h>

#include "paths.hpp"
#include "uri++.hpp"
#include "util.hpp"
#include "main.hpp"
#include "shell.hpp"
#include "stock.hpp"
#include "database.hpp"

#include "x_lastfm.hpp"
#include "x_library.hpp"

#include "lastfm_recommend.hpp"

namespace Bmp
{
  namespace LastFM
  {
      RecommendDialog*
      RecommendDialog::create ()
      {
        const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "lastfm_recommend.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        RecommendDialog *i = 0;
        glade_xml->get_widget_derived ("recommend", i);
        return i;
      }

      RecommendDialog::~RecommendDialog () {}
      RecommendDialog::RecommendDialog (BaseObjectType*                       cobject,
                                        const Glib::RefPtr<Gnome::Glade::Xml> &xml)
            : Gtk::Dialog (cobject),
              m_ref_xml     (xml)
      {
        m_ref_xml->get_widget ("e_artist", e_artist);
        m_ref_xml->get_widget ("e_album",  e_album);
        m_ref_xml->get_widget ("e_title",  e_title);

        m_ref_xml->get_widget ("l_artist", l_artist);
        m_ref_xml->get_widget ("l_album",  l_album);
        m_ref_xml->get_widget ("l_title",  l_title);

        m_ref_xml->get_widget ("cbox_recommend",  cbox_recommend);
        m_ref_xml->get_widget ("view_notes",  view_notes);

        m_ref_xml->get_widget ("notebook", notebook);
        m_ref_xml->get_widget ("image", image);
        image->set (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "wait.gif");

        m_ref_xml->get_widget ("image_lastfm", image_lastfm);
        image_lastfm->set (Gtk::StockID (BMP_STOCK_LASTFM), Gtk::ICON_SIZE_SMALL_TOOLBAR);

        m_ref_xml->get_widget ("treeview", treeview);
        treeview->set_headers_clickable (false);
        treeview->set_headers_visible (false);

        Gtk::CellRendererPixbuf *cell_pixbuf = Gtk::manage (new Gtk::CellRendererPixbuf ());
        cell_pixbuf->property_ypad () = 8;
        cell_pixbuf->property_xpad () = 8;
        treeview->append_column ("Avatar", *cell_pixbuf);
        treeview->get_column (0)->add_attribute (*cell_pixbuf, "pixbuf", 0);
        treeview->get_column (0)->set_resizable (false);

        Gtk::CellRendererText *cell_text = Gtk::manage (new Gtk::CellRendererText ());
        cell_text->property_yalign() = 0.5; 
        treeview->append_column ("Name", *cell_text);
        treeview->get_column (1)->add_attribute (*cell_text, "markup", 1);
        treeview->get_column (1)->set_resizable (false);

        mcs->subscribe ("LastFM-Recommend", "lastfm", "submit-enable", sigc::mem_fun (this, &Bmp::LastFM::RecommendDialog::lastfm_enable_changed));

        notebook->set_current_page (0);
        cbox_recommend->signal_changed().connect (sigc::mem_fun (this, &Bmp::LastFM::RecommendDialog::recommend_changed));
      }

      void
      RecommendDialog::recommend_changed ()
      {
        int current = cbox_recommend->get_active_row_number();

        e_artist->set_sensitive (false);
        e_album->set_sensitive (false);
        e_title->set_sensitive (false);

        l_artist->set_sensitive (false);
        l_album->set_sensitive (false);
        l_title->set_sensitive (false);

        switch (current)
        {
          case 0:
          {
            e_artist->set_sensitive (true);
            l_artist->set_sensitive (true);
            break;
          }

          case 1:
          {
            e_artist->set_sensitive (true);
            l_artist->set_sensitive (true);

            e_album->set_sensitive (true);
            l_album->set_sensitive (true);
            break;
          }

          case 2:
          {
            e_artist->set_sensitive (true);
            l_artist->set_sensitive (true);

            e_title->set_sensitive (true);
            l_title->set_sensitive (true);
            break;
          }
        }
      }

      void
      RecommendDialog::lastfm_enable_changed  (MCS_CB_DEFAULT_SIGNATURE)
      {
        bool active = boost::get<bool>(value);
  
        if (active)
          {
            if (is_visible())
              {
                update_view ();
              }
          }
        else
          {
            liststore = Gtk::ListStore::create (friends_cr); 
            treeview->set_model (liststore);
          }
      }

      void
      RecommendDialog::update_view ()
      {
        notebook->set_current_page (1);
      
        liststore = Gtk::ListStore::create (friends_cr); 
        treeview->set_model (liststore);

        //Get Friends list
        LastFM::Friends friends;
        lastfm->get_friends (mcs->key_get<std::string>("lastfm", "username"), friends); 

        for (LastFM::Friends::const_iterator i = friends.begin (); i != friends.end(); ++i)
        {
          LastFM::Friend const& f = (*i); //Darn you C++ with your 'friend' keyword!
          Gtk::TreeModel::iterator m_Iter = liststore->append();
 
          char *image_url = g_strdup (f.image.c_str()); 
          g_strstrip (image_url);
    
          Glib::RefPtr<Gdk::Pixbuf> image = Util::get_image_from_uri (image_url);
          if (image) (*m_Iter)[friends_cr.avatar] = image->scale_simple ( 32, 32, Gdk::INTERP_HYPER );

          Glib::ustring text;

          text  .append ("<b>")
                .append (f.username)
                .append ("</b>\n<i>")
                .append (f.url)
                .append ("</i>");

          (*m_Iter)[friends_cr.name] = text; 
          (*m_Iter)[friends_cr.username] = f.username; 
          free (image_url);
      
          while (gtk_events_pending()) gtk_main_iteration();
        }
        cbox_recommend->set_active (0);
        notebook->set_current_page (0);
      }

      void
      RecommendDialog::run (Glib::ustring const& uri)
      {
        Bmp::DB::Row row; 
        library->get_metadata (uri, row);
        Bmp::Library::Track track (row);

        if (track.artist) e_artist->set_text (track.artist.get());
        if (track.album)  e_album->set_text  (track.album.get());
        if (track.title)  e_title->set_text  (track.title.get());
      
        show ();
        update_view ();
        if (Gtk::Dialog::run() == GTK_RESPONSE_OK)
          {
            recommend ();
          }
        view_notes->get_buffer()->set_text("");
        hide ();
      }

      void
      RecommendDialog::recommend ()
      {
          using namespace Glib;
          using namespace std;

          static ustring param_pre  = "<param><value><string>";
          static ustring param_post = "</string></value></param>";

          string username = mcs->key_get<std::string>("lastfm", "username"); 
          string password = mcs->key_get<std::string>("lastfm", "password"); 
          string passmd5  = Bmp::Util::md5_hex ((char*)password.c_str(), strlen(password.c_str())); 

          ustring artist = e_artist->get_text ();
          ustring album  = e_album->get_text ();
          ustring title  = e_title->get_text ();

          int type = cbox_recommend->get_active_row_number ();

          ustring challenge = "12345678"; 
          ustring xmldata;

          Gtk::TreeModel::iterator m_Iter = treeview->get_selection()->get_selected();
          ustring user_to = (*m_Iter)[friends_cr.username];
         
          const char * reply = (passmd5+challenge).c_str();
          string md5reply = Bmp::Util::md5_hex ((char*)reply, strlen(reply));

          switch (type)
          {
              case 0:
                {
                  xmldata = "<?xml version=\"1.0\"?><methodCall><methodName>recommendArtist</methodName><params>";
                  xmldata.append (param_pre).append (username).append (param_post);
                  xmldata.append (param_pre).append (challenge).append (param_post);
                  xmldata.append (param_pre).append (md5reply).append (param_post);
                  xmldata.append (param_pre).append (artist).append (param_post);
                  xmldata.append (param_pre).append (user_to).append (param_post);
                  xmldata.append (param_pre).append (view_notes->get_buffer()->get_text(false)).append (param_post);
                  xmldata.append ("</params></methodCall>");
                  break;
                }

              case 1:
                {
                  xmldata = "<?xml version=\"1.0\"?><methodCall><methodName>recommendAlbum</methodName><params>";
                  xmldata.append (param_pre).append (username).append (param_post);
                  xmldata.append (param_pre).append (challenge).append (param_post);
                  xmldata.append (param_pre).append (md5reply).append (param_post);
                  xmldata.append (param_pre).append (artist).append (param_post);
                  xmldata.append (param_pre).append (album).append (param_post);
                  xmldata.append (param_pre).append (user_to).append (param_post);
                  xmldata.append (param_pre).append (view_notes->get_buffer()->get_text(false)).append (param_post);
                  xmldata.append ("</params></methodCall>");
                  break;
                }

              case 2:
                {
                  xmldata = "<?xml version=\"1.0\"?><methodCall><methodName>recommendTrack</methodName><params>";
                  xmldata.append (param_pre).append (username).append (param_post);
                  xmldata.append (param_pre).append (challenge).append (param_post);
                  xmldata.append (param_pre).append (md5reply).append (param_post);
                  xmldata.append (param_pre).append (artist).append (param_post);
                  xmldata.append (param_pre).append (title).append (param_post);
                  xmldata.append (param_pre).append (user_to).append (param_post);
                  xmldata.append (param_pre).append (view_notes->get_buffer()->get_text(false)).append (param_post);
                  xmldata.append ("</params></methodCall>");
                  break;
                }
          }

          ne_session *sess;
          ne_request *reqs;

          sess = ne_session_create ("http", "ws.audioscrobbler.com", 80);
          reqs = ne_request_create (sess, "POST", "/1.0/rw/xmlrpc.php");

          ne_add_request_header (reqs, "Content-Type", "text/xml");
          ne_set_request_body_buffer (reqs, xmldata.c_str (), (size_t)xmldata.length());

#if 0
          char   *memory = 0;    
          char    tmp[1024];
          size_t  size = 0;

          ne_begin_request (reqs);
          bool read = true;
     
          while (read) 
            {
              int len = ne_read_response_block (reqs, tmp, 1024);
              if (!len)
                {
                  read = false;
                  break;
                }

              memory = static_cast<char *> (realloc (memory, size+len));
              std::memcpy (memory+size, tmp, len);
              size += len;
            }
#endif

          int req = ne_request_dispatch (reqs);
          if (req != NE_OK)
            {
              if (req != NE_ERROR)
                {
                  g_message ("NEON Error: %d", req);
                }
              else
                {
                  g_message ("NEON Error: %d, %s", req, ne_get_error (sess));
                }
            }

          ne_request_destroy (reqs);
          ne_session_destroy (sess);
      }
  }
}
