//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_LIBRARY_UI_HPP 
#define BMP_LIBRARY_UI_HPP

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <glibmm/markup.h>
#include <gtkmm.h>
#include <libglademm.h>

#include <mcs/mcs.h>

#ifdef HAVE_HAL
#  include "hal.hpp"
#endif

#include "amazon.hpp"
#include "library.hpp"
#include "vfs.hpp"
#include "playback_source.hpp"

#ifdef HAVE_MOOD
#  include "mood_dialog.hpp"
#endif

#include "x_system_control.hpp"

namespace Bmp
{
  class TreeViewXDS;

  class LibraryUI : public Gtk::Window
  {
    protected:

      virtual bool
      on_key_press_event (GdkEventKey*);

    private:

#ifdef HAVE_MOOD
      friend class Bmp::MooDialog;
#endif

      typedef std::map < Glib::ustring, Gtk::TreeModel::iterator> IterMap;

      class Tracks : public Gtk::TreeModel::ColumnRecord
      {
        public:
          Gtk::TreeModelColumn<int>           track;
          Gtk::TreeModelColumn<Glib::ustring> title;
          Gtk::TreeModelColumn<int>           rating;
          Gtk::TreeModelColumn<Glib::ustring> artist;
          Gtk::TreeModelColumn<Glib::ustring> album;
          Gtk::TreeModelColumn<Glib::ustring> genre;
          Gtk::TreeModelColumn<int>           bitrate;
          Gtk::TreeModelColumn<int>           samplerate;
          Gtk::TreeModelColumn<Glib::ustring> location;
#ifdef HAVE_HAL
          Gtk::TreeModelColumn<bool>          present;
#endif
          Tracks ()
          {
            add (track);
            add (title);
            add (rating);
            add (artist);
            add (album);
            add (genre);
            add (bitrate);
            add (samplerate);
            add (location);
#ifdef HAVE_HAL
            add (present);
#endif
          }
      };
      Tracks tracks;

      enum AlbumFlags
      {
        A_NONE            = 0,
        A_HAS_COVER       = 1 << 0,
        A_HAS_SORTNAME    = 1 << 1
      };

      class Albums : public Gtk::TreeModel::ColumnRecord
      {
        public:
          Gtk::TreeModelColumn<Glib::RefPtr<Gdk::Pixbuf> >  cover;
          Gtk::TreeModelColumn<Glib::ustring>               title;
          Gtk::TreeModelColumn<Glib::ustring>               artist;
          Gtk::TreeModelColumn<Glib::ustring>               album;
          Gtk::TreeModelColumn<AlbumFlags>                  flags;
          Gtk::TreeModelColumn<unsigned int>                rowuid;
          Albums ()
          {
            add (cover);
            add (title);
            add (artist);
            add (album);
            add (flags);
            add (rowuid);
          }
      };
      Albums albums; 

      class CBoxSearchCR : public Gtk::TreeModel::ColumnRecord
      {
        public:
          Gtk::TreeModelColumn<Glib::ustring> name;
          Gtk::TreeModelColumn<std::string> id;
          CBoxSearchCR ()
          {
            add (name);
            add (id);
          }
      };
      CBoxSearchCR cr_search; 

      // Albums bookkeeping 
      typedef std::pair    < Glib::ustring, Glib::ustring >               RowKeyAlbum; 
      typedef std::map     < RowKeyAlbum, Gtk::TreeModel::RowReference>   RowMapAlbums;
      typedef std::pair    < unsigned int, Gtk::TreeModel::RowReference > AlbumRowSelectionPair; 
      typedef std::list    < AlbumRowSelectionPair >                      AlbumsRowSelectionList; 

      RowMapAlbums m_row_map_albums;
      AlbumsRowSelectionList m_selection_list_albums;

      // Tracks bookkeeping 
      typedef std::map     < Glib::ustring, Gtk::TreeModel::RowReference>   RowMapTracks;
      RowMapTracks m_row_map_tracks;

      void
      check_update (const Glib::ustring&      location,
                    Bmp::Library::UpdateTrack track);

#ifdef HAVE_HAL
      void
      hal_volume_removed  (Library::HALVolume volume);
      void
      hal_volume_added    (Library::HALVolume volume);
      void
      tracks_set_row_sensitive   (Gtk::CellRenderer* cell, const Gtk::TreeModel::iterator& m_iter);
      bool
      tracks_row_can_be_selected (const Glib::RefPtr<Gtk::TreeModel>& model, const Gtk::TreeModel::Path& path, bool selected);
      bool
      all_tracks_present  ();
#endif

      enum LibraryPage
      {
        PAGE_ALBUMS,
        PAGE_SEARCH,
      };

      //---- Page Switch
      void
      update_track_view_from_albums ();
      void
      update_track_view_from_search ();
      void
      update_track_view (int id);

      //---- Bottom controls
      void
      enqueue ();
      void
      play ();
      void
      remove ();

      //---- Removal/Adding/Modifications
      void
      import_tracks_into_album ();
      void
      import_tracks ();
      void
      import_tracks_as_album ();
      bool
      import_tracks_as_album_X (const VFS::UriList& urilist);

      void
      approve_all ();
      void
      approve_selected ();
      void
      drop_items ();
      void
      remove_albums (bool force = false);
      void
      remove_albums_action_cb ();
      void
      remove_selected_album (Glib::ustring const& key_A, Glib::ustring const& key_B, Gtk::TreeModel::iterator const& m_Iter);
      void
      fetch_cover (const Gtk::TreePath &path);
      void
      relocate_album ();
      void
      albums_update ();
      void
      albums_modify ();
      void
      tracks_modify ();
      void
      albums_refilter ();
      bool
      albums_row_is_visible (const Gtk::TreeModel::const_iterator& iter);
      bool
      albums_event (GdkEvent * ev);
      bool
      tracks_event (GdkEvent * ev);

      int 
      albums_default_sort_func
      (const Gtk::TreeModel::iterator& iter_a, const Gtk::TreeModel::iterator& iter_b);

      bool
      albums_record_selection_change
      (const Glib::RefPtr<Gtk::TreeModel>& model, const Gtk::TreeModel::Path& path, bool selected);

      void
      notebook_page_switch
      (GtkNotebookPage *page, unsigned int page);

      void
      tracks_selection_changed ();
     
      //---- Auxilliaries
      void
      activate_default (const Gtk::TreeModel::Path& path, Gtk::TreeViewColumn* column);

      void
      fill_tracks_from_rows (const Bmp::DB::VRow& vector);

      Glib::RefPtr<Gnome::Glade::Xml>     m_ref_xml;
      Glib::RefPtr<Gtk::ListStore>        store_tracks;
      Glib::RefPtr<Gtk::ListStore>        store_albums;
      Glib::RefPtr<Gtk::TreeModelFilter>  store_albums_filter;

      Bmp::TreeViewXDS        *view_albums;
      Gtk::TreeView           *view_tracks;

      Gtk::Entry              *albums_filter;
      Gtk::Notebook           *notebook_library;
      Gtk::Notebook           *notebook_add;
      Gtk::Notebook           *notebook_albums;

      Gtk::ProgressBar        *progress_modify;
      Gtk::Label              *label_current_album;

      Gtk::Label              *l_play;
      Gtk::Label              *l_enqueue;

      Gtk::CellRendererText   *cell_album_text;
   
      enum FilterID
      {
        FILTER_1, 
        FILTER_2,
        FILTER_3,

        N_FILTERS,
      };

      // Search 
      void
      filter_toggled (FilterID id);
      void
      cb_rating_toggled ();
      void
      cb_played_toggled ();
      bool
      evb_rating_button_press (GdkEventButton * ev);

      Gtk::CheckButton                *cbutton[N_FILTERS];
      Gtk::ComboBox                   *cbox[N_FILTERS];
      Gtk::Entry                      *entry[N_FILTERS];
      Glib::RefPtr<Gtk::ListStore>     cs_store;

      // Rating
      void
      rating_cell_data_func (Gtk::CellRenderer                *cell_,
                             const Gtk::TreeModel::iterator   &m_Iter);

      Gtk::CellRenderer               *cell_rating;
      Gtk::CheckButton                *cb_rating;
      Gtk::EventBox                   *evb_rating;
      Gtk::Image                      *i_rating;
      Glib::RefPtr<Gdk::Pixbuf>        rating[6];
      unsigned int                     m_current_rating;

      // Played
      Gtk::CheckButton                *cb_played;
      Gtk::ComboBox                   *cbox_playdate;

      // Shutdown
      bool m_updated;
      bool m_busy;
      sigc::connection conn_notebook;
      static gboolean shutdown_request (BmpSystemControl *control,
                                        Bmp::LibraryUI   *ui)
      {
        return ui->m_busy;    
      }

      // View size
      void
      mcs_library_view_size_changed (MCS_CB_DEFAULT_SIGNATURE);

      Gtk::ComboBox              *cb_view_size;
      Glib::RefPtr<Gdk::Pixbuf>   m_disc;
      Glib::RefPtr<Gdk::Pixbuf>   m_disc_small;

      enum ViewSize
      {
        NORMAL,
        SMALL,
      };
      ViewSize m_viewsize; 

      // The final ultimate flag
      bool m_init;

      // -- Our UIManager
      Glib::RefPtr<Gtk::UIManager> m_ui_manager;
    
      enum ActionGroups
      {
        IMPORT,
        ALBUM,
        APPROVE,
        PLAYBACK,

        N_ACTION_GROUPS,
      };

      Glib::RefPtr<Gtk::ActionGroup> m_library_actions[N_ACTION_GROUPS];

    public:

      LibraryUI (BaseObjectType                       *cobject,
                 const Glib::RefPtr<Gnome::Glade::Xml> &xml);

      static LibraryUI*
      create ();

      ~LibraryUI ();

      virtual void
      present ()
      {
        Gtk::Window::present ();
        if (!m_updated)
        {
          albums_update ();
          m_updated = true;
        }
      }
  };
}
#endif
