#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <boost/format.hpp>

#include "file_utils.hpp"
#include "util.hpp"
#include "paths.hpp"
#include "main.hpp"

#include "x_library.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

#include "musicbrainz_xml.hpp"
#include "library_ui_modify.hpp"

namespace
{
  void
  cbe_set_sensitive (Gtk::CheckButton *cb, Gtk::ComboBoxEntry *cbe)
  {
      cbe->set_sensitive (cb->get_active());
  }

  void
  fcb_set_sensitive (Gtk::CheckButton *cb, Gtk::FileChooserButton *fcb)
  {
      fcb->set_sensitive (cb->get_active());
  }
}

namespace Bmp
{
      LibraryUIModify::LibraryUIModify (BaseObjectType                        *cobject,
                                        const Glib::RefPtr<Gnome::Glade::Xml> &xml)
                    : Gtk::Dialog (cobject), ref_xml (xml),
                      disc (Gdk::Pixbuf::create_from_file
                      (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "audio.png"))
      {
          ref_xml->get_widget ("merge_cb_artist",   merge_cb_artist);
          ref_xml->get_widget ("merge_cb_album",    merge_cb_album);

          ref_xml->get_widget ("merge_coverart",    merge_coverart);

          ref_xml->get_widget ("merge_artist_id",   merge_artist_id);
          ref_xml->get_widget ("merge_release_id",  merge_release_id);

          ref_xml->get_widget ("merge_asin",	      merge_asin);
          ref_xml->get_widget ("merge_score",	      merge_score);

          ref_xml->get_widget ("b_refresh",	        b_refresh);
          ref_xml->get_widget ("l_tracks",	        l_tracks);

          ref_xml->get_widget ("view_release",	    view_release);
          ref_xml->get_widget ("view_local",	      view_local);

          ref_xml->get_widget ("b_move_up",	        b_move_up);
          ref_xml->get_widget ("b_move_down",	      b_move_down);

          ref_xml->get_widget ("cancelbutton1",	    cancelbutton1);
          ref_xml->get_widget ("okbutton2",	        okbutton2);
   
          merge_cb_artist_entry = merge_cb_artist->get_entry ();
          merge_cb_album_entry  = merge_cb_album->get_entry ();

          merge_cb_artist_entry->set_activates_default (true);
          merge_cb_album_entry->set_activates_default (true);

          merge_cb_artist_entry->set_text (_("<select here>"));
          merge_cb_album_entry->set_text ("");
          merge_coverart->set (disc);

          b_refresh->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::refresh_releases));
          b_move_up->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::on_button_up));
          b_move_down->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::on_button_down));

          struct { 
              char    *title;
              double   xalign;
              int	     width;
          } headers_tracks[] = {
              { _("#"),	    1.0, 0,  },
              { _("Title"), 0.0, 100 },
          };

          // Release
          for (unsigned int n = 0; n < G_N_ELEMENTS (headers_tracks); ++n)
            {
              Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
              cell->property_xalign() = headers_tracks[n].xalign;
              view_release->append_column (headers_tracks[n].title, *cell);
              Gtk::TreeViewColumn *column = view_release->get_column (n);
              column->add_attribute (*cell, "text", n);
              column->set_resizable (true);
            }

          store_release = Gtk::ListStore::create (tracks_cr);
          view_release->set_model (store_release);
          view_release->get_selection()->set_mode (Gtk::SELECTION_NONE);

          // Local
          {
            Gtk::CellRendererToggle *cell = Gtk::manage ( new Gtk::CellRendererToggle() );
            view_local->append_column ("", *cell);
            Gtk::TreeViewColumn *column = view_local->get_column (0);
            column->add_attribute (*cell, "active", 0);
            column->set_resizable (false);
            cell->signal_toggled().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::on_mod_toggled));
          }

          for (unsigned int n = 0; n < G_N_ELEMENTS (headers_tracks); ++n)
          {
            Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
            cell->property_xalign() = headers_tracks[n].xalign;
            view_local->append_column (headers_tracks[n].title, *cell);
            Gtk::TreeViewColumn *column = view_local->get_column (n+1);
            if (n == 0)
              column->set_cell_data_func (*cell, sigc::mem_fun (*this, &Bmp::LibraryUIModify::cell_data_func_nr));
            else
              column->add_attribute (*cell, "text", n+1);
            column->set_resizable (true);
          }
          store_local = Gtk::ListStore::create (update_tracks_cr);
          view_local->set_model (store_local);
          view_local->get_selection()->set_mode (Gtk::SELECTION_SINGLE);
          view_local->get_selection()->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::local_selection_changed));

          okbutton2->set_sensitive (false);
      }

      LibraryUIModify*
      LibraryUIModify::create ()
      {
        const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "library_ui_modify.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        LibraryUIModify *p = 0;
        glade_xml->get_widget_derived ("dialog_modify", p);
        return p;
      }

      void
      LibraryUIModify::cell_data_func_nr (Gtk::CellRenderer* cell_, const Gtk::TreeModel::iterator& iter)
      {
        Gtk::CellRendererText *cell = reinterpret_cast<Gtk::CellRendererText*>(cell_);
        static boost::format fint ("%d");
        unsigned int nr = (*iter)[update_tracks_cr.tracknr];
        if (nr == 0)
          cell->property_text() = "";
        else
          cell->property_text() = (fint % nr).str();
      }

      void
      LibraryUIModify::on_mod_toggled (const Glib::ustring& path)
      {
        Gtk::TreeModel::iterator m_Iter = store_local->get_iter (path);
        (*m_Iter)[update_tracks_cr.mod] = !(*m_Iter)[update_tracks_cr.mod];
      }

     void
      LibraryUIModify::local_selection_changed ()
      {
        bool selected = view_local->get_selection()->count_selected_rows ();
        b_move_up->set_sensitive (selected);
        b_move_down->set_sensitive (selected);
      }

      void
      LibraryUIModify::on_button_up ()
      {
        Gtk::TreeModel::iterator m_Iter = view_local->get_selection()->get_selected ();
        if (m_Iter == store_local->children().begin()) return;
        Gtk::TreeModel::iterator d_Iter = m_Iter;
        d_Iter--; // yes, this is slow, but it doesn't matter much as a release doesn't have that many tracks
        store_local->iter_swap (m_Iter, d_Iter);
      }

      void
      LibraryUIModify::on_button_down ()
      {
        Gtk::TreeModel::iterator m_Iter = view_local->get_selection()->get_selected ();
        if (m_Iter == store_local->children().end()) return;
        Gtk::TreeModel::iterator d_Iter = m_Iter;
        d_Iter++;
        store_local->iter_swap (m_Iter, d_Iter);
      }

      void
      LibraryUIModify::refresh_releases ()
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          set_sensitive (false);
          
          MusicBrainzReleases mb_releases; 
          get_releases (merge_cb_artist_entry->get_text (), merge_cb_album_entry->get_text (), mb_releases);

          merge_cb_artist_store = ListStore::create (artists);
          merge_cb_album_store  = ListStore::create (albums);
          merge_artist_release_map.clear ();
          merge_coverart->set (disc);
          merge_artist_id->set_text   ("");
          merge_release_id->set_text  ("");
          merge_asin->set_text        ("");
          merge_score->set_text       ("");
          l_tracks->set_text          ("");
          store_release->clear ();
          store_local->clear ();

          for (MusicBrainzReleases::const_iterator r_Iter     = mb_releases.begin (),
                                                   r_IterEnd  = mb_releases.end ();
                                                   r_Iter    != r_IterEnd; ++r_Iter)
          {
            MusicBrainzRelease const& release = (*r_Iter);
            merge_releases.insert (make_pair (release.mb_release_id, release));
            ArtistReleaseMap::iterator map_Iter = merge_artist_release_map.find (release.mb_artist_id);

            if (map_Iter != merge_artist_release_map.end ())
            {
              ReleaseIds &ids = map_Iter->second;
              ids.push_back (release.mb_release_id);
            }
            else
            {
              TreeModel::iterator m_Iter  = merge_cb_artist_store->append ();
              (*m_Iter)[artists.name]	    = release.mb_artist_name;
              (*m_Iter)[artists.id]	      = release.mb_artist_id;

              ReleaseIds ids; 
              ids.push_back (release.mb_release_id);
              merge_artist_release_map.insert (make_pair (release.mb_artist_id, ids)); 
            }
          }

          merge_cb_artist->set_model  (merge_cb_artist_store);
          merge_cb_album->set_model   (merge_cb_album_store);

          merge_cb_artist_entry->set_text (_("<select here>"));
          merge_cb_album_entry->set_text  ("");
          okbutton2->set_sensitive (false);
          set_sensitive (true);
      }

      void
      LibraryUIModify::album_changed ()
      {
          using namespace Bmp::DB;
          using namespace Bmp::Library;
          using namespace Bmp::MusicBrainzXML;
          using namespace std;
          using namespace Gtk;
          using namespace Glib;

          if (merge_cb_album->get_active_row_number () < 0) return;

          set_sensitive (false);
          store_release->clear ();
          store_local->clear ();
        
          //--- Set Up Release 
          TreeModel::const_iterator m_Iter  = merge_cb_album->get_active ();
          ReleaseMap::iterator r_Iter       = merge_releases.find ((*m_Iter)[albums.id]);
          MusicBrainzRelease const& release = r_Iter->second;
          try
            {
              Glib::RefPtr<Gdk::Pixbuf> cover;
              Amazon::get_cover_by_asin (release.asin, cover);
              merge_coverart->set (cover);
            }
          catch (...)
            { 
              merge_coverart->set (disc);
            }
          merge_artist_id->set_text   (release.mb_artist_id);
          merge_release_id->set_text  (release.mb_release_id);
          merge_asin->set_text        (release.asin);
          merge_score->set_text       ((boost::format ("%d") % release.score).str());
          l_tracks->set_text          ((boost::format ("%d") % release.track_list).str());

          tracks = MusicBrainzTracks();

          get_tracks (release.mb_release_id, tracks);

          for (Tracks::iterator iter  = tracks.tracks.begin  (),
                                end   = tracks.tracks.end    ();
                                iter != end; ++iter)
            {
              TreeModel::iterator iter_r = store_release->append();
              if (iter->mb_artist)
              {
                ustring title;
                title .append (iter->mb_artist.get())
                      .append (" - ") 
                      .append (iter->mb_title);
                (*iter_r)[tracks_cr.title]  = title; 
              }
              else
              {
                (*iter_r)[tracks_cr.title]  = iter->mb_title; 
              }

              (*iter_r)[tracks_cr.tracknr]  = iter->tracknum; 
              (*iter_r)[tracks_cr.mb_track] = (*iter);
            }
          view_release->columns_autosize ();


          //--- Set Up Local 
          string id_location = library->get_metadatum_id (DATUM_LOCATION);

          for (::Bmp::VFS::UriList::const_iterator i = m_list.begin (); i != m_list.end(); ++i)
            {
              AttributeList attrs; 
              attrs.push_back (Attribute (DB::EXACT, id_location, (*i)));

              VRow vector;
              library->query (attrs, vector);

              for (VRow::const_iterator v = vector.begin (), end = vector.end (); v != end; ++v) 
                { 
                  Track track (*v);

                  TreeModel::iterator iter_l = store_local->append ();

                  if (track.title)
                      (*iter_l)[update_tracks_cr.title]     = track.title.get(); 

                  if (track.tracknumber)
                      (*iter_l)[update_tracks_cr.tracknr]   = track.tracknumber.get();
                  else
                      (*iter_l)[update_tracks_cr.tracknr]   = 0;

                  if (track.location)
                      (*iter_l)[update_tracks_cr.location]  = track.location.get(); 

                  (*iter_l)[update_tracks_cr.mod] = true;
                }
            }

          view_local->columns_autosize ();

          TreeNodeChildren const& nodes_release = store_release->children();
          TreeNodeChildren const& nodes_local   = store_local->children();

          if (nodes_local.size () < nodes_release.size()) while (nodes_local.size () < nodes_release.size ()) store_local->append();

          okbutton2->set_sensitive (true);
          set_sensitive (true);
      }

      void
      LibraryUIModify::artist_changed ()
      {
          using namespace Gtk;
          using namespace MusicBrainzXML;

          if (merge_cb_artist->get_active_row_number () < 0) return;

          set_sensitive (false);
          // Get releases from selected artist
          TreeModel::const_iterator  m_Iter   = merge_cb_artist->get_active ();
          ArtistReleaseMap::iterator map_Iter = merge_artist_release_map.find ((*m_Iter)[artists.id]);
          ReleaseIds                 &ids     = map_Iter->second;

          merge_cb_album_store->clear ();

          for (ReleaseIds::const_iterator r_Iter    = ids.begin (),
                                          r_IterEnd = ids.end ();
                                          r_Iter != r_IterEnd; ++r_Iter)
            {
              // Iter from releases map
              ReleaseMap::iterator map_Iter = merge_releases.find (*r_Iter);

              // Get the release based on id 
              MusicBrainzRelease const& release = map_Iter->second; 

              TreeModel::iterator c_Iter  = merge_cb_album_store->append ();
              (*c_Iter)[albums.name]      = release.mb_release_name;
              (*c_Iter)[albums.id]        = release.mb_release_id;
            }
          set_sensitive (true);
      }

      int
      LibraryUIModify::run (MusicBrainzXML::MusicBrainzReleases const& releases,
                            Bmp::VFS::UriList const& list)
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;
          using namespace std;
          using namespace Glib;

          m_list = list;

          merge_cb_artist_store = ListStore::create (artists);
          merge_cb_album_store  = ListStore::create (albums);

          for (MusicBrainzReleases::const_iterator r_Iter     = releases.begin(),
                                                   r_IterEnd  = releases.end();
                                                   r_Iter    != r_IterEnd; ++r_Iter)
          {

            MusicBrainzRelease const& release = (*r_Iter);
            merge_releases.insert (std::make_pair (release.mb_release_id, release));
        
            ArtistReleaseMap::iterator map_Iter = merge_artist_release_map.find (release.mb_artist_id);

            if (map_Iter != merge_artist_release_map.end ())
              {
                ReleaseIds &ids = map_Iter->second;
                ids.push_back (release.mb_release_id);
              }
            else
              {
                TreeModel::iterator m_Iter  = merge_cb_artist_store->append ();

                (*m_Iter)[artists.name]	  = release.mb_artist_name;
                (*m_Iter)[artists.id]	  = release.mb_artist_id;
              
                ReleaseIds ids; 
                ids.push_back (release.mb_release_id);
                merge_artist_release_map.insert (std::make_pair (release.mb_artist_id, ids)); 
              }
          }

          // Setup 'Artist' ComboBox 
          merge_cb_artist->set_model (merge_cb_artist_store);
          merge_cb_artist->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::artist_changed));
          merge_cb_artist->set_text_column (0);

          // Setup 'Album' ComboBox 
          merge_cb_album->set_model (merge_cb_album_store);
          merge_cb_album->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModify::album_changed));
          merge_cb_album->set_text_column (0);

          if ( Gtk::Dialog::run ()  == GTK_RESPONSE_OK )
            {
              TreeNodeChildren const& nodes_release = store_release->children();
              TreeNodeChildren const& nodes_local   = store_local->children();

              unsigned int max_nodes = nodes_release.size();
              if (nodes_local.size() < max_nodes) max_nodes = nodes_local.size();

              TreeNodeChildren::const_iterator iter_r = nodes_release.begin();
              TreeNodeChildren::const_iterator iter_l = nodes_local.begin();

              for (unsigned int n = 0; n < max_nodes; ++n)
                {
                  if ((*iter_l)[update_tracks_cr.mod]) 
                    {
                      Library::UpdateTrack update_track;
                      MusicBrainzXML::MusicBrainzTrack const& mb_track = (*iter_r)[tracks_cr.mb_track];
                      MusicBrainzXML::mb_track_to_update_track (mb_track, update_track);

                      update_track.asin                      = tracks.asin; 
                      update_track.tracknumber               = (*iter_r)[tracks_cr.tracknr];
                      update_track.location_cur              = (*iter_l)[update_tracks_cr.location];


                      update_track.artist                    = mb_track.mb_artist ? mb_track.mb_artist.get() : tracks.mb_album_artist; 
                      update_track.title                     = mb_track.mb_title;

                      update_track.album                     = tracks.mb_album_name; 
                      update_track.mb_album_artist           = tracks.mb_album_artist; 
                      update_track.mb_album_artist_sort_name = tracks.mb_album_artist_sort_name;
                      update_track.mb_album_artist_id        = tracks.mb_album_artist_id;
                      update_track.mb_album_id               = tracks.mb_release_id;

                      update_track.mb_track_id               = mb_track.mb_track_id;
                      update_track.mb_artist_sort_name       = mb_track.mb_artist_sort_name;
                      update_track.mb_artist_id              = mb_track.mb_artist_id; 
                      
                      modify_tracks.update_tracks.push_back (update_track);
                    }
                  ++iter_r;
                  ++iter_l;
                }

              modify_tracks.asin                      = tracks.asin;
              modify_tracks.mb_album_artist_sort_name = tracks.mb_album_artist_sort_name;
              modify_tracks.mb_release_id             = tracks.mb_release_id;              
              modify_tracks.mb_album_name             = tracks.mb_album_name;
              modify_tracks.mb_album_artist           = tracks.mb_album_artist;
              modify_tracks.mb_album_artist_id        = tracks.mb_album_artist_id;

              return GTK_RESPONSE_OK;
            }
          else
              return GTK_RESPONSE_CANCEL;
      }


}
