//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_LIBRARY_UI_MODIFY_HPP
#define BMP_LIBRARY_UI_MODIFY_HPP

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <glibmm/markup.h>
#include <gtkmm.h>
#include <libglademm.h>

#include "amazon.hpp"
#include "library.hpp"
#include "musicbrainz_xml.hpp"
#include "vfs.hpp"

namespace Bmp
{
  class LibraryUIModify : public Gtk::Dialog
  {

    public:

        struct ModifyTracks
        {
          Library::UpdateTracks           update_tracks;

          boost::optional<Glib::ustring>  asin;
          boost::optional<Glib::ustring>	mb_album_artist_sort_name;

          Glib::ustring mb_release_id;
          Glib::ustring mb_album_name;
          Glib::ustring mb_album_artist;
          Glib::ustring mb_album_artist_id;

        };

        LibraryUIModify (BaseObjectType                       *cobject,
                         const Glib::RefPtr<Gnome::Glade::Xml> &xml);
        static LibraryUIModify* create ();
       
        int
        run   (MusicBrainzXML::MusicBrainzReleases const& releases,
               Bmp::VFS::UriList const& list); 

        ModifyTracks modify_tracks;

    private:

        void
        artist_changed	();

        void
        album_changed	();

        void
        refresh_releases	();
    
        // ColumnRecords
        class ColumnRecordArtist : public Gtk::TreeModel::ColumnRecord
        {
          public:
            Gtk::TreeModelColumn<Glib::ustring> name;
            Gtk::TreeModelColumn<Glib::ustring> id; // Artist ID
            ColumnRecordArtist ()
            {
              add (name);
              add (id);
            }
        };
        ColumnRecordArtist artists;

        class ColumnRecordAlbum : public Gtk::TreeModel::ColumnRecord
        {
          public:
            Gtk::TreeModelColumn<Glib::ustring> name;
            Gtk::TreeModelColumn<Glib::ustring> id; // Release ID
            ColumnRecordAlbum ()
            {
              add (name);
              add (id);
            }
        };
        ColumnRecordAlbum albums;

        class ColumnRecordTrack : public Gtk::TreeModel::ColumnRecord
        {
          public:
            Gtk::TreeModelColumn<unsigned int>                          tracknr;
            Gtk::TreeModelColumn<Glib::ustring>                         title;
            Gtk::TreeModelColumn<Bmp::MusicBrainzXML::MusicBrainzTrack> mb_track;
            ColumnRecordTrack ()
            {
              add (tracknr);
              add (title);
              add (mb_track);
            }
        };
        ColumnRecordTrack tracks_cr;

        class ColumnRecordUpdateTrack : public Gtk::TreeModel::ColumnRecord
        {
          public:
            Gtk::TreeModelColumn<bool>	        mod;
            Gtk::TreeModelColumn<unsigned int>  tracknr;
            Gtk::TreeModelColumn<Glib::ustring> title;
            Gtk::TreeModelColumn<Glib::ustring> location;
            ColumnRecordUpdateTrack ()
            {
              add (mod);
              add (tracknr);
              add (title);
              add (location);
            }
        };
        ColumnRecordUpdateTrack update_tracks_cr;



        typedef std::map < Glib::ustring, MusicBrainzXML::MusicBrainzRelease > ReleaseMap;
        ReleaseMap merge_releases;

        typedef std::vector< Glib::ustring >          ReleaseIds;
        typedef std::map< Glib::ustring, ReleaseIds > ArtistReleaseMap; 
        ArtistReleaseMap merge_artist_release_map;

        Glib::RefPtr<Gnome::Glade::Xml> ref_xml;

        Gtk::ComboBoxEntry	    *merge_cb_artist;
        Gtk::ComboBoxEntry	    *merge_cb_album;

        Gtk::Entry		          *merge_cb_artist_entry;
        Gtk::Entry		          *merge_cb_album_entry;

        Glib::RefPtr<Gtk::ListStore>   merge_cb_artist_store;
        Glib::RefPtr<Gtk::ListStore>   merge_cb_album_store;

        Gtk::Image		                *merge_coverart;
        Glib::RefPtr<Gdk::Pixbuf>	     disc;

        Gtk::Entry		    *merge_artist_id;
        Gtk::Entry		    *merge_release_id;
        Gtk::Entry		    *merge_asin;
        Gtk::Entry		    *merge_score;
        Gtk::Label		    *l_tracks;
        Gtk::Button		    *b_refresh;

        Gtk::Button		    *cancelbutton1;
        Gtk::Button		    *okbutton2;

        Gtk::TreeView		  *view_release;
        Gtk::TreeView		  *view_local;
        Gtk::Button		    *b_move_up;
        Gtk::Button		    *b_move_down;

        Glib::RefPtr<Gtk::ListStore>   store_release;
        Glib::RefPtr<Gtk::ListStore>   store_local;

        Bmp::VFS::UriList m_list; 
        MusicBrainzXML::MusicBrainzTracks tracks; 

        void
        cell_data_func_nr (Gtk::CellRenderer* cell, const Gtk::TreeModel::iterator& iter);
        void
        local_selection_changed ();
        void
        on_button_down ();
        void
        on_button_up   ();
        void
        on_mod_toggled    (Glib::ustring const& path);
  };
}
#endif


