#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <boost/format.hpp>

#include "file_utils.hpp"
#include "util.hpp"
#include "paths.hpp"
#include "main.hpp"

#include "musicbrainz_xml.hpp"
#include "library_ui_modify_albums.hpp"

#include "x_library.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

namespace
{
  void
  cbe_set_sensitive (Gtk::CheckButton *cb, Gtk::ComboBoxEntry *cbe)
  {
      cbe->set_sensitive (cb->get_active());
  }

  void
  fcb_set_sensitive (Gtk::CheckButton *cb, Gtk::FileChooserButton *fcb)
  {
      fcb->set_sensitive (cb->get_active());
  }
}

namespace Bmp
{
      LibraryUIModifyAlbums::LibraryUIModifyAlbums (BaseObjectType                        *cobject,
			                                              const Glib::RefPtr<Gnome::Glade::Xml> &xml) : Gtk::Dialog (cobject), ref_xml (xml),
                                                                                                  disc (Gdk::Pixbuf::create_from_file
                                                                                                  (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "audio.png"))
      {
          ref_xml->get_widget ("merge_cb_artist",   merge_cb_artist);
          ref_xml->get_widget ("merge_cb_album",    merge_cb_album);
          ref_xml->get_widget ("merge_coverart",    merge_coverart);
          ref_xml->get_widget ("merge_album",	      merge_album);
          ref_xml->get_widget ("merge_artist",	    merge_artist);
          ref_xml->get_widget ("merge_artistid",   merge_artistid);
          ref_xml->get_widget ("merge_ASIN",	      merge_ASIN);
          ref_xml->get_widget ("merge_release_ID",  merge_release_ID);
          ref_xml->get_widget ("merge_score",	      merge_score);
          ref_xml->get_widget ("b_refresh",	        b_refresh);
          ref_xml->get_widget ("l_tracks",	        l_tracks);

          merge_cb_artist_entry = merge_cb_artist->get_entry ();
          merge_cb_album_entry  = merge_cb_album->get_entry ();

          merge_cb_artist_entry->set_activates_default (true);
          merge_cb_album_entry->set_activates_default (true);

          merge_cb_artist_entry->set_text (_("<select here>"));
          merge_cb_album_entry->set_text ("");
          merge_coverart->set (disc);

          b_refresh->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyAlbums::refresh_releases));
      }

      LibraryUIModifyAlbums*
      LibraryUIModifyAlbums::create ()
      {
        const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "library_ui_modify_albums.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        LibraryUIModifyAlbums *p = 0;
        glade_xml->get_widget_derived ("dialog_modify_albums", p);
        return p;
      }

      void
      LibraryUIModifyAlbums::refresh_releases ()
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;

          set_sensitive (false);
          
          MusicBrainzXML::MusicBrainzReleases mb_releases; 
          MusicBrainzXML::get_releases (merge_cb_artist_entry->get_text (),
                     merge_cb_album_entry->get_text (), mb_releases);

          merge_cb_artist_store = ListStore::create (artists);
          merge_cb_album_store  = ListStore::create (albums);
          merge_artist_release_map.clear ();

          for (MusicBrainzReleases::const_iterator r_Iter     = mb_releases.begin (),
                                          r_IterEnd  = mb_releases.end ();
                                          r_Iter    != r_IterEnd; ++r_Iter)
          {

            MusicBrainzRelease const& release = (*r_Iter);
            merge_releases.insert (std::make_pair (release.mb_release_id, release));
            ArtistReleaseMap::iterator map_Iter = merge_artist_release_map.find (release.mb_artist_id);
            if (map_Iter != merge_artist_release_map.end ())
            {
              ReleaseIds &ids = map_Iter->second;
              ids.push_back (release.mb_release_id);
            }
            else
            {
              TreeModel::iterator m_Iter  = merge_cb_artist_store->append ();
              (*m_Iter)[artists.name]	  = release.mb_artist_name;
              (*m_Iter)[artists.id]	  = release.mb_artist_id;
              ReleaseIds ids; 
              ids.push_back (release.mb_release_id);
              merge_artist_release_map.insert (std::make_pair (release.mb_artist_id, ids)); 
            }
          }
          merge_cb_artist->set_model (merge_cb_artist_store);
          merge_cb_album->set_model (merge_cb_album_store);

          merge_cb_artist_entry->set_text (_("<select here>"));
          merge_cb_album_entry->set_text ("");

          set_sensitive (true);
      }

      int
      LibraryUIModifyAlbums::run (const MusicBrainzXML::MusicBrainzReleases& releases_)
      {
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;

          merge_cb_artist_store = ListStore::create (artists);
          merge_cb_album_store  = ListStore::create (albums);

          for (MusicBrainzReleases::const_iterator r_Iter     = releases_.begin (),
                   r_IterEnd  = releases_.end ();
                   r_Iter    != r_IterEnd; ++r_Iter)
          {

            MusicBrainzRelease const& release = (*r_Iter);
            merge_releases.insert (std::make_pair (release.mb_release_id, release));
        
            ArtistReleaseMap::iterator map_Iter = merge_artist_release_map.find (release.mb_artist_id);
            if (map_Iter != merge_artist_release_map.end ())
              {
                ReleaseIds &ids = map_Iter->second;
                ids.push_back (release.mb_release_id);
              }
            else
              {
                TreeModel::iterator m_Iter  = merge_cb_artist_store->append ();

                (*m_Iter)[artists.name]	  = release.mb_artist_name;
                (*m_Iter)[artists.id]	  = release.mb_artist_id;
              
                ReleaseIds ids; 
                ids.push_back (release.mb_release_id);
                merge_artist_release_map.insert (std::make_pair (release.mb_artist_id, ids)); 
              }
          }

          // Setup 'Artist' ComboBox 
          merge_cb_artist->set_model (merge_cb_artist_store);
          merge_cb_artist->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyAlbums::artist_changed));
          merge_cb_artist->set_text_column (0);

          // Setup 'Album' ComboBox 
          merge_cb_album->set_model (merge_cb_album_store);
          merge_cb_album->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyAlbums::album_changed));
          merge_cb_album->set_text_column (0);

          return Gtk::Dialog::run ();
      }

      int
      LibraryUIModifyAlbums::rerun ()
      {
	      return Gtk::Dialog::run ();
      }

      void
      LibraryUIModifyAlbums::album_changed ()
      {
          using namespace Gtk;
          using namespace MusicBrainzXML;

          if (merge_cb_album->get_active_row_number () < 0) return;

          set_sensitive (false);
        
          // Get releases from selected artist
          TreeModel::const_iterator m_Iter  = merge_cb_album->get_active ();
          ReleaseMap::iterator r_Iter       = merge_releases.find ((*m_Iter)[albums.id]);
          MusicBrainzRelease const& release = r_Iter->second;

          try
            {
              Glib::RefPtr<Gdk::Pixbuf> cover;
              Amazon::get_cover_by_asin (release.asin, cover);
              merge_coverart->set (cover);
            }
          catch (...)
            { 
              merge_coverart->set (disc);
            }

          merge_album->set_text       (release.mb_release_name);
          merge_artist->set_text      (release.mb_artist_name);
          merge_artistid->set_text    (release.mb_artist_id);
          merge_release_ID->set_text  (release.mb_release_id);
          merge_ASIN->set_text        (release.asin);
          merge_score->set_text       ((boost::format ("%d") % release.score).str());
          l_tracks->set_text          ((boost::format ("%d") % release.track_list).str());
          selected = release;
          set_sensitive (true);
      }

      void
      LibraryUIModifyAlbums::artist_changed ()
      {
          using namespace Gtk;
          using namespace MusicBrainzXML;

          if (merge_cb_artist->get_active_row_number () < 0) return;

          set_sensitive (false);
        
          // Get releases from selected artist
          TreeModel::const_iterator  m_Iter   = merge_cb_artist->get_active ();
          ArtistReleaseMap::iterator map_Iter = merge_artist_release_map.find ((*m_Iter)[artists.id]);
          ReleaseIds                 &ids     = map_Iter->second;

          merge_cb_album_store->clear ();

          for (ReleaseIds::const_iterator r_Iter    = ids.begin (),
                                          r_IterEnd = ids.end ();
                                          r_Iter != r_IterEnd; ++r_Iter)
            {
              // Iter from releases map
              ReleaseMap::iterator map_Iter = merge_releases.find (*r_Iter);

              // Get the release based on id 
              MusicBrainzRelease const& release = map_Iter->second; 

              TreeModel::iterator c_Iter  = merge_cb_album_store->append ();
              (*c_Iter)[albums.name]      = release.mb_release_name;
              (*c_Iter)[albums.id]        = release.mb_release_id;
            }
          set_sensitive (true);
      }
}
