#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <boost/format.hpp>
#include <mcs/mcs.h>

#include "file_utils.hpp"
#include "util.hpp"
#include "paths.hpp"
#include "main.hpp"
#include "uri++.hpp"

#include "library_ui_relocate_album.hpp"
#include "x_library.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

namespace
{
  void
  cbe_set_sensitive (Gtk::CheckButton *cb, Gtk::ComboBoxEntry *cbe)
  {
      cbe->set_sensitive (cb->get_active());
  }

  void
  fcb_set_sensitive (Gtk::CheckButton *cb, Gtk::FileChooserButton *fcb)
  {
      fcb->set_sensitive (cb->get_active());
  }

  bool
  folder_writeable (const std::string& foldername)
  {
    struct stat info;
    if (stat (foldername.c_str (), &info))
      {
        return false;
      }

    // Get stat info
    uid_t  stat_uid  = info.st_uid;
    gid_t  stat_gid  = info.st_gid;
    mode_t stat_mode = info.st_mode;

    // Check if world-writable
    if (stat_mode & S_IWOTH)
      {
        return true;
      } 

    // Get the process uid/gid
    uid_t p_uid = geteuid ();
    gid_t p_gid = getegid ();

    // Check whether group has write permission
    if ((p_gid == stat_gid) && (stat_mode & S_IWGRP))
      {
        return true;
      }

    // Check whether user has write permission
    if ((p_uid == stat_uid) && (stat_mode & S_IWUSR))
      {  
        return true;
      }

    return false;
  }
}

namespace Bmp
{
      LibraryUIMoveAlbum::~LibraryUIMoveAlbum ()
      {
        mcs->key_set<std::string>("bmp","file-chooser-path-relocate-album", fcb_move->get_current_folder ());
      }

      LibraryUIMoveAlbum::LibraryUIMoveAlbum (BaseObjectType*                       cobject,
                                              const Glib::RefPtr<Gnome::Glade::Xml> &xml)
            : Gtk::Dialog (cobject),
              ref_xml     (xml),
              disc        (Gdk::Pixbuf::create_from_file (BMP_IMAGE_DIR G_DIR_SEPARATOR_S "audio.png"))
      {
          ref_xml->get_widget ("coverart",
                                coverart);

          ref_xml->get_widget ("e_album",
                                e_album);

          ref_xml->get_widget ("e_artist",
                                e_artist);

          ref_xml->get_widget ("cb_append_albumname",
                                cb_append_albumname);

          ref_xml->get_widget ("fcb_move",
                                fcb_move);
          ref_xml->get_widget ("hbox_warning",
                                hbox_warning);
          ref_xml->get_widget ("okbutton2",
                                okbutton2);

          fcb_move->signal_current_folder_changed ().connect (sigc::mem_fun (*this, &LibraryUIMoveAlbum::fcb_current_folder_changed));
          fcb_move->set_current_folder (mcs->key_get<std::string>("bmp","file-chooser-path-relocate-album"));
      }

      void
      LibraryUIMoveAlbum::fcb_current_folder_changed ()
      {
          GSList *filenames = gtk_file_chooser_get_filenames (GTK_FILE_CHOOSER (fcb_move->gobj())); 
          bool  writable = folder_writeable ((char*)filenames->data);
          g_slist_free (filenames);
  
          if (writable)
              hbox_warning->hide ();
          else
              hbox_warning->show_all ();

          okbutton2->set_sensitive (writable);
      }

      LibraryUIMoveAlbum*
      LibraryUIMoveAlbum::create ()
      {
        const std::string path (BMP_GLADE_DIR G_DIR_SEPARATOR_S "library_ui_relocate_album.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        LibraryUIMoveAlbum *library_ui = 0;
        glade_xml->get_widget_derived ("dialog_relocate_album", library_ui);
        return library_ui;
      }

      int
      LibraryUIMoveAlbum::run (Glib::ustring const& artist, Glib::ustring const& album) 
      {
          using namespace Glib;
          using namespace Gtk;
          using namespace Bmp::DB;
          using namespace Bmp::Library;
          using namespace std;

          AttributeList attrs; 
          string id_artist = library->get_metadatum_id (DATUM_ARTIST);
          string id_album  = library->get_metadatum_id (DATUM_ALBUM);
          attrs.push_back (Attribute (DB::EXACT, id_artist, artist));
          attrs.push_back (Attribute (DB::EXACT, id_album,  album));

          library->query (attrs, vector);
          if (vector.empty()) return GTK_RESPONSE_CANCEL;

          e_artist->set_text (artist);
          e_album->set_text (album);

          try {
            RefPtr<Gdk::Pixbuf> _cover;
            ustring asin = library->get_asin_for_album (artist, album);
            Amazon::get_cover_by_asin (asin, _cover); 
            coverart->set (_cover);
          } catch (...) {
            coverart->set (disc);
          }

          return Gtk::Dialog::run ();
      }
}
