#include <glibmm.h>
#include <glib/gi18n.h>

#include <ne_basic.h>
#include <ne_compress.h>
#include <ne_socket.h>
#include <ne_utils.h>
#include <ne_props.h>
#include <ne_session.h>
#include <ne_request.h>
#include <ne_uri.h>

#include <vector>

#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/replace.hpp>

#include <iostream>
#include <sstream>
#include <fstream>

#include "util.hpp"
#include "uri++.hpp"
#include "xml.hpp"
#include "main.hpp"
#include "paths.hpp"

#include "lyrics.hpp"
#include "x_vfs.hpp"

namespace
{
    const char* param[] = {
      "act",
      "cln",
      "clv",
      "ar",
      "ti",
      "al",
      "co",
      "yr",
      "txt",
      "and",
      "id",
    };

    enum ParamID
    {
      LT_ACTION,
      LT_CLIENT,
      LT_CLVERSION,
      LT_ARTIST,
      LT_TITLE,
      LT_ALBUM,
      LT_COMPOSER,
      LT_YEAR,
      LT_TEXT,
      LT_AND,
      LT_ID,
    };
}

namespace Bmp
{
  namespace Lyrictracker
  {
    bool get_lyrics (const Bmp::Library::Track& track,
                     Glib::ustring&	            lyrics,
                     bool		                    cached_only)
    {
      using namespace Glib;
      using namespace std;

      static const string LT_BASE_URI = "http://lyrictracker.com/soap.php";
      if (!(track.artist && track.title)) return false;

      string data = track.artist.get().casefold().raw() + string("_") + track.title.get().casefold().raw();
      string md5_hex = Util::md5_hex ((char* const)data.c_str(), strlen(data.c_str()));
      string text_path = build_filename (BMP_PATH_LYRIC_CACHE_DIR, md5_hex + ustring(".txt")); 

      if (file_test (text_path, FILE_TEST_EXISTS))
        { 
          try {
              lyrics = file_get_contents (text_path);
              return true;
            }
          catch (...) { return false; }
        }
  
      if (cached_only) return false;

      string _title  = track.title.get().raw(); 
      string _artist = track.artist.get().raw();

      boost::replace_all (_title, "'", "\\'");
      boost::replace_all (_artist, "'", "\\'");

      ustring uri_query_id;
      uri_query_id 
        .append( LT_BASE_URI )
        .append( "?" )
        .append( param[LT_CLIENT] )
        .append( "=bmp&" )
        .append( param[LT_CLVERSION] )
        .append( "=0.20&" )
        .append( param[LT_TITLE] )
        .append( "=" )
        .append( _title )
        .append( "&" )
        .append( param[LT_ARTIST] )
        .append( "=" )
        .append( _artist )
        .append( "&" )
        .append( param[LT_ACTION] )
        .append( "=query&" )
        .append( param[LT_AND] )
        .append( "=1" );

      Bmp::VFS::Handle handle_id;	
      try 
        {
          Bmp::URI u_query_id (uri_query_id); 
          u_query_id.escape ();
          handle_id = Bmp::VFS::Handle::create (ustring(u_query_id));
          vfs->read (handle_id, Bmp::VFS::TRANSPORT);
        }
      catch (...) { return false; }

      if (!handle_id.get_buffer()) return false;

      xmlDocPtr doc = 0;
      xmlXPathObjectPtr xpathobj = 0;
      xmlNodeSetPtr nv = 0;

      doc = xmlParseMemory (reinterpret_cast<const char*>(handle_id.get_buffer()), handle_id.get_buffer_size());
      if (!doc) return false;

      xpathobj = xml_execute_xpath_expression (doc, BAD_CAST "//result", NULL);
      nv = xpathobj->nodesetval;
      if ((!nv) || (!nv->nodeNr)) return false;
      xmlChar *id;
      id = xmlGetProp (nv->nodeTab[0], BAD_CAST "id");
      if (!id) return false;

      ustring uri_query_lyrics;
      uri_query_lyrics
        .append( LT_BASE_URI )
        .append( "?" )
        .append( param[LT_CLIENT] ) 
        .append( "=bmp&" ) 
        .append( param[LT_CLVERSION] )
        .append( "=0.20&" )
        .append( param[LT_ID] )
        .append( "=" ) 
        .append( (const char*)id )
        .append( "&" )
        .append( param[LT_ACTION] )
        .append( "=detail" );

      free (id);
 
      Bmp::VFS::Handle handle_lyrics; 
      try  
        {
          Bmp::URI u_query_lyrics (uri_query_lyrics);
          u_query_lyrics.escape ();
          handle_lyrics = Bmp::VFS::Handle::create (ustring(u_query_lyrics));
          vfs->read (handle_lyrics, Bmp::VFS::TRANSPORT);
        }
      catch (...) { return false; }
        
      if (!handle_lyrics.get_buffer()) return false;

      lyrics = reinterpret_cast<const char*>(handle_lyrics.get_buffer ());

      try
        {
          std::ofstream outfile;      
          outfile.open (text_path.c_str ());
          outfile << std::string(lyrics); 
          outfile.close ();
        }
      catch (ConvertError& cxe)
        {
          g_warning ("%s: Unable to save lyrics for %s", G_STRLOC, track.location.get().c_str());
        }
      return true;
    }
  }
}
