//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <glibmm/markup.h>
#include <iostream>

#include "uri++.hpp"
#include "debug.hpp"
#include "library.hpp"

#include "x_vfs.hpp"

#include "musicbrainz_xml.hpp"

namespace Bmp
{
  namespace MusicBrainzXML
  {
      void
      mb_track_to_update_track (MusicBrainzTrack const& track, Bmp::Library::UpdateTrack &update_track)
      {
        using namespace Bmp::Library;

        update_track.artist              = track.mb_artist;
        update_track.title               = track.mb_title;
        update_track.mb_track_id         = track.mb_track_id;
        update_track.mb_album_id         = track.mb_release_id;
        update_track.mb_artist_id        = track.mb_artist_id;
        update_track.mb_artist_sort_name = track.mb_artist_sort_name;
        update_track.tracknumber        = track.tracknum;
      }

	    void
	    get_releases (Glib::ustring const& artist,
                    Glib::ustring const& album,
                    MusicBrainzReleases& releases)
	    {
        using namespace Glib;

	      ustring uri;

	      uri .append ("http://musicbrainz.org/ws/1/release/?type=xml&title=")
            .append (album)
            .append ("&artist=")
            .append (artist)
            .append ("&releasetypes=Official&limit=25");

	      Bmp::URI u (uri);
	      u.escape ();
	      Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (Glib::ustring(u));

	      try {
          vfs->read (handle, Bmp::VFS::TRANSPORT);
          MusicBrainzReleaseParser parser;
          Glib::Markup::ParseContext context (parser);
          Glib::ustring data = reinterpret_cast<const char *>(handle.get_buffer()); 	  
          context.parse (data);
          context.end_parse ();
          releases = parser.releases;
	      }
	      catch (Bmp::VFS::Exception& cxe)
	      {
          //FIXME: VFS exception
	      }
	      catch (Glib::ConvertError& cxe)
	      {
    		  //FIXME: Glib convert error (ustring) 
	      }
	      catch (Glib::MarkupError& cxe)
	      {
    		  //FIXME: Glib Markup error
	      }
	    }
  
      void
      MusicBrainzReleaseParser::on_start_element  (Glib::Markup::ParseContext& context,
                                                   Glib::ustring const& elementname,
                                                   AttributeMap const& attributes)
	    {
          if (elementname == "release")
          {
            state |= E_RELEASE;
            current = MusicBrainzRelease();
            current.mb_release_id = (*(attributes.find ("id"))).second;
            current.score = atoi ((*(attributes.find ("ext:score"))).second.c_str());
            debug ("musicbrainz-xml", "<release id='%s' score='%d'>", current.mb_release_id.c_str(), current.score);
            return;
          }

          if (elementname == "title")
          {
            state |= E_TITLE;
            debug ("musicbrainz-xml", "<title>");
            return;
          }

          if (elementname == "asin")
          {
            state |= E_ASIN;
            debug ("musicbrainz-xml", "<asin>");
            return;
          }

          if (elementname == "artist")
          {
            state |= E_ARTIST;
            current.mb_artist_id = (*(attributes.find ("id"))).second;
            debug ("musicbrainz-xml", "<artist id='%s'>", current.mb_artist_id.c_str());
            return;
          }

          if (elementname == "name")
          {
            state |= E_NAME;
            debug ("musicbrainz-xml", "<name>");
            return;
          }

          if (elementname == "disc-list")
          {
            state |= E_DISC_LIST;
            current.disc_list = atoi ((*(attributes.find ("count"))).second.c_str());
            debug ("musicbrainz-xml", "<disc-list count='%d'>", current.disc_list);
            return;
          }

          if (elementname == "track-list")
          {
            state |= E_TRACK_LIST;
            current.track_list = atoi ((*(attributes.find ("count"))).second.c_str());
            debug ("musicbrainz-xml", "<track-list count='%d'>", current.track_list);
            return;
          }
	    }

      void
      MusicBrainzReleaseParser::on_end_element (Glib::Markup::ParseContext& context,
                                                Glib::ustring const& elementname)
	    {
          if (elementname == "release")
          {
            state &= ~ E_RELEASE;
            debug ("musicbrainz-xml", "</release>");

            releases.push_back (current); 

            if (state)
            {
              g_warning (G_STRLOC ": State should be 0, but is %d", state);
              state = 0;
            }
            return;
          }

          if (elementname == "title")
          {
            state &= ~ E_TITLE;
            debug ("musicbrainz-xml", "</title>");
            return;
          }

          if (elementname == "asin")
          {
            state &= ~ E_ASIN;
            debug ("musicbrainz-xml", "</asin>");
            return;
          }

          if (elementname == "artist")
          {
            state &= ~ E_ARTIST;
            debug ("musicbrainz-xml", "</artist>");
            return;
          }

          if (elementname == "name")
          {
            state &= ~ E_NAME;
            debug ("musicbrainz-xml", "</name>");
            return;
          }

          if (elementname == "disc-list")
          {
            state &= ~ E_DISC_LIST;
            debug ("musicbrainz-xml", "</disc-list>");
            return;
          }

          if (elementname == "track-list")
          {
            state &= ~ E_TRACK_LIST;
            debug ("musicbrainz-xml", "</track-list>");
            return;
          }
	    }

      void
      MusicBrainzReleaseParser::on_text   (Glib::Markup::ParseContext& context,
                                           Glib::ustring const& text)
	    {
          if (state & E_TITLE)
          {
            current.mb_release_name = text;
            debug ("musicbrainz-xml", "%s", text.c_str());
            return;
          }

          if (state & E_ASIN)
          {
            current.asin = text;
            debug ("musicbrainz-xml", "%s", text.c_str());
            return;
          }

          if ( (state & E_ARTIST) && (state & E_NAME) )
          {
            current.mb_artist_name = text;
            debug ("musicbrainz-xml", "%s", text.c_str());
            return;
          }

	    }

      void
      MusicBrainzReleaseParser::on_passtrough     (Glib::Markup::ParseContext& context,
                                                   Glib::ustring const& text)
	    {
	    }

      void
      MusicBrainzReleaseParser::on_error          (Glib::Markup::ParseContext& context,
                                                   const Glib::MarkupError& error)
	    {
	    }

    }
}

namespace Bmp
{
  namespace MusicBrainzXML
  {
	    void
	    get_tracks (Glib::ustring const& mb_release_id,
			            MusicBrainzTracks&   tracks)
	    {
        using namespace Glib;

	      ustring uri;

	      uri .append ("http://musicbrainz.org/ws/1/release/")
            .append (mb_release_id)
            .append ("?type=xml&inc=tracks artist");

	      Bmp::URI u (uri);
	      u.escape ();
	      Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (ustring(u));
	      try
          {
            vfs->read (handle, Bmp::VFS::TRANSPORT);
            MusicBrainzTrackParser parser (tracks);
            Markup::ParseContext context (parser);

            ustring data = reinterpret_cast<const char *>(handle.get_buffer()); 	  
            context.parse (data);
            context.end_parse ();

            if (parser.state)
              {
                g_warning (G_STRLOC ": State should be 0, but is %d", parser.state);
              }
          }
        catch (Bmp::VFS::Exception& cxe)
          {
            //FIXME: VFS exception
          }
       catch (ConvertError& cxe)
          {
            //FIXME: Glib convert error (ustring) 
          }
       catch (MarkupError& cxe)
          {
            //FIXME: Glib Markup error
          }
	    }
  
    void
    MusicBrainzTrackParser::on_start_element  (Glib::Markup::ParseContext& context,
                                               Glib::ustring const& elementname,
                                               const AttributeMap& attributes)
	    {
          if (elementname == "release")
            {
              state |= E_RELEASE;
              release_id = (*(attributes.find ("id"))).second;
              return;
            }

          if (elementname == "asin")
            {
              state |= E_ASIN;
              return;
            }

          if (elementname == "artist")
            {
              state |= E_ARTIST;

              if (state & E_TRACK)
                {
                  current.mb_artist_id = (*(attributes.find ("id"))).second;
                } 
              else
                {
                  album_artist_id = (*(attributes.find ("id"))).second;
                }
              return;
            }

          if (elementname == "title")
            {
              state |= E_TITLE;
              return;
            }

          if (elementname == "name")
            {
              state |= E_NAME;
              return;
            }

          if (elementname == "sort-name")
            {
              state |= E_SORT_NAME;
              return;
            }

          if (elementname == "track")
            {
              state |= E_TRACK;
              current = MusicBrainzTrack();
              current.mb_track_id = (*(attributes.find ("id"))).second;
              current.tracknum = tracknum++;
              return;
            }
	    }

    void
    MusicBrainzTrackParser::on_end_element    (Glib::Markup::ParseContext& context,
                                               Glib::ustring const& elementname)
	    {
          if (elementname == "release")
            {
              state &= ~ E_RELEASE;
              tracks.mb_release_id              = release_id;
              tracks.asin                       = asin;
              tracks.mb_album_name              = album_name;
              tracks.mb_album_artist            = album_artist;
              tracks.mb_album_artist_id         = album_artist_id;
              tracks.mb_album_artist_sort_name  = album_artist_sort_name;
              return;
            }

          if (elementname == "asin")
            {
              state &= ~ E_ASIN;
              return;
            }

          if (elementname == "artist")
            {
              state &= ~ E_ARTIST;
              return;
            }

          if (elementname == "title")
            {
              state &= ~ E_TITLE;
              return;
            }

          if (elementname == "name")
            {
              state &= ~ E_NAME;
              return;
            }

          if (elementname == "sort-name")
            {
              state &= ~ E_SORT_NAME;
              return;
            }

          if (elementname == "track")
            {
              state &= ~ E_TRACK;
              current.mb_release_id = release_id; 
              tracks.tracks.push_back (current); 
              return;
            }
	    }

    void
    MusicBrainzTrackParser::on_text  (Glib::Markup::ParseContext& context,
                                      Glib::ustring const& text)
	    {
          if (state & E_ASIN)
            {
              asin = text;
              return;
            }

          if (state & E_TITLE)
            {
              if (state & E_TRACK)
                {
                  current.mb_title = text;
                }
              else
                {
                  album_name = text;
                }
              return;
            }

          if (state & E_NAME)
            {
              if (state & E_TRACK)
                {
                  current.mb_artist = text;
                }
              else
                {
                  album_artist = text;
                }
              return;
            }

          if (state & E_SORT_NAME)
            {
              if (state & E_TRACK)
                {
                  current.mb_artist_sort_name = text;
                }
              else
                {
                  album_artist_sort_name = text;
                }
              return;
            }
	    }

    void
    MusicBrainzTrackParser::on_passtrough (Glib::Markup::ParseContext& context,
                                           Glib::ustring const& text)
	    {
	    }

    void
    MusicBrainzTrackParser::on_error      (Glib::Markup::ParseContext& context,
                                           Glib::MarkupError const& error)
	    {
	    }
  }
}

