//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_MBWS_HPP 
#define BMP_MBWS_HPP

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <glibmm/markup.h>

#include <boost/format.hpp>
#include <boost/optional.hpp>

#include "library.hpp"

namespace Bmp
{
  namespace MusicBrainzXML
  {

      //////////////////// Tracks for a Release

      struct MusicBrainzTrack
      { 
        boost::optional<Glib::ustring>	mb_artist;
        boost::optional<Glib::ustring>  mb_artist_id;
        boost::optional<Glib::ustring>	mb_artist_sort_name;

        Glib::ustring	                  mb_title;
        Glib::ustring	                  mb_track_id;
        Glib::ustring                   mb_release_id;

        int                             tracknum;

        MusicBrainzTrack () : tracknum (0) {};
      };

      typedef std::vector< MusicBrainzTrack > Tracks;

      struct MusicBrainzTracks
      {
          Tracks tracks;

          Glib::ustring  mb_release_id;
          Glib::ustring  mb_album_name;
          Glib::ustring	 mb_album_artist;
          Glib::ustring  mb_album_artist_id;

          boost::optional<Glib::ustring>  asin;
          boost::optional<Glib::ustring>	mb_album_artist_sort_name;
      };


      class MusicBrainzTrackParser : public Glib::Markup::Parser
      {
          public:

            MusicBrainzTrackParser (MusicBrainzTracks &tracks) : tracks (tracks), state (0), tracknum (1) {}
            virtual ~MusicBrainzTrackParser () {}

            MusicBrainzTracks &tracks;
            int                state;

          protected:

            void
            on_start_element  (Glib::Markup::ParseContext& context,
                               Glib::ustring const& elementname,
                               const AttributeMap& attributes);
            void
            on_end_element    (Glib::Markup::ParseContext& context,
                               Glib::ustring const& elementname);
            void
            on_text	          (Glib::Markup::ParseContext& context,
                               Glib::ustring const& text);
            void
            on_passtrough     (Glib::Markup::ParseContext& context,
                               Glib::ustring const& text);
            void
            on_error	        (Glib::Markup::ParseContext& context,
                               const Glib::MarkupError& error);
          private:

              enum Element
              {
                E_NONE	            = 0, 
                E_RELEASE           = 1 << 0,
                E_TRACK	            = 1 << 1,
                E_TITLE	            = 1 << 2,
                E_ARTIST	          = 1 << 3,
                E_NAME	            = 1 << 4,
                E_SORT_NAME         = 1 << 5,
                E_ASIN              = 1 << 6,
              };

              MusicBrainzTrack    current;
              int		              tracknum;

              Glib::ustring release_id;
              Glib::ustring album_name;
              Glib::ustring album_artist;
              Glib::ustring album_artist_id;

              boost::optional<Glib::ustring> album_artist_sort_name;
              boost::optional<Glib::ustring> asin;
      };

      void
      get_tracks   (Glib::ustring const&  releaseid,
                    MusicBrainzTracks&    tracks);
  }
}

namespace Bmp
{
  namespace MusicBrainzXML
  {
      //////////////////// Release

      struct MusicBrainzRelease
      { 
        Glib::ustring	mb_release_id;
        Glib::ustring	mb_release_name;
        Glib::ustring	mb_artist_id;
        Glib::ustring	mb_artist_name;
        Glib::ustring	asin;

        unsigned int	score;
        unsigned int	disc_list;
        unsigned int	track_list;

        MusicBrainzRelease () : score (0), disc_list (0), track_list (0) {}
      };

      typedef std::vector< MusicBrainzRelease > MusicBrainzReleases;

      class MusicBrainzReleaseParser : public Glib::Markup::Parser
      {
          public:

            MusicBrainzReleaseParser () : state (0) {}
            virtual ~MusicBrainzReleaseParser () {}

            MusicBrainzReleases releases;

          protected:

            void
            on_start_element  (Glib::Markup::ParseContext& context,
                               Glib::ustring const& elementname,
                               const AttributeMap& attributes);
            void
            on_end_element    (Glib::Markup::ParseContext& context,
                               Glib::ustring const& elementname);
            void
            on_text	          (Glib::Markup::ParseContext& context,
                               Glib::ustring const& text);
            void
            on_passtrough     (Glib::Markup::ParseContext& context,
                               Glib::ustring const& text);
            void
            on_error	        (Glib::Markup::ParseContext& context,
                               const Glib::MarkupError& error);
          private:

            enum Element
            {
              E_NONE	      = 1 << 0,
              E_RELEASE	    = 1 << 1,
              E_TITLE	      = 1 << 2,
              E_ASIN	      = 1 << 3,
              E_ARTIST	    = 1 << 4,
              E_NAME	      = 1 << 5,
              E_DISC_LIST   = 1 << 6,
              E_TRACK_LIST  = 1 << 7,
              E_SORT_NAME   = 1 << 8,
            };

            MusicBrainzRelease  current;
            int                 state;
        };

        void
        get_releases (Glib::ustring const& artist,
                      Glib::ustring const& album,
                      MusicBrainzReleases& releases);
        
    }
}

namespace Bmp
{
  namespace MusicBrainzXML
  {
        void 
        mb_track_to_update_track (MusicBrainzTrack const& track, Bmp::Library::UpdateTrack &update_track);
  }
}
#endif
