//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <cstring>

#include <iostream>
#include <sstream>

#include <ne_basic.h>
#include <ne_compress.h>
#include <ne_socket.h>
#include <ne_utils.h>
#include <ne_props.h>
#include <ne_session.h>
#include <ne_request.h>
#include <ne_uri.h>

#include "util.hpp"
#include "uri++.hpp"

#include <build.h>
#include <revision.h>

#define SERVER_TEST_NAME  "google.com" 
#define SERVER_NAME       "beep-media-player.org" 
#define SERVER_PORT        80

#include "network.hpp"

namespace
{
  bool connected = false;
  bool attempted_connect = false;
}

namespace Bmp
{
  namespace Network
  {
    void
    send_statistic (StatisticAction action)
    {
      std::stringstream          post;
      ne_session                *sess;
      ne_request                *reqs;
     
      const char *action_str = (action == ACTION_QUIT) ? "quit":"start"; 
      char *aux = ne_path_escape (BUILD_ARCH);
      post << "http://" << SERVER_NAME << "/stats.php" << "?version=" << VERSION << "&svnrev=" << RV_REVISION << "&system=" << aux << "&action=" << action_str;
      free (aux);

      ne_uri u;
      ne_uri_parse (post.str().c_str(), &u);
  
      sess = ne_session_create (u.scheme,
			        u.host, 
				80);	

      reqs = ne_request_create (sess, "GET", u.path); 
      ne_add_request_header (reqs, "Content-Type", "application/x-www-form-urlencoded");
      ne_begin_request (reqs);
      if (ne_discard_response (reqs))
      {
      	g_warning (G_STRLOC ": Couldn't send statistic. libneon error: %s", ne_get_error (sess));
      }
      ne_end_request (reqs);

      ne_uri_free (&u);
      ne_request_destroy (reqs);
      ne_session_destroy (sess);
    }

    bool
    is_connected (bool force)
    {
      int    sd, rc;
      struct sockaddr_in localAddr, servAddr;
      struct hostent *h;

      if (attempted_connect && !force) return connected;

      attempted_connect = true;
      h = gethostbyname (SERVER_TEST_NAME);
      if ( h == NULL )
      {
      	return false;
      }

      servAddr.sin_family = h->h_addrtype;
      std::memcpy((char *) &servAddr.sin_addr.s_addr, h->h_addr_list[0], h->h_length);
      servAddr.sin_port = htons (SERVER_PORT);

      // Create socket
      sd = socket (AF_INET, SOCK_STREAM, 0);
      if(sd<0)
      {
        return false;
      }

      // Bind Any Port Number
      localAddr.sin_family = AF_INET;
      localAddr.sin_addr.s_addr = htonl (INADDR_ANY);
      localAddr.sin_port = htons(0);
      
      rc = bind (sd, (struct sockaddr *) &localAddr, sizeof(localAddr));
      if ( rc < 0 )
      {
      	return false;
      }
				    
      // Connect To Server
      rc = connect (sd, (struct sockaddr *) &servAddr, sizeof(servAddr));
      if (rc < 0)
      {
      	return false;
      }

      shutdown (sd, SHUT_RDWR);
      
      connected = true;
      return true;
    }
  }
}

