//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_PLAYLIST_HPP
#define BMP_PLAYLIST_HPP

#include <time.h>
#include <glib-object.h>
#include <gtk/gtk.h>

#ifndef BMP_PLUGIN_BUILD
#  include "library.hpp"
#  include "database.hpp"
#endif

#include "interfaces.h"
#include "guid.hpp"
#include "vfs.hpp"
#include "playlist++.hpp"

#define BMP_TYPE_PLAYLIST                 (bmp_playlist_get_type ())
#define BMP_PLAYLIST(obj)                 (G_TYPE_CHECK_INSTANCE_CAST ((obj), BMP_TYPE_PLAYLIST, BmpPlaylist))
#define BMP_PLAYLIST_CLASS(klass)         (G_TYPE_CHECK_CLASS_CAST ((klass), BMP_TYPE_PLAYLIST, BmpPlaylistClass))
#define BMP_IS_PLAYLIST(obj)              (G_TYPE_CHECK_INSTANCE_TYPE ((obj), BMP_TYPE_PLAYLIST))
#define BMP_IS_PLAYLIST_CLASS(klass)      (G_TYPE_CHECK_CLASS_TYPE ((klass), BMP_TYPE_PLAYLIST))
#define BMP_PLAYLIST_GET_CLASS(obj)       (G_TYPE_INSTANCE_GET_CLASS ((obj), BMP_TYPE_PLAYLIST, BmpPlaylistClass))

struct BmpPlaylistPrivate;

struct BmpPlaylist
{
        GObject parent;
        BmpPlaylistPrivate *priv;
};

struct BmpPlaylistClass
{
        GObjectClass parent;
        void (*tracklist_list_sorted)    (BmpPlaylist *playlist, int i, gpointer array, int n_items);
        void (*tracklist_items_removed)  (BmpPlaylist *playlist, int i, gpointer array, int n_items);
        void (*tracklist_items_added)    (BmpPlaylist *playlist, int i, gpointer array, int n_items);
};

#ifndef BMP_PLUGIN_BUILD
unsigned int
vcolumns_get_n_items  (void);

Bmp::Library::Datum
vcolumns_get_item_nth (unsigned int n);
#endif

enum TracklistColumn 
{
    COLUMN_ARTIST = 0,
    COLUMN_ALBUM,
    COLUMN_TRACK,
    COLUMN_TITLE,
    COLUMN_GENRE,
    COLUMN_DATE,
    COLUMN_DURATION,
    COLUMN_GUID,
    COLUMN_URI,
};

GType
bmp_playlist_get_type                   (void);

BmpPlaylist*
bmp_playlist_new                        (void);

void
bmp_playlist_setup             (BmpPlaylist *playlist);

GtkListStore*
bmp_playlist_get_playlist      (BmpPlaylist *playlist);

Glib::RefPtr<Bmp::Playlist>
bmp_playlist_get_playlist_pp   (BmpPlaylist *playlist);

int
bmp_playlist_get_length        (BmpPlaylist *playlist);

int
bmp_playlist_get_row_by_guid   (BmpPlaylist *playlist,
                                RowGUID     *guid);

char*
bmp_playlist_get_uri_by_index  (BmpPlaylist  *playlist,
                                int           index);

char*
bmp_playlist_get_title_nth      (BmpPlaylist  *playlist,
                                 int          index);
char*
bmp_playlist_get_current_title  (BmpPlaylist  *playlist);

char*
bmp_playlist_get_uri_nth        (BmpPlaylist  *playlist,
                                 int           index);
char*
bmp_playlist_get_current_uri    (BmpPlaylist  *playlist);

int
bmp_playlist_insert_items	      (BmpPlaylist  *playlist,
                                 const Bmp::VFS::UriList& list,
                                 int index); 

void
bmp_playlist_remove_all		(BmpPlaylist      *playlist);

void
bmp_playlist_row_removing	    (BmpPlaylist  *playlist,
                               GtkTreePath	*path);
 
void
bmp_playlist_load_playlist	  (BmpPlaylist  *playlist);

void
bmp_playlist_connect_library	(BmpPlaylist  *playlist);

void
bmp_playlist_unsort           (BmpPlaylist  *playlist);
 

#endif // BMP_PLAYLIST_HPP
