//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <gtkmm.h>
#include "stock.hpp"

namespace Bmp
{
  using namespace std;
  using namespace Glib;
  using namespace Gtk;
  using namespace Gdk;

  void
  register_stock_icons ()
  {
      RefPtr<Gtk::IconFactory> factory = IconFactory::create ();

      struct StockIconSpec
      {
        char* filename;
        char* stock_id;
      };

      StockIconSpec list[] =
      {   { "library.png",                 BMP_STOCK_LIBRARY          },
#ifdef HAVE_MOOD
          { "museek.png",                  BMP_STOCK_MUSEEK,          },
#endif
          { "streams.png",                 BMP_STOCK_STREAMS          },
          { "bmp.png",                     BMP_STOCK_ABOUT            },
          { "bookmarks.png",               BMP_STOCK_BOOKMARKS        },
          { "xiph.png",                    BMP_STOCK_XIPH             },
          { "shoutcast.png",               BMP_STOCK_SHOUTCAST        },
          { "musicbrainz.png",             BMP_STOCK_MUSICBRAINZ      },
          { "amazon.png",                  BMP_STOCK_AMAZON           },
          { "search.png",                  BMP_STOCK_SEARCH           },
          { "tachikoma.png",               BMP_STOCK_TACHIKOMA        },
          { "audio.png",                   BMP_STOCK_AUDIO            },
          { "tag.png",                     BMP_STOCK_TAG              },
          { "lastfm.png",                  BMP_STOCK_LASTFM           },
          { "playlist.png",                BMP_STOCK_PLAYLIST         },
          { "approve.png",                 BMP_STOCK_APPROVE          },
          { "silk-mod-server_down.png",    BMP_STOCK_XFER_DOWN        },
          { "silk-mod-server_up.png",      BMP_STOCK_XFER_UP          },
          { "silk-cart_put.png",           BMP_STOCK_DOWNLOAD         },
          { "silk-cd_edit.png",            BMP_STOCK_EDIT_ALBUM       },
          { "silk-cd_go.png",              BMP_STOCK_MOVE_ALBUM       },
          { "silk-cd_add.png",             BMP_STOCK_ADD_MUSIC        },
          { "silk-cd_delete.png",          BMP_STOCK_REMOVE_MUSIC     },
          { "silk-application_edit.png",   BMP_STOCK_PREFERENCES      },
          { "silk-control_play_blue.png",  BMP_STOCK_MEDIA_PLAY       },
          { "silk-control_stop_blue.png",  BMP_STOCK_MEDIA_STOP       },
          { "silk-control_pause_blue.png", BMP_STOCK_MEDIA_PAUSE      },
          { "silk-add.png",                BMP_STOCK_ADD              },
          { "silk-delete.png",             BMP_STOCK_DELETE           },
          { "silk-plugin.png",             BMP_STOCK_PLUGIN           },
          { "silk-plugin_disabled.png",    BMP_STOCK_PLUGIN_DISABLED  },  };

    for (unsigned int n = 0; n < G_N_ELEMENTS (list); ++n)
    {
      string filename = build_filename (DATA_DIR G_DIR_SEPARATOR_S "images" G_DIR_SEPARATOR_S "stock", list[n].filename);
      RefPtr<Pixbuf> pixbuf = Pixbuf::create_from_file (filename);
      factory->add (StockID (list[n].stock_id), IconSet(pixbuf));
    }

#ifdef HAVE_GAIM
      StockIconSpec list_gaim[] =
      {   { "oscar.png",          BMP_STOCK_GAIM_OSCAR       },
          { "gadu-gadu.png",      BMP_STOCK_GAIM_GADU        },
          { "jabber.png",         BMP_STOCK_GAIM_JABBER      },
          { "msn.png",            BMP_STOCK_GAIM_MSN         },
          { "napster.png",        BMP_STOCK_GAIM_NAPSTER     },
          { "yahoo.png",          BMP_STOCK_GAIM_YAHOO       },
          { "zephyr.png",         BMP_STOCK_GAIM_ZEPHYR      },
          { "gaim.png",           BMP_STOCK_GAIM             }, };

    for (unsigned int n = 0; n < G_N_ELEMENTS (list_gaim); ++n)
    {
      string filename = Glib::build_filename (DATA_DIR G_DIR_SEPARATOR_S "images" G_DIR_SEPARATOR_S "stock" G_DIR_SEPARATOR_S "gaim", list_gaim[n].filename);
      RefPtr<Gdk::Pixbuf> pixbuf = Pixbuf::create_from_file (filename);
      factory->add (StockID (list_gaim[n].stock_id), IconSet(pixbuf));
    }
#endif
    factory->add_default ();
  }
}
