//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <gdkmm/wrap_init.h>
#include <gtkmm/wrap_init.h>
#include <chroma/chroma-list.h>
#include <chroma/chroma-list-selection.h>
#include "shell.hpp"
#include "ui_util.hpp"
#include "action_defines.hpp"

#include <glibmm/i18n.h>
#include <cstring>

#include <dbus/dbus-glib.h>
#include <bmp/dbus.h>

#include <goa/libgoa.h>

#include "bmp-marshalers.h"
#include "plugin.h"
#include "interfaces.h"

#include "util.hpp"
#include "loader.hpp"
#include "network.hpp"

#include "play.hpp"
#include "playlist.hpp"
#include "history.hpp"

#include "x_bookmarks.hpp"
#include "x_lastfm.hpp"
#include "x_library.hpp"
#include "x_play.hpp"
#include "x_playlist.hpp"
#include "x_vfs.hpp"

#include "playlist++.hpp"

#include "system_control.hpp"
#include "main.hpp"

#define MPRIS_IDENTITY "BMPx"

#include "gmacros.hpp"
BMP_DEFINE_TYPE (BmpSystemControl, bmp_system_control, G_TYPE_OBJECT)

using namespace Bmp;

namespace
{
  bool no_remote_L = false; 
  bool proceed = true;
}

enum BmpSystemControlSignals 
{
    BMP_SYSTEM_CONTROL_SIGNAL_APP_BUSY,
    BMP_SYSTEM_CONTROL_SIGNAL_APP_IDLE,

    BMP_SYSTEM_CONTROL_SIGNAL_SEEK_EVENT,

    BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_SHUFFLE,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_REPEAT,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_BITRATE,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_SAMPLERATE,
    BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_URI,

    BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE,

    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED,
    BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED,

    BMP_SYSTEM_CONTROL_SIGNAL_SET_STOP_AFTER_CURRENT,

    BMP_SYSTEM_CONTROL_SIGNAL_STARTUP_COMPLETE,
    BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_COMPLETE,
    BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_REQUEST,

    BMP_SYSTEM_CONTROL_N_SIGNALS
};

namespace { guint signals[BMP_SYSTEM_CONTROL_N_SIGNALS] = { 0 }; }

enum BmpSystemControlProperties
{
    BMP_SYSTEM_CONTROL_PROPERTY_REPEAT = 1,
    BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE,

    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TITLE,
    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_URI,
    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_INDEX,
    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_BITRATE,
    BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_SAMPLERATE,

    BMP_SYSTEM_CONTROL_PROPERTY_STOP_AFTER_CURRENT,

    BMP_SYSTEM_CONTROL_N_PROPERTIES
};

struct BmpSystemControlPrivate
{
    gboolean             dispose_has_run;

    GList               *message_domains;
    gint                 msg_domain_id;

    GAsyncQueue         *message_queue;
    guint                message_queue_source_id;

    BmpPlaybackHistory  *history;
    GObject             *plugin; // Flow plugin

    gint                 sc_msg_domain_id;

    SystemShutdownFunc   shutdown_func;
    gpointer             shutdown_data;

    SystemStartupFunc    startup_func;
    gpointer             startup_data;

    /* Properties */
    char           *uri;
    char           *title;
    int             index;
    int             duration;
    bool            played;

    int             bitrate;
    int             samplerate;
    int             stream_pos;
    int             playstatus;

    gboolean        repeat;
    gboolean        shuffle;
    gboolean        stop_after_current;

    /*< private >*/
    gboolean        anonymous;
};

#include "system_control_glue.h"

namespace
{

  void
  play_track      (BmpSystemControl     *control,
                   gint              position);

  void
  play_seek       (BmpSystemControl     *control,
                   gint              position);

  // MCS Callbacks
  void
  mcs_set_bmp_volume (MCS_CB_DEFAULT_SIGNATURE, BmpSystemControl *control)
  {
    int volume = boost::get<int>(value);
    play->property_volume() = volume; 
    g_signal_emit (control, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME], 0, volume);
  }

  gboolean
  send_shutdown_complete    (BmpSystemControl *control)
  {
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL(control), FALSE);

    g_signal_emit (G_OBJECT (control),
           signals[BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_COMPLETE], 0);

    if (Bmp::Network::is_connected ())
    {
      if (mcs->key_get<bool>("bmp","send-statistics"))
      {
        Bmp::Network::send_statistic (Bmp::Network::ACTION_QUIT);
      }
    }
    return TRUE;
  }

  gboolean
  send_startup_complete (BmpSystemControl *control)
  {
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL(control), FALSE);

    if (Bmp::Network::is_connected () && (mcs->key_get<bool>("bmp","send-statistics")))
    {
      Bmp::Network::send_statistic (Bmp::Network::ACTION_START);
    }

    g_signal_emit (G_OBJECT (control),
           signals[BMP_SYSTEM_CONTROL_SIGNAL_STARTUP_COMPLETE], 0);

    return FALSE;
  }

  void
  emit_row_changed_on_index (BmpSystemControl   *control,
                             int                 index)
  {
    GtkListStore *model = bmp_playlist_get_playlist (bmp_playlist);
    GtkTreePath  *path;
    GtkTreeIter   iter;

    if (index < 0) return;

    path = gtk_tree_path_new_from_indices (index, -1);
    gtk_tree_model_get_iter (GTK_TREE_MODEL(model), &iter, path);
    gtk_tree_model_row_changed (GTK_TREE_MODEL(model), path, &iter);
    gtk_tree_path_free (path);
  }

  void
  change_current_row (BmpSystemControl *control)
  {
    GtkListStore      *list;
    GtkTreePath       *path;
    gpointer           data;

    list = bmp_playlist_get_playlist (bmp_playlist);
    data = g_object_get_data (G_OBJECT (list), "current");
    g_object_set_data (G_OBJECT (list), "current", NULL);

    if (data)
    {
      path = gtk_tree_row_reference_get_path ((GtkTreeRowReference*)(data));
      if (path)
      {
        emit_row_changed_on_index (control, gtk_tree_path_get_indices (path)[0]);
        g_object_set_data (G_OBJECT (list), "current", data);
        gtk_tree_path_free (path);
      }
      else
      {
        gtk_tree_row_reference_free ((GtkTreeRowReference*)data);
      }
    }
  }

  void
  play_anonymous
             (BmpSystemControl*   control,
              const char*         uri)
  {
    change_current_row (control);

    g_object_set (G_OBJECT(control), "current_index", -1, NULL);

    play->property_status() = PLAYSTATUS_WAITING;
    play->property_stream() = uri; 
    play->property_status() = PLAYSTATUS_PLAYING;

    control->priv->anonymous = TRUE;

    g_signal_emit (G_OBJECT(control), signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE], 0);

    control->priv->duration = 0;
    control->priv->played   = true;
  
    if (control->priv->title) 
        g_free (control->priv->title);
    if (control->priv->uri)
        g_free (control->priv->uri);

    control->priv->title = 0;
    g_object_notify (G_OBJECT (control), "current-title");

    control->priv->uri = g_strdup (uri);
    g_object_notify (G_OBJECT (control), "current-uri");
  }

  void
  play_track (BmpSystemControl      *control,
              int                    position)
  {
    change_current_row (control);

    char *uri = bmp_playlist_get_uri_by_index (bmp_playlist, position);

    if (!uri) return;

    g_object_set (G_OBJECT(control), "current_index", position, NULL);
    bmp_playback_history_set (control->priv->history, position);

    play->property_status() = PLAYSTATUS_WAITING;
    play->property_stream() = uri; 
    play->property_status() = PLAYSTATUS_PLAYING;

    free (uri);

    control->priv->anonymous = FALSE;
    emit_row_changed_on_index (control, position);
    bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_TRACK_CHANGE, NULL, NULL);
  }

  void
  play_current (BmpSystemControl *control)
  {
    GtkTreePath *path = 0;
    if (wTracklist)
    {
      path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION(wTracklist));
    }
    else
    {
      GtkTreeRowReference *reference;

      reference = static_cast<GtkTreeRowReference *>
          (g_object_get_data (G_OBJECT(bmp_playlist_get_playlist(bmp_playlist)), "current"));

      if (reference)
      {
        GtkTreePath *path = 0;
        path = gtk_tree_row_reference_get_path (reference);
        if (!path)
          {
            gtk_tree_row_reference_free (reference);
            g_object_set_data (G_OBJECT(bmp_playlist_get_playlist(bmp_playlist)), "current", NULL);
            return;
          }
      }
    }

    int current = 0;
    if (path)
    {
        current = gtk_tree_path_get_indices (path)[0];
        gtk_tree_path_free (path);
        play_track (control, current);
    }
  }

  void
  play_next (BmpSystemControl *control)
  {
    if (control->priv->stop_after_current)
      {
        play->property_status() = PLAYSTATUS_STOPPED;
        g_object_set (G_OBJECT (control), "stop_after_current", FALSE, NULL);
        return;
      }

    change_current_row (control);

    bool dirty = control->priv->history->dirty;
    gboolean repeat = g_object_get_boolean        (control, "repeat");
    int position    = bmp_playback_history_next   (control->priv->history,
                                                  (BmpPluginFlowInterface*)control->priv->plugin);
    if (position < 0)
      {
        if (bmp_playlist_get_length (bmp_playlist))
          {
            if (!dirty && repeat)
              {
                bmp_playback_history_set (control->priv->history, 0);
                position = 0;
              }
            else
            if (!repeat)
              {
                play->property_status() = PLAYSTATUS_STOPPED;
                return;
              }
            else
              {
                position = bmp_playback_history_rewind (control->priv->history);
              }
          }
        else
          {
            play->property_status() = PLAYSTATUS_STOPPED;
            return;
          }
      }

    char *uri = bmp_playlist_get_uri_by_index (bmp_playlist, position);
    if (!uri) return;

    play->property_status() = PLAYSTATUS_WAITING;
    play->property_stream() = uri; 
    play->property_status() = PLAYSTATUS_PLAYING;

    free (uri);

    control->priv->anonymous = FALSE;
    g_object_set (G_OBJECT(control), "current_index", position, NULL);
    emit_row_changed_on_index (control, position);
    bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_TRACK_CHANGE, NULL, NULL);
  }

  void
  play_prev (BmpSystemControl *control)
  {
    if (control->priv->stop_after_current)
      {
        g_object_set (G_OBJECT (control), "stop_after_current", FALSE, NULL);
        return;
      }

    change_current_row (control);

    int position = bmp_playback_history_prev (control->priv->history,
                      (BmpPluginFlowInterface *) (control->priv->plugin));

    if (position < 0) position = 0;

    char *uri = bmp_playlist_get_uri_by_index (bmp_playlist, position);
    if (!uri) return;

    play->property_status() = PLAYSTATUS_WAITING;
    play->property_stream() = uri;
    play->property_status() = PLAYSTATUS_PLAYING;

    free (uri);

    control->priv->anonymous = FALSE;
    g_object_set (G_OBJECT(control), "current_index", position, NULL);
    emit_row_changed_on_index (control, position);
    bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_TRACK_CHANGE, NULL, NULL);
  }

  void
  play_pause (BmpSystemControl *control)
  {
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (control));
    play->property_status() = PLAYSTATUS_PAUSED;
  }

  void
  play_stop (BmpSystemControl *control)
  {
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (control));
    play->property_status() = PLAYSTATUS_STOPPED;
  } 

  void
  play_seek   (BmpSystemControl *control,
               gint              position)
  {
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (control));
    play->property_position() = position; 
  }

  void
  play_seek_percent (BmpSystemControl   *control,
           gdouble       position)
  {
    g_return_if_fail (BMP_IS_SYSTEM_CONTROL (control));

    gdouble length = play->property_length();
    gdouble percent = length / 100.;
    int position_new = int (position * percent);
    play->property_position() = position_new; 
  }

  gboolean
  process_messages (BmpSystemControl *control)
  {
    BmpSystemMessage *message;
    gint queue_length;

    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL (control), FALSE);

    message = static_cast<BmpSystemMessage *> (g_async_queue_try_pop (control->priv->message_queue));

    if (!message)
      {
        control->priv->message_queue_source_id = 0;
        return FALSE;
      }

    switch (message->message_type)
      {

        case BMP_SYSTEM_MESSAGE_PAUSE:
            {
                play_pause (control);
                break;
            }

        case BMP_SYSTEM_MESSAGE_STOP:
            {
                play_stop (control);
                break;
            }

        case BMP_SYSTEM_MESSAGE_NEXT:
            {
                play_next (control);
                break;
            }

        case BMP_SYSTEM_MESSAGE_CURRENT:
            {
                play_current (control);
                break;
            }

        case BMP_SYSTEM_MESSAGE_PREV:
            {
                play_prev (control);
                break;
            }

        case BMP_SYSTEM_MESSAGE_PLAY_TRACK:
            {
                play_track (control, GPOINTER_TO_INT (message->data));
                break;
            }


        case BMP_SYSTEM_MESSAGE_TRACK_CHANGE:
            {
                g_signal_emit (G_OBJECT(control), signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE], 0);
                control->priv->played = false;
                // We intentionally fall trough here
            }

        case BMP_SYSTEM_MESSAGE_UPDATE:
            {
                GtkListStore        *tracklist;
                GtkTreeRowReference *reference;
                GtkTreePath         *path;
                GtkTreeIter          iter;
                char                *title;
                char                *uri;
                int                  current;

                if (play->property_status() == PLAYSTATUS_WAITING) break;
                if (play->property_status() == PLAYSTATUS_PAUSED) break;

                tracklist = bmp_playlist_get_playlist (bmp_playlist);
                reference = static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(tracklist), "current"));
                path = 0;

                if ((play->property_status() == PLAYSTATUS_PLAYING) && reference)
                {
                    path = gtk_tree_row_reference_get_path (reference);
                    if (!path)
                    {
                      gtk_tree_row_reference_free (reference);
                      g_object_set_data (G_OBJECT(tracklist), "current", NULL);
                      control->priv->anonymous = TRUE;
                      break;
                    }
                }
                else
                if (play->property_status() == PLAYSTATUS_STOPPED)
                {

                    control->priv->bitrate = 0;
                    g_object_notify (G_OBJECT (control), "current-bitrate");

                    control->priv->samplerate = 0;
                    g_object_notify (G_OBJECT (control), "current-samplerate");

                    control->priv->duration = 0;

                  if (Bmp::shell_running ())
                  {
                    path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION(wTracklist));
                  }

                  if (!path)
                  {
                    if (!reference || !(path = gtk_tree_row_reference_get_path (reference)))
                    {
                      if (control->priv->title)
                        g_free (control->priv->title);

                      if (control->priv->uri)
                        g_free (control->priv->uri);

                      control->priv->title = 0;
                      g_object_notify (G_OBJECT (control), "current-title");

                      control->priv->uri = 0;
                      g_object_notify (G_OBJECT (control), "current-uri");
                      break;
                    }
                  }
                }

                if (path) current = gtk_tree_path_get_indices (path)[0]; else break;

                title = bmp_playlist_get_title_nth (bmp_playlist, current);
                gtk_tree_model_get_iter (GTK_TREE_MODEL(tracklist), &iter, path);
                gtk_tree_path_free (path);

                if (title && control->priv->title && (!strcmp(title, control->priv->title)))
                  {
                    g_free (title);
                    break;
                  }

                gtk_tree_model_get (GTK_TREE_MODEL(tracklist), &iter,
                                    COLUMN_URI,
                                    &uri,
                                    COLUMN_DURATION,
                                    &(control->priv->duration), -1);

                if (control->priv->title) 
                    free (control->priv->title);
                if (control->priv->uri)
                    free (control->priv->uri);

                control->priv->title = title;
                g_object_notify (G_OBJECT (control), "current-title");

                control->priv->uri = uri;
                g_object_notify (G_OBJECT (control), "current-uri");
                break;
              }

        case BMP_SYSTEM_MESSAGE_STOP_AFTER_CURRENT:
            {
                gboolean stop;
                stop = GPOINTER_TO_INT (message->data);
                g_object_set (G_OBJECT (control), "stop_after_current", stop, NULL);
                break;
            }

        case BMP_SYSTEM_MESSAGE_REPEAT:
            {
                gboolean repeat = GPOINTER_TO_INT (message->data);
                g_object_set (G_OBJECT (control), "repeat", repeat, NULL);
                break;
            }

        case BMP_SYSTEM_MESSAGE_SHUFFLE:
            {
                gboolean shuffle = GPOINTER_TO_INT (message->data);
                g_object_set (G_OBJECT (control), "shuffle", shuffle, NULL);
                break;
            }

        case BMP_SYSTEM_MESSAGE_QUIT:
            {
                gboolean rval;

                rval = FALSE;
                g_signal_emit (G_OBJECT(control), signals[BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_REQUEST], 0, &rval);

                if (!rval)
                {
                  g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "SHUTTING DOWN");
                  control->priv->shutdown_func (control->priv->shutdown_data);
                }
                break;
            }

          case BMP_SYSTEM_MESSAGE_UI_START:
            {
                static bool running = false;

                GdkDisplay *display;

                if (!running)
                {
                  gtk_init (0, NULL);
                  Gdk::wrap_init();
                  Gtk::wrap_init(); 
                  running = true;
                }

                if (message->data)
                  {
                    display = gdk_display_open ((gchar*)message->data);
                    if (display)
                      {
                        gdk_display_manager_set_default_display (gdk_display_manager_get (), display);
                      }
                    else
                      {
                        gdk_display_open_default_libgtk_only();
                      }
                  }
                else
                  {
                    gdk_display_open_default_libgtk_only();
                  }

                g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Starting UI");

                Bmp::shell_init (); 
                g_signal_emit (G_OBJECT (control),
                               signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS],
                               0, play->property_status().get_value());
                bmp_system_control_update (control);
                break;
            }

          case BMP_SYSTEM_MESSAGE_UI_STOP:
            {
                Bmp::shell_stop ();
                break;
            }

      }

    free (message);
    queue_length = g_async_queue_length (control->priv->message_queue);
    if (queue_length == 0)
      {
        control->priv->message_queue_source_id = 0;
        return FALSE;
      }
    return TRUE;
  }

  gboolean
  check_playback_sanity      (BmpSystemControl        *control,
                              GError                 **error)
  {
    if (play->property_sane().get_value() == false)
    {
      bmp_system_control_message_dispatch (control,
                           control->priv->sc_msg_domain_id,
                           GTK_MESSAGE_ERROR,
                           _("The audio system is not properly initialized. Please check the audio settings in the Preferences panel."));
      if (error != NULL)
        {
          *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE),
                    0,
                    _("Unable to initiate playback: check audio settings"));
        }
      return FALSE;
    }

    return TRUE;
  }

  //GObject
  void
  bmp_system_control_init (BmpSystemControl *control)
  {
    control->priv = g_new (BmpSystemControlPrivate, 1);
    control->priv->dispose_has_run = FALSE;

    control->priv->message_queue = g_async_queue_new ();
    control->priv->message_queue_source_id = 0;

    control->priv->bitrate            = 0;
    control->priv->samplerate         = 0;
    control->priv->title              = 0;
    control->priv->uri                = 0;
    control->priv->duration           = 0;
    control->priv->played             = FALSE;

    control->priv->repeat             = FALSE;
    control->priv->shuffle            = FALSE;
    control->priv->index              = -1;

    control->priv->stop_after_current = FALSE;
    control->priv->anonymous          = FALSE;
    control->priv->history            = bmp_playback_history_new();
    control->priv->plugin             = G_OBJECT (g_hash_table_lookup (Plugins::table[Plugins::TYPE_FLOW], "linear"));

    control->priv->message_domains    = NULL;
    control->priv->msg_domain_id      = 0;

    g_assert (control->priv->plugin != NULL);

    control->priv->sc_msg_domain_id =
    bmp_system_control_message_domain_register (control,
                                                "bmp",
                                                "BMP");

    g_message ("running: %s", G_OBJECT_TYPE_NAME(control));
  }

  void
  bmp_system_control_set_property (GObject * object,
                 guint property_id,
                 const GValue * value, GParamSpec * pspec)
  {
    BmpSystemControl *control = BMP_SYSTEM_CONTROL (object);

    switch (property_id)
    {
      case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_INDEX:
          {
              GtkListStore      *list;
              GtkTreePath       *path;
              GtkTreeRowReference   *reference;
              gint           index = g_value_get_int (value);

              list = bmp_playlist_get_playlist (bmp_playlist);
              control->priv->index = index;

              reference = static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(list), "current"));
              gtk_tree_row_reference_free (reference);
              g_object_set_data (G_OBJECT(list), "current", NULL);

              if (index != -1)
                {
                  path = gtk_tree_path_new_from_indices (index, -1);
                  reference = gtk_tree_row_reference_new (GTK_TREE_MODEL(list), path);
                  g_object_set_data (G_OBJECT(list), "current", reference);
                }
          }
          break;

      case BMP_SYSTEM_CONTROL_PROPERTY_STOP_AFTER_CURRENT:
          {
              control->priv->stop_after_current = g_value_get_boolean (value);

              g_signal_emit (control,
                     signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STOP_AFTER_CURRENT],
                     0, control->priv->stop_after_current);
          }
          break;


      case BMP_SYSTEM_CONTROL_PROPERTY_REPEAT:
          {
              control->priv->repeat = g_value_get_boolean (value);
              mcs->key_set<bool>("bmp", "repeat", control->priv->repeat);
              g_signal_emit (control,
                     signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_REPEAT],
                     0, control->priv->repeat);
          }
          break;

      case BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE:
          {
              control->priv->shuffle = g_value_get_boolean (value);
              control->priv->plugin = G_OBJECT (g_hash_table_lookup (Plugins::table[Plugins::TYPE_FLOW],
                                                                                  control->priv->shuffle ? "shuffle" : "linear"));

              mcs->key_set<bool>("bmp", "shuffle", control->priv->shuffle);
              g_signal_emit (control,
                     signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_SHUFFLE],
                     0, control->priv->shuffle);
          }
          break;

      default: break;
    }
  }

  void
  bmp_system_control_get_property (GObject      *object,
                                   unsigned int  property_id,
                                   GValue       *value,
                                   GParamSpec   *pspec)
  {
    BmpSystemControl *control = BMP_SYSTEM_CONTROL (object);

    switch (property_id)
    {
        case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_INDEX:
            {
                g_value_set_int (value, control->priv->index);
                break;
            }

        case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_BITRATE:
            {
                g_value_set_int (value, control->priv->bitrate);
                break;
            }

        case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_SAMPLERATE:
            {
                g_value_set_int (value, control->priv->samplerate);
                break;
            }

        case BMP_SYSTEM_CONTROL_PROPERTY_REPEAT:
            {
                g_value_set_boolean (value, control->priv->repeat);
                break;
            }

        case BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE:
            {
                g_value_set_boolean (value, control->priv->shuffle);
                break;
            }

        case BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TITLE:
            {
                g_value_set_string (value, control->priv->title);
                break;
            }

        default: break;
    }
  }

  GObject *
  bmp_system_control_constructor (GType type,
                guint n_construct_properties,
                GObjectConstructParam * construct_properties)
  {
    GObject *obj;

    {
        /* Invoke parent constructor. */
        BmpSystemControlClass *klass;
        GObjectClass *parent_class;
        klass =
          BMP_SYSTEM_CONTROL_CLASS (g_type_class_peek
                    (BMP_TYPE_SYSTEM_CONTROL));
        parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
        obj = parent_class->constructor (type,
                         n_construct_properties,
                         construct_properties);
    }
    return obj;
  }

  void
  bmp_system_control_dispose (GObject *obj)
  {
    BmpSystemControl *control = (BmpSystemControl*) obj;

    if (control->priv->dispose_has_run) return;
    control->priv->dispose_has_run = TRUE;
   
    if (Bmp::shell_running ()) Bmp::shell_stop (); 

    send_shutdown_complete (control);
    g_message ("stopped: %s", G_OBJECT_TYPE_NAME(control));
  }

  void
  bmp_system_control_finalize (GObject * obj)
  {
    BmpSystemControl *control = (BmpSystemControl *) obj;
    g_free (control->priv);
  }

  void
  bmp_system_control_class_init (BmpSystemControlClass * g_class)
  {
    GObjectClass *gobject_class = G_OBJECT_CLASS (g_class);

    gobject_class->set_property = bmp_system_control_set_property;
    gobject_class->get_property = bmp_system_control_get_property;
    gobject_class->dispose = bmp_system_control_dispose;
    gobject_class->finalize = bmp_system_control_finalize;
    gobject_class->constructor = bmp_system_control_constructor;

    g_object_class_install_property (gobject_class,
                     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_BITRATE,
                     g_param_spec_int
                     ("current-bitrate",
                      "current-bitrate",
                      "Current stream's bitrate.",
                       0, G_MAXINT, 0, GParamFlags (G_PARAM_READABLE)));

    g_object_class_install_property (gobject_class,
                     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_SAMPLERATE,
                     g_param_spec_int
                     ("current-samplerate",
                      "current-samplerate",
                      "Current stream's samplerate.",
                       0, G_MAXINT, 0, GParamFlags (G_PARAM_READABLE)));

    g_object_class_install_property (gobject_class,
                     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_TITLE,
                     g_param_spec_string
                     ("current-title",
                      "current-title",
                      "Current title",
                       NULL, GParamFlags (G_PARAM_READABLE)));

    g_object_class_install_property (gobject_class,
                     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_URI,
                     g_param_spec_string
                     ("current-uri",
                      "current-uri",
                      "Current URI",
                       NULL, GParamFlags (G_PARAM_READABLE)));

    g_object_class_install_property (gobject_class,
                     BMP_SYSTEM_CONTROL_PROPERTY_CURRENT_INDEX,
                     g_param_spec_int
                     ("current_index",
                      "current_index",
                      "Current track index",
                       -1, G_MAXINT, 0, GParamFlags (G_PARAM_READWRITE)));

    g_object_class_install_property (gobject_class,
                     BMP_SYSTEM_CONTROL_PROPERTY_REPEAT,
                     g_param_spec_boolean
                     ("repeat",
                      "repeat",
                      "Player repeat state",
                       FALSE, GParamFlags (G_PARAM_READWRITE)));

    g_object_class_install_property (gobject_class,
                     BMP_SYSTEM_CONTROL_PROPERTY_STOP_AFTER_CURRENT,
                     g_param_spec_boolean
                     ("stop_after_current",
                      "stop_after_current",
                      "Stop after current track",
                       FALSE, GParamFlags (G_PARAM_READWRITE)));

    g_object_class_install_property (gobject_class,
                     BMP_SYSTEM_CONTROL_PROPERTY_SHUFFLE,
                     g_param_spec_boolean
                     ("shuffle",
                      "shuffle",
                      "Player shuffle state",
                       FALSE, GParamFlags (G_PARAM_READWRITE)));

    signals[BMP_SYSTEM_CONTROL_SIGNAL_APP_BUSY] =
      g_signal_new ("app-busy",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, app_busy),
            NULL, NULL, g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_APP_IDLE] =
      g_signal_new ("app-idle",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, app_idle),
            NULL, NULL, g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SEEK_EVENT] =
      g_signal_new ("seek-event",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, seek_event),
            NULL, NULL, g_cclosure_marshal_VOID__INT,
            G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_STARTUP_COMPLETE] =
      g_signal_new ("startup-complete",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, startup_complete),
            NULL, NULL, g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_COMPLETE] =
      g_signal_new ("shutdown-complete",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, shutdown_complete),
            NULL, NULL, g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SHUTDOWN_REQUEST] =
      g_signal_new ("shutdown-request",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (0),
            G_STRUCT_OFFSET (BmpSystemControlClass, shutdown_request),
            NULL, NULL, bmpx_BOOLEAN__VOID,
            G_TYPE_BOOLEAN, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_VOLUME] =
      g_signal_new ("set-volume",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, set_volume),
            NULL, NULL, g_cclosure_marshal_VOID__INT,
            G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_SHUFFLE] =
      g_signal_new ("set-shuffle",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, set_shuffle),
            NULL, NULL, g_cclosure_marshal_VOID__BOOLEAN,
            G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_REPEAT] =
      g_signal_new ("set-repeat",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, set_repeat),
            NULL, NULL, g_cclosure_marshal_VOID__BOOLEAN,
            G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS] =
      g_signal_new ("set-playstatus",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, set_playstatus),
            NULL, NULL, g_cclosure_marshal_VOID__INT,
            G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACK_CHANGE] =
      g_signal_new ("track-change",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, track_change),
            NULL, NULL, g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE, 0);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_URI] =
      g_signal_new ("set-current-uri",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, set_current_uri),
            NULL, NULL, g_cclosure_marshal_VOID__STRING,
            G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_BITRATE] =
      g_signal_new ("set-current-bitrate",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, set_current_bitrate),
            NULL, NULL, g_cclosure_marshal_VOID__INT,
            G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_SAMPLERATE] =
      g_signal_new ("set-current-samplerate",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, set_current_samplerate),
            NULL, NULL, g_cclosure_marshal_VOID__INT,
            G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS] =
      g_signal_new ("set-stream-pos",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass, set_stream_pos),
            NULL, NULL, g_cclosure_marshal_VOID__INT,
            G_TYPE_NONE, 1, G_TYPE_INT);

    signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STOP_AFTER_CURRENT] =
      g_signal_new ("set-stop-after-current",
            G_OBJECT_CLASS_TYPE (gobject_class),
            GSignalFlags (G_SIGNAL_RUN_FIRST | G_SIGNAL_DETAILED),
            G_STRUCT_OFFSET (BmpSystemControlClass,
                     set_stop_after_current), NULL, NULL,
            g_cclosure_marshal_VOID__BOOLEAN, G_TYPE_NONE, 1,
            G_TYPE_BOOLEAN);

    if (no_remote_L)
     {
        signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
            g_signal_new ("tracklist-items-removed",
                          G_OBJECT_CLASS_TYPE (gobject_class),
                          G_SIGNAL_RUN_FIRST,
                          G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_removed),
                          NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                          2, G_TYPE_POINTER, G_TYPE_INT);


        signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED] =
            g_signal_new ("tracklist-items-added",
                          G_OBJECT_CLASS_TYPE (gobject_class),
                          G_SIGNAL_RUN_FIRST,
                          G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_added),
                          NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                          2, G_TYPE_POINTER, G_TYPE_INT);
    }
  else
    {
        signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED] =
            g_signal_new ("tracklist-items-removed",
                          G_OBJECT_CLASS_TYPE (gobject_class),
                          G_SIGNAL_RUN_FIRST,
                          G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_removed),
                          NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                          2, DBUS_TYPE_G_INT_ARRAY, G_TYPE_INT);


        signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED] =
            g_signal_new ("tracklist-items-added",
                          G_OBJECT_CLASS_TYPE (gobject_class),
                          G_SIGNAL_RUN_FIRST,
                          G_STRUCT_OFFSET (BmpSystemControlClass, tracklist_items_added),
                          NULL, NULL, bmpx_VOID__POINTER_INT, G_TYPE_NONE,
                          2, DBUS_TYPE_G_INT_ARRAY, G_TYPE_INT);
    }
  }

  void
  on_play_eos (BmpSystemControl *control)
  {
    bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
  }

  void
  on_play_playstatus (BmpSystemControl *control)
  {
    control->priv->playstatus = play->property_status().get_value();
    bmp_system_control_update (control);
    g_signal_emit (G_OBJECT (control),
           signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS],
           0, play->property_status().get_value());
  }

  void
  on_play_position (int               position,
                    BmpSystemControl *control)
  { 
    control->priv->stream_pos = position; 

    // Check if the track qualifies as played
    if ((control->priv->duration > 30) && !control->priv->played)
      {
        if ((position >= 240) || (position >= (control->priv->duration/2)))
          {
            control->priv->played = true;
            if (mcs->key_get<bool>("lastfm", "queue-enable"))
              {
                lastfm->send_song_information (position);
              }
            library->increase_play_count (control->priv->uri);
          }
      }

    g_signal_emit (control, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_STREAM_POS], 0, position);
  }

  void
  on_play_seek    (int                 position,
                   BmpSystemControl   *control)
  {
    if (!control->priv->played && ((position >= 240) || (position >= (control->priv->duration/2))))
      {
        control->priv->played = true;
      }

    g_signal_emit (G_OBJECT (control), signals[BMP_SYSTEM_CONTROL_SIGNAL_SEEK_EVENT], 0, position);
  }

  void
  on_play_title   (char               *title,
                   BmpSystemControl   *control)
  {
    if (control->priv->title)
        free (control->priv->title);

    control->priv->title = strdup (title); 
    g_object_notify (G_OBJECT (control), "current-title");
  }

  void
  on_play_bitrate     (unsigned int        bitrate,
                       BmpSystemControl   *control)
  {
    if (bitrate == 0) return;
    control->priv->bitrate = bitrate;
    g_object_notify (G_OBJECT (control), "current-bitrate");
  }

  void
  on_play_samplerate  (unsigned int samplerate,
                       BmpSystemControl   *control)
  {
    if (samplerate == 0) return;
    control->priv->samplerate = samplerate;
    g_object_notify (G_OBJECT (control), "current-samplerate");
  }

  void
  playlist_callback_tracklist_items_removed (BmpPlaylist      *playlist,
                                             GArray           *removed_rows,
                                             gint              n_removed_rows,
                                             BmpSystemControl *control)
  {
    g_signal_emit (G_OBJECT (control),
                   signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_REMOVED],
                   0, removed_rows, n_removed_rows);

    if (gtk_tree_model_iter_n_children (reinterpret_cast<GtkTreeModel*>(bmp_playlist_get_playlist (playlist)), NULL) == 0)
        bmp_system_control_clear_playback_history (control); 

    bmp_system_control_update (control);
  }

  void
  playlist_callback_tracklist_items_added (BmpPlaylist      *playlist,
                                           GArray           *added_rows,
                                           int               n_added_rows,
                                           BmpSystemControl *control)
  {
    g_signal_emit (G_OBJECT (control), signals[BMP_SYSTEM_CONTROL_SIGNAL_TRACKLIST_ITEMS_ADDED], 0);
  }

  void
  set_current_uri
              (BmpSystemControl *control)
  {
      g_signal_emit (control, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_URI],
                              0, control->priv->uri); 
  }

  void
  set_current_samplerate
              (BmpSystemControl *control)
  {
      g_signal_emit (control, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_SAMPLERATE],
                              0, control->priv->samplerate); 
  }
  
  void
  set_current_bitrate
              (BmpSystemControl *control)
  {
      g_signal_emit (control, signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_CURRENT_SAMPLERATE],
                              0, control->priv->bitrate); 
  }

}

gboolean
bmp_system_control_identity (BmpSystemControl  *control,
                             gchar            **name,
                             GError           **error)
{
  *name = g_strdup(MPRIS_IDENTITY);
   return TRUE;
}

gboolean
bmp_system_control_add_uri_list (BmpSystemControl   *control,
                                 const char        **strv,
                                 gint                position,
                                 gboolean            clear,
                                 gboolean            start_playback,
                                 gint                index,
                                 gint               *position_out,
                                 GError            **error)
{
    g_return_val_if_fail (BMP_IS_SYSTEM_CONTROL (control), TRUE);

#if 0
    Glib::RefPtr<Bmp::Playlist> p = bmp_playlist_get_playlist_pp (bmp_playlist);
    VFS::UriList list;
    Bmp::VFS::VFS::strv_to_uri_list (list, (char**)strv);
    int out = p->insert_items (list, -1);
    return TRUE;
#endif
    
    bmp_system_control_app_busy (control);

    bmp_playlist_unsort (bmp_playlist);

    if (clear)
      {
        bmp_playlist_remove_all (bmp_playlist);
      }

    if (index == -1)
      {
        index = bmp_playlist_get_length (bmp_playlist);
      }

    VFS::UriList list;
    Bmp::VFS::VFS::strv_to_uri_list (list, (char**)strv);

    *position_out = bmp_playlist_insert_items (bmp_playlist, list, (position < 0) ? bmp_playlist_get_length (bmp_playlist) : position); 

    if (start_playback)
      {
        bmp_system_control_play_track (control, index, NULL);
      }

    bmp_system_control_app_idle (control);

    return TRUE;
}

gboolean
bmp_system_control_get_metadata_for_list_item (BmpSystemControl   *control,
                                               int                 item_idx,
                                               GHashTable        **metadata,
                                               GError            **error)
{
    char *uri;
    Bmp::DB::Row row;
    uri = bmp_playlist_get_uri_by_index (bmp_playlist, item_idx);
    if (!uri)
      {
          *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE), 0, "Index out of range");
      return FALSE;
      }
    library->get_metadata (uri, row);
    *metadata = library->get_hashtable (row); 
    free (uri);
    return TRUE;
}

gboolean
bmp_system_control_get_metadata_for_uri       (BmpSystemControl   *control,
                                               char           *uri,
                                               GHashTable    **metadata,
                                               GError        **error)
{
    Bmp::DB::Row row;
    library->get_metadata (uri, row);
    *metadata = library->get_hashtable (row); 
    return TRUE;
}

gboolean
bmp_system_control_send_status (BmpSystemControl *control, GError ** error)
{
    g_signal_emit (G_OBJECT (control),
           signals[BMP_SYSTEM_CONTROL_SIGNAL_SET_PLAYSTATUS],
           0, play->property_status().get_value());

    return TRUE;
}

gboolean
bmp_system_control_volume_get (BmpSystemControl  *control,
                               int               *volume,
                               GError           **error)
{
  *volume = mcs->key_get<int>("bmp", "volume"); 
  return TRUE;
}

gboolean
bmp_system_control_volume_set (BmpSystemControl  *control,
                               int                volume,
                               GError           **error)
{
  mcs->key_set<int>("bmp", "volume", volume);
  return TRUE;
}

gboolean
bmp_system_control_get_current_title (BmpSystemControl  *control,
                                      gchar            **title,
                                      GError           **error)
{
    *title = g_strdup (control->priv->title);
    return TRUE;
}

gboolean
bmp_system_control_get_current_stream_pos
                                     (BmpSystemControl  *control,
                                      int               *position,
                                      GError           **error)
{
  *position = control->priv->stream_pos; 
  return TRUE;
}

gboolean
bmp_system_control_get_current_play_status
                                      (BmpSystemControl  *control,
                                       int               *playstatus,
                                       GError           **error)
{
  *playstatus = control->priv->playstatus; 
  return TRUE;
}

gboolean
bmp_system_control_get_current_bitrate
                                      (BmpSystemControl  *control,
                                       int               *bitrate,
                                       GError           **error)
{
    *bitrate = control->priv->bitrate;
    return TRUE;
}

gboolean
bmp_system_control_get_current_samplerate
                                      (BmpSystemControl  *control,
                                       int               *samplerate,
                                       GError           **error)
{
    *samplerate = control->priv->samplerate;
    return TRUE;
}

gboolean
bmp_system_control_get_current_uri
                                     (BmpSystemControl    *control,
                                      gchar              **uri,
                                      GError             **error)
{
    if (!control->priv->uri)
      {
          *error = g_error_new (g_quark_from_static_string (BMP_DBUS_INTERFACE), 0, "No current URI");
          return FALSE;
      }
    *uri = g_strdup (control->priv->uri);
    return TRUE;
}

gboolean
bmp_system_control_get_current_track (BmpSystemControl  *control,
                                      gint              *idx,
                                      GError           **error)
{
  *idx = g_object_get_int (control, "current_index");
  return TRUE;
}

gboolean
bmp_system_control_get_bookmarks		(BmpSystemControl	 *control,
                        						 GArray		        **array,
                        						 GError		        **error)
{
 *array = bookmarks->get_array (); 
  return TRUE;
}

gboolean
bmp_system_control_get_tracklist (BmpSystemControl *control,
                                  char ***array,
                                  GError ** error)
{
  GtkListStore      *list;
  GtkTreeIter        iter;
  gint               n, l;
  gboolean           loop;

  list = bmp_playlist_get_playlist (bmp_playlist);
  l = gtk_tree_model_iter_n_children (GTK_TREE_MODEL(list), NULL);
  *array = g_new0 (gchar*, l+1);
  loop = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(list), &iter);
  n = 0;

  //FIXME: This is not completely safe because rows might get removed while we're doing this.
  while (loop)
    {
      char *uri;
      gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, COLUMN_URI, &uri, -1);
      (*array)[n] = uri; 
      loop = gtk_tree_model_iter_next (GTK_TREE_MODEL(list), &iter);
      n++;
    }

  (*array)[l] = NULL;
  return TRUE;
}

gboolean
bmp_system_control_export_tracklist (BmpSystemControl *control,
                                     const gchar      *container,
                                     gboolean          all,
                                     const gchar      *filename,
                                     GError          **error)
{
  GtkListStore        *list;
  GtkTreeIter          iter;
  bool                 loop;
  Bmp::VFS::UriList    u_list;

  Bmp::VFS::Handle handle = Bmp::VFS::Handle::create (filename);
  list = bmp_playlist_get_playlist (bmp_playlist);

  if ((all) || (!Bmp::shell_running()))
    {
      loop = gtk_tree_model_get_iter_first (GTK_TREE_MODEL(list), &iter);
      while (loop)
      {
        char *uri;
        gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, COLUMN_URI, &uri, -1);
        u_list.push_back (uri);
        free (uri);
        loop = gtk_tree_model_iter_next (GTK_TREE_MODEL(list), &iter);
      }
    }
  else
    {
      GList *_iter;
      _iter = chroma_list_selection_get_selected_paths (CHROMA_LIST_SELECTION (wTracklist));
      for ( ; _iter ; _iter = _iter->next )
      {
        GtkTreeIter  iter;
        char        *uri;

        gtk_tree_model_get_iter (GTK_TREE_MODEL(list), &iter, (GtkTreePath*)_iter->data);
        gtk_tree_model_get (GTK_TREE_MODEL(list), &iter, COLUMN_URI, &uri, -1);
        u_list.push_back (uri);
        free (uri);
      }
      g_list_foreach (g_list_first (_iter), (GFunc)gtk_tree_path_free, NULL);
    }

  try 
    {
      vfs->write (u_list, Glib::filename_to_uri(filename), container);
    }
  catch (Bmp::VFS::Exception& cxe)
    {
      bmp_system_control_message_dispatch (control,
                                           control->priv->sc_msg_domain_id,
                                           GTK_MESSAGE_ERROR,
                                           _("Unable to export playlist: An VFS error occured."));
    }

  return TRUE;
}

void
bmp_system_message_send (BmpSystemControl       *control,
                         BmpSystemMessageType    message_type,
                         gpointer                data,
                         GDestroyNotify          data_destroy)
{
  g_return_if_fail (BMP_IS_SYSTEM_CONTROL (control));

  BmpSystemMessage *message;
  message = g_new0 (BmpSystemMessage, 1);
  message->message_type = message_type;
  message->data = data;
  message->data_destroy = data_destroy;
  g_async_queue_push (control->priv->message_queue, message);
  if (control->priv->message_queue_source_id != 0) return;
  control->priv->message_queue_source_id = g_idle_add ((GSourceFunc) process_messages, control);
}

void
bmp_system_control_update (BmpSystemControl *control)

{
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_UPDATE, NULL, NULL);
}

gboolean
bmp_system_control_startup (BmpSystemControl *control, GError **error)
{
  return TRUE;
}

gboolean
bmp_system_control_stop_after_current_set (BmpSystemControl *control, gboolean stop, GError **error)
{
  bmp_system_message_send (control,
                 BMP_SYSTEM_MESSAGE_STOP_AFTER_CURRENT,
                 (gpointer)GINT_TO_POINTER(stop),
                 NULL);

  return TRUE;
}

gboolean
bmp_system_control_repeat_set (BmpSystemControl *control, gboolean repeat, GError **error)
{
  bmp_system_message_send (control,
                 BMP_SYSTEM_MESSAGE_REPEAT,
                 (gpointer)GINT_TO_POINTER(repeat),
                 NULL);

  return TRUE;
}

gboolean
bmp_system_control_repeat_get (BmpSystemControl *control, gboolean *repeat, GError **error)
{
  *repeat = control->priv->repeat;
  return TRUE;
}

gboolean
bmp_system_control_shuffle_set (BmpSystemControl *control, gboolean shuffle, GError **error)
{
  bmp_system_message_send (control,
                           BMP_SYSTEM_MESSAGE_SHUFFLE,
                           (gpointer)GINT_TO_POINTER(shuffle),
                           NULL);
  return TRUE;
}

gboolean
bmp_system_control_shuffle_get (BmpSystemControl *control, gboolean *shuffle, GError **error)
{
  *shuffle = control->priv->shuffle;
  return TRUE;
}

gboolean
bmp_system_control_ui_start (BmpSystemControl   *control,
                             gchar              *display,
                             GError            **error)
{
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_UI_START, display, NULL);
  return TRUE;
}

gboolean
bmp_system_control_ui_stop  (BmpSystemControl   *control,
                             GError        **error)
{
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_UI_STOP, NULL, NULL);
  return TRUE;
}

gboolean
bmp_system_control_ui_raise (BmpSystemControl   *control,
                             GError            **error)
{
  Bmp::raise_windows ();
  return TRUE;
}

gboolean
bmp_system_control_quit     (BmpSystemControl   *control,
                             GError            **error)
{
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_QUIT, NULL, NULL);
  return TRUE;
}

gboolean
bmp_system_control_play_track (BmpSystemControl   *control,
                               gint                index,
                               GError            **error)
{
  if (!check_playback_sanity (control, error)) return FALSE;
  bmp_system_message_send (control,
               BMP_SYSTEM_MESSAGE_PLAY_TRACK,
               (gpointer)GINT_TO_POINTER(index),
               NULL);
  return TRUE;
}

gboolean
bmp_system_control_seek (BmpSystemControl     *control,
                         gint                  position,
                         GError              **error)
{
  if (!check_playback_sanity (control, error)) return FALSE;
  play_seek (control, position);
  return TRUE;
}

gboolean
bmp_system_control_seek_percent (BmpSystemControl *control,
                                 gdouble           position,
                                 GError          **error)
{
  if (!check_playback_sanity (control, error)) return FALSE;
  play_seek_percent (control, position);
  return TRUE;
}

gboolean
bmp_system_control_pause      (BmpSystemControl   *control,
                               GError            **error)
{
  if (!check_playback_sanity (control, error)) return FALSE;
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_PAUSE, NULL, NULL);
  return TRUE;
}


gboolean
bmp_system_control_stop      (BmpSystemControl    *control,
                              GError             **error)
{
  if (!check_playback_sanity (control, error)) return FALSE;
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_STOP, NULL, NULL);
  return TRUE;
}


gboolean
bmp_system_control_go_prev   (BmpSystemControl    *control,
                              GError             **error)
{
  if (!check_playback_sanity (control, error)) return FALSE;
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_PREV, NULL, NULL);
  return TRUE;
}

gboolean
bmp_system_control_go_next   (BmpSystemControl    *control,
                              GError             **error)
{
  if (!check_playback_sanity (control, error)) return FALSE;
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_NEXT, NULL, NULL);
  return TRUE;
}

gboolean
bmp_system_control_play      (BmpSystemControl     *control,
                              GError              **error)
{
  if (!check_playback_sanity (control, error)) return FALSE;
  bmp_system_message_send (control, BMP_SYSTEM_MESSAGE_CURRENT, NULL, NULL);
  return TRUE;
}

void
bmp_system_control_app_busy (BmpSystemControl *control)
{
  g_signal_emit (G_OBJECT (control),
                 signals[BMP_SYSTEM_CONTROL_SIGNAL_APP_BUSY],
                 0);
}

void
bmp_system_control_app_idle (BmpSystemControl *control)
{
  g_signal_emit (G_OBJECT (control),
                 signals[BMP_SYSTEM_CONTROL_SIGNAL_APP_IDLE],
                 0);
}

void
bmp_system_control_clear_playback_history (BmpSystemControl *control)
{
    bmp_playback_history_free (control->priv->history);
    control->priv->history = bmp_playback_history_new ();
    bmp_plugin_flow_reset_state (BMP_PLUGIN_FLOW (control->priv->plugin));
}

struct BmpUserMessageDomain
{
    int  id;
    char *name_id;
    char *description;
};

gint
bmp_system_control_message_domain_register (BmpSystemControl    *control,
                                            const gchar         *id,
                                            const gchar         *description)
{
    BmpUserMessageDomain *domain;
    gint new_id;

    new_id = control->priv->msg_domain_id;
    control->priv->msg_domain_id++;

    domain = g_new0 (BmpUserMessageDomain, 1);

    domain->id = new_id;
    domain->name_id = g_strdup (id);
    domain->description = g_strdup (description);

    control->priv->message_domains = g_list_append (control->priv->message_domains, domain);

    return new_id;
}

gint
bmp_system_control_message_dispatch     (BmpSystemControl   *control,
                                         gint               domain_id,
                                         GtkMessageType     msgtype,
                                         const gchar       *text)
{
    BmpUserMessageDomain *domain;
    GtkWidget            *dialog;
    int                   response;

    domain = static_cast<BmpUserMessageDomain *> (g_list_nth_data (control->priv->message_domains, domain_id));

    if (msgtype == GTK_MESSAGE_QUESTION)
    {
        dialog = gtk_message_dialog_new (NULL,
                         GTK_DIALOG_MODAL,
                         msgtype,
                         GTK_BUTTONS_YES_NO,
                         NULL);
    }
    else
    {
        dialog = gtk_message_dialog_new (NULL,
                         GTK_DIALOG_MODAL,
                         msgtype,
                         GTK_BUTTONS_OK,
                         NULL);
    }

    gtk_message_dialog_set_markup (GTK_MESSAGE_DIALOG(dialog), text);
    gtk_widget_realize (dialog);

    gtk_window_set_title (GTK_WINDOW(dialog), domain->description);
    gtk_window_set_position (GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);

    response = gtk_dialog_run (GTK_DIALOG(dialog));
    gtk_widget_destroy (dialog);

    if (msgtype != GTK_MESSAGE_QUESTION) // Pointless if the GUI isn't running
    {
        static gchar *m_strings[] = {
        "INFO",
        "WARNING",
        "QUESTION",
        "ERROR"
        };

        g_log (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s: [%s] from '%s': %s",
                            G_STRLOC,
                            m_strings[msgtype],
                            domain->description,
                            text);
    }
    return response;
}

void
bmp_system_control_initialize_dbus_connection   (BmpSystemControl *control)
{
  DBusGConnection *bus;
  DBusGProxy *bus_proxy;
  gboolean connected = FALSE;
  guint request_name_result;
  GError *error = NULL;

  dbus_g_thread_init ();

  bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
  if (!bus)
    {
      g_message ("running: %s (DBUS Error: '%s')", G_OBJECT_TYPE_NAME(control), error->message);
      g_error_free (error);
      return;
    }

  bus_proxy =
    dbus_g_proxy_new_for_name (bus,
               "org.freedesktop.DBus",
               "/org/freedesktop/DBus",
               "org.freedesktop.DBus");

  if (!dbus_g_proxy_call (bus_proxy, "RequestName", &error,
              G_TYPE_STRING, BMP_DBUS_INTERFACE,
              G_TYPE_UINT, 0,
              G_TYPE_INVALID,
              G_TYPE_UINT, &request_name_result,
              G_TYPE_INVALID))
    {
      g_error ("%s: Failed RequestName request: %s", G_STRFUNC, error->message);
      g_error_free (error);
      error = NULL;
    }

  switch (request_name_result)
    {
      case DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER:
        {
          dbus_g_connection_register_g_object (bus, BMP_DBUS_PATH,
                           G_OBJECT (control));
          connected = TRUE;
          break;
        }

      case DBUS_REQUEST_NAME_REPLY_EXISTS:
        {
          g_message
            ("%s: BMPx session already running, not registering SystemControl object to DBus.",
             G_STRFUNC);
          connected = FALSE;
          proceed = FALSE;
          break;
        }
    }

  if (connected)
    {
      g_message ("%s (DBus:%s)", G_OBJECT_TYPE_NAME(control), BMP_DBUS_INTERFACE);
    }
}

BmpSystemControl*
bmp_system_control_new (gboolean no_remote)
{
  BmpSystemControl *control;

  if (!no_remote)
  dbus_g_object_type_install_info (BMP_TYPE_SYSTEM_CONTROL,
                   &dbus_glib_bmp_system_control_object_info);
  no_remote_L = no_remote;

  control = static_cast<BmpSystemControl *> (g_object_new (bmp_system_control_get_type (), NULL));

  /* Proxy BmpPlaylist signals */
  g_object_connect(G_OBJECT(bmp_playlist),
      "signal::tracklist-items-removed",
      G_CALLBACK(playlist_callback_tracklist_items_removed),
      control,
      "signal::tracklist-items-added",
      G_CALLBACK(playlist_callback_tracklist_items_added),
      control,
      NULL);

  if (!proceed)
  return NULL;
  else
  return control;
}

void
bmp_system_control_init_finalize (BmpSystemControl *control)
{
  play->signal_eos()        .connect  (sigc::bind (&on_play_eos,        control));
  play->signal_position()   .connect  (sigc::bind (&on_play_position,   control));
  play->signal_seek()       .connect  (sigc::bind (&on_play_seek,       control));
  play->signal_title()      .connect  (sigc::bind (&on_play_title,      control));
  play->signal_bitrate()    .connect  (sigc::bind (&on_play_bitrate,    control));
  play->signal_samplerate() .connect  (sigc::bind (&on_play_samplerate, control));

  play->property_status().signal_changed()
                            .connect  (sigc::bind (&on_play_playstatus, control));

  mcs->subscribe ("SystemControl", "bmp", "volume", sigc::bind (sigc::ptr_fun (&mcs_set_bmp_volume), control));

  g_object_connect (G_OBJECT (control),
                    "signal::notify::current-uri",
                    G_CALLBACK (set_current_uri),
                    0,
                    "signal::notify::current-bitrate",
                    G_CALLBACK (set_current_bitrate),
                    0,
                    "signal::notify::current-samplerate",
                    G_CALLBACK (set_current_samplerate),
                    0, NULL);
}

void
bmp_system_control_register_shutdown_func    (BmpSystemControl    *control,
                                              SystemShutdownFunc   shutdown_func,
                                              gpointer             shutdown_data)
{
  g_return_if_fail (BMP_IS_SYSTEM_CONTROL (control));
  control->priv->shutdown_func = shutdown_func;
  control->priv->shutdown_data = shutdown_data;
}

void
bmp_system_control_register_startup_func      (BmpSystemControl     *control,
                                               SystemStartupFunc     startup_func,
                                               gpointer              startup_data)
{
  g_return_if_fail (BMP_IS_SYSTEM_CONTROL (control));
  control->priv->startup_func = startup_func;
  control->priv->startup_data = startup_data;
}

void
bmp_system_control_application_run      (BmpSystemControl     *control)
{
  g_return_if_fail (BMP_IS_SYSTEM_CONTROL (control));
  send_startup_complete (control);
  control->priv->startup_func (control->priv->startup_data);
}

void
bmp_system_control_remove_guid          (BmpSystemControl   *control,
                                         RowGUID            *guid)
{
  bmp_playback_history_remove_guid (control->priv->history, guid);
}

GObject*
bmp_system_control_get_current_plugin   (BmpSystemControl *control)
{
  return control->priv->plugin;
}

void
bmp_system_control_play_anonymous       (BmpSystemControl *control,
                                         const char       *uri)
{
  play_anonymous (control, uri);
}
