//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_SYSTEM_CONTROL_HPP
#define BMP_SYSTEM_CONTROL_HPP

/*
 * Potentially, include other headers on which this header depends.
 */
#include <glib-object.h>
#include <gtk/gtk.h>
#include "guid.hpp"

#define BMP_TYPE_SYSTEM_CONTROL     (bmp_system_control_get_type ())
#define BMP_SYSTEM_CONTROL(obj)     (G_TYPE_CHECK_INSTANCE_CAST ((obj), BMP_TYPE_SYSTEM_CONTROL, BmpSystemControl))
#define BMP_SYSTEM_CONTROL_CLASS(klass)   (G_TYPE_CHECK_CLASS_CAST ((klass), BMP_TYPE_SYSTEM_CONTROL, BmpSystemControlClass))
#define BMP_IS_SYSTEM_CONTROL(obj)    (G_TYPE_CHECK_INSTANCE_TYPE ((obj), BMP_TYPE_SYSTEM_CONTROL))
#define BMP_IS_SYSTEM_CONTROL_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), BMP_TYPE_SYSTEM_CONTROL))
#define BMP_SYSTEM_CONTROL_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), BMP_TYPE_SYSTEM_CONTROL, BmpSystemControlClass))

struct BmpSystemControlPrivate;

struct BmpSystemControl
{
    GObject parent;
    BmpSystemControlPrivate *priv;
};

struct BmpSystemControlClass
{
    GObjectClass parent;

    // Signals

    void     (*app_busy)                  (BmpSystemControl *control);
    void     (*app_idle)                  (BmpSystemControl *control);

    void     (*set_stop_after_current)    (BmpSystemControl *control, gboolean stop);
    void     (*seek_event)                (BmpSystemControl *control, int stream_pos);

    void     (*set_stream_pos)            (BmpSystemControl *control, int stream_pos);
    void     (*set_playstatus)            (BmpSystemControl *control, int playstatus);
    void     (*set_current_bitrate)       (BmpSystemControl *control, int bitrate);
    void     (*set_current_samplerate)    (BmpSystemControl *control, int samplerate);
    void     (*set_current_uri)           (BmpSystemControl *control, char *uri);

    void     (*set_volume)                (BmpSystemControl *control, gint volume);
    void     (*set_shuffle)               (BmpSystemControl *control, gboolean shuffle);
    void     (*set_repeat)                (BmpSystemControl *control, gboolean repeat);

    void     (*tracklist_items_removed)   (BmpSystemControl *control, gpointer array, gint n_items);
    void     (*tracklist_items_added)     (BmpSystemControl *control, gpointer array, gint n_items);

    void     (*startup_complete)          (BmpSystemControl *control);
    void     (*shutdown_complete)         (BmpSystemControl *control);

    gboolean (*shutdown_request)          (BmpSystemControl *control);

    void     (*track_change)              (BmpSystemControl *control);
};

enum BmpSystemMessageType
{
    BMP_SYSTEM_MESSAGE_CURRENT,
    BMP_SYSTEM_MESSAGE_NEXT,
    BMP_SYSTEM_MESSAGE_PREV,
    BMP_SYSTEM_MESSAGE_STOP,
    BMP_SYSTEM_MESSAGE_PAUSE,
    BMP_SYSTEM_MESSAGE_PLAY_TRACK,
    BMP_SYSTEM_MESSAGE_STOP_AFTER_CURRENT,
    BMP_SYSTEM_MESSAGE_REPEAT,
    BMP_SYSTEM_MESSAGE_SHUFFLE,
    BMP_SYSTEM_MESSAGE_UI_START,
    BMP_SYSTEM_MESSAGE_UI_STOP,
    BMP_SYSTEM_MESSAGE_TRACK_CHANGE,
    BMP_SYSTEM_MESSAGE_UPDATE,
    BMP_SYSTEM_MESSAGE_QUIT
};

struct BmpSystemMessage
{
    BmpSystemMessageType message_type;
    gpointer             data;
    GDestroyNotify       data_destroy;
};

GType bmp_system_control_get_type (void);
BmpSystemControl *bmp_system_control_new (gboolean no_remote);

void bmp_system_control_init_finalize (BmpSystemControl *control);

void bmp_system_message_send (BmpSystemControl     *control,
                              BmpSystemMessageType  message_type,
                              gpointer              data,
                              GDestroyNotify        data_destroy);

void bmp_system_control_update    (BmpSystemControl *control);

void bmp_system_control_app_busy  (BmpSystemControl *control);

void bmp_system_control_app_idle  (BmpSystemControl *control);

void bmp_system_control_clear_playback_history (BmpSystemControl *control);
  
gboolean
bmp_system_control_ui_start   (BmpSystemControl   *control,
                               gchar              *display_name,
                               GError            **error);

gboolean
bmp_system_control_ui_stop    (BmpSystemControl   *control,
                               GError            **error);
gboolean
bmp_system_control_ui_raise   (BmpSystemControl   *control,
                               GError            **error);

gboolean
bmp_system_control_seek_percent   (BmpSystemControl   *control,
                                   gdouble             position,
                                   GError            **error);

gboolean
bmp_system_control_play_track   (BmpSystemControl   *control,
                                 gint                index,
                                 GError            **error);

void
bmp_system_control_play_anonymous       (BmpSystemControl *control,
                                         const char       *uri);

gboolean
bmp_system_control_export_tracklist (BmpSystemControl *control,
                                     const gchar      *container_plugin,
                                     gboolean          all,
                                     const gchar      *filename,
                                     GError          **error);

gboolean
bmp_system_control_volume_set     (BmpSystemControl *control,
                                   int               volume,
                                   GError          **error);

gboolean
bmp_system_control_go_prev  (BmpSystemControl *control, GError **error);

gboolean
bmp_system_control_go_next  (BmpSystemControl *control, GError **error);

gboolean
bmp_system_control_pause    (BmpSystemControl *control, GError **error);

gboolean
bmp_system_control_play     (BmpSystemControl *control, GError **error);

gboolean
bmp_system_control_stop     (BmpSystemControl *control, GError **error);

gboolean
bmp_system_control_add_uri_list (BmpSystemControl  *control,
                                 const gchar      **uri_list,
                                 gint               position,
                                 gboolean           clear,
                                 gboolean           start_playback,
                                 gint               start_payback_position,
                                 gint              *position_out,
                                 GError           **error);


gboolean
bmp_system_control_stop_after_current_set (BmpSystemControl   *control,
                                           gboolean            stop,
                                           GError            **error);

gboolean
bmp_system_control_repeat_set   (BmpSystemControl   *control,
                                 gboolean            repeat,
                                 GError            **error);

gboolean
bmp_system_control_repeat_get   (BmpSystemControl   *control,
                                 gboolean           *repeat,
                                 GError            **error);

gboolean
bmp_system_control_shuffle_set    (BmpSystemControl   *control,
                                   gboolean            shuffle,
                                   GError            **error);

gboolean
bmp_system_control_shuffle_get    (BmpSystemControl  *control,
                                   gboolean          *shuffle,
                                   GError           **error);

gboolean
bmp_system_control_identity (BmpSystemControl   *control,
                             gchar             **name,
                             GError            **error);


gboolean
bmp_system_control_get_tracklist    (BmpSystemControl    *control,
                                     char              ***array,
                                     GError             **error);

gboolean
bmp_system_control_get_bookmarks    (BmpSystemControl  *control,
                                     GArray           **array,
                                     GError           **error);

gboolean
bmp_system_control_get_current_title    (BmpSystemControl  *control,
                                         gchar             **title,
                                         GError            **error);
gboolean
bmp_system_control_get_current_uri        (BmpSystemControl *control,
                                           char            **uri,
                                           GError          **error);
gboolean
bmp_system_control_get_current_track      (BmpSystemControl *control,
                                           int              *idx,
                                           GError          **error);

gboolean
bmp_system_control_get_current_samplerate (BmpSystemControl  *control,
                                           int               *samplerate,
                                           GError           **error);
 
gboolean
bmp_system_control_get_current_bitrate    (BmpSystemControl  *control,
                                           int               *bitrate,
                                           GError           **error);

gboolean
bmp_system_control_get_current_stream_pos (BmpSystemControl  *control,
                                           int               *position,
                                           GError           **error);

gboolean
bmp_system_control_get_current_play_status (BmpSystemControl  *control,
                                           int               *playstatus,
                                           GError           **error);

gboolean
bmp_system_control_get_metadata_for_list_item (BmpSystemControl   *control,
                                               int                 item,
                                               GHashTable        **value,
                                               GError            **error);

gboolean
bmp_system_control_get_metadata_for_uri       (BmpSystemControl   *control,
                                               char               *uri,
                                               GHashTable        **value,
                                               GError            **error);

gboolean
bmp_system_control_send_status                (BmpSystemControl    *control,
                                               GError             **error);

gboolean
bmp_system_control_volume_get (BmpSystemControl   *control,
                              int                 *volume,
                              GError             **error);

gboolean
bmp_system_control_volume_set (BmpSystemControl   *control,
                               int                 volume,
                               GError            **error);

gboolean
bmp_system_control_seek       (BmpSystemControl   *control,
                               int                 position,
                               GError            **error);

void
bmp_system_control_initialize_dbus_connection (BmpSystemControl *control);

/* System startup/shutdown stuff */
typedef GThreadFunc SystemShutdownFunc;
typedef GThreadFunc SystemStartupFunc;
void
bmp_system_control_register_shutdown_func     (BmpSystemControl     *control,
                                               SystemShutdownFunc    shutdown_func,
                                               gpointer              shutdown_data);

void
bmp_system_control_register_startup_func      (BmpSystemControl     *control,
                                               SystemStartupFunc     startup_func,
                                               gpointer              startup_data);

void
bmp_system_control_application_run (BmpSystemControl      *control);

gboolean
bmp_system_control_quit (BmpSystemControl  *control,
                         GError           **error);

gint
bmp_system_control_message_domain_register (BmpSystemControl  *control,
                                            const gchar       *id,
                                            const gchar       *description);

gint
bmp_system_control_message_dispatch     (BmpSystemControl     *control,
                                         int                   domain_id,
                                         GtkMessageType        msgtype,
                                         const gchar          *text);

void
bmp_system_control_remove_guid        (BmpSystemControl     *control,
                                       RowGUID              *guid);

// Dummy method for D-BUS activation
gboolean
bmp_system_control_startup  (BmpSystemControl *control, GError **error);

// FIXME: That's not so good here, Al
GObject*
bmp_system_control_get_current_plugin (BmpSystemControl *control);

#endif // BMP_SYSTEM_CONTROL_HPP
