//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License Version 2
//  as published by the Free Software Foundation.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_TRACK_INFO_HPP
#define BMP_TRACK_INFO_HPP

#include <gtkmm.h>
#include <libglademm.h>
#include <boost/optional.hpp>

#include "database.hpp"
#include "library.hpp"
#include "lyrics.hpp"

namespace Bmp
{
  class EventBoxXDS;

  class TrackInfo
      : public Gtk::Window
  {
  public:

      TrackInfo (BaseObjectType                        *cobject,
                 const Glib::RefPtr<Gnome::Glade::Xml> &xml);

      static TrackInfo *
      create ();

      ~TrackInfo ();

      void
      show (Glib::ustring const& uri);

      void
      clear ();

  private:

      const std::string path_image_default;

      enum Rating
      {
        RATING_0,
        RATING_1,
        RATING_2,
        RATING_3,
        RATING_4,
        RATING_5,
  
        N_RATINGS,
      };

      Glib::RefPtr<Gdk::Pixbuf> rating[N_RATINGS];

      void
      check_update (Glib::ustring const&      pkey,
                    Bmp::Library::UpdateTrack track);

      bool
      rating_button_press (GdkEventButton *event);
 
      void
      ok_clicked ();

      void
      get_lyrics ();

      void
      on_drag_start (int n_tracks);
      void
      on_drag_step  ();
      void
      on_drag_end   ();

      Glib::RefPtr<Gnome::Glade::Xml> ref_xml;

      Gtk::Button   *ok;
      Gtk::Image    *image;
      Gtk::TextView *lyrics;

      Gtk::Button   *b_lyrics;

      Gtk::Entry    *entry_artist;
      Gtk::Entry    *entry_album;
      Gtk::Entry    *entry_title;
      Gtk::Entry    *entry_file;
      Gtk::Entry    *entry_comment;
      Gtk::Entry    *entry_genre;

      Gtk::Label    *label_date;
      Gtk::Label    *label_track;
      Gtk::Label    *label_bitrate;
      Gtk::Label    *label_count;

      Gtk::Image    *i_rating;
      Gtk::EventBox *evb_rating;
      
      Gtk::Table    *table_mutable_info;

      Bmp::EventBoxXDS *ebox;

      boost::optional<Glib::ustring> location;
      boost::optional<Glib::ustring> artist;
      boost::optional<Glib::ustring> album;

      Bmp::Library::Track track;

      int n_drag_tracks;
      int n_drag_current;

      Gtk::ProgressBar *copy_progress;
     
      
  };

} // namespace Bmp

#endif // BMP_TRACK_INFO_HPP

