//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <cstring>
#include <glib/gi18n.h>
#include <glibmm.h>
#include <gtkmm.h>
#include <gdk/gdkx.h>
#include <gdk/gdkkeysyms.h>

#include <chroma/chroma-list.h>
#include <chroma/chroma-list-selection.h>
#include <chroma/chroma-vscrollbar.h>
#include <chroma/chroma-entry.h>
#include <goa/libgoa.h>
#include <boost/format.hpp>

#include "widgets/bmp_window.h"

#include "winamp2_skin.hpp"
#include "winamp2_main.hpp"
#include "winamp2_playlist.hpp"

#include "action_defines.hpp"

#include "main.hpp"
#include "paths.hpp"
#include "ui_util.hpp"

#include "x_library.hpp"
#include "x_play.hpp"
#include "x_playlist.hpp"
#include "x_system_control.hpp"

#include "shell.hpp"

using namespace Bmp;

#include "gmacros.hpp"
BMP_DEFINE_TYPE (BmpWindowPlaylist, bmp_window_playlist, G_TYPE_OBJECT)

#define BMP_UIC_PL_INFORMATION        "bmp_component_pl_information"

#define BMP_UIC_PL_TITLE              "bmp_component_pl_title"
#define BMP_UIC_PL_TITLE_NOFOCUS      "bmp_component_pl_title_nf"

#define BMP_UIC_PL_PAD_TOP            "bmp_component_pl_pad_top"
#define BMP_UIC_PL_PAD_TOP_NOFOCUS    "bmp_component_pl_pad_top_nf"

#define BMP_UIC_PL_PAD_BOTTOM         "bmp_component_pl_pad_bottom"

#define BMP_UIC_PL_CORNER_UL          "bmp_component_pl_corner_ul"
#define BMP_UIC_PL_CORNER_UL_NOFOCUS  "bmp_component_pl_corner_ul_nf"

#define BMP_UIC_PL_CORNER_UR          "bmp_component_pl_corner_ur"
#define BMP_UIC_PL_CORNER_UR_NOFOCUS  "bmp_component_pl_corner_ur_nf"

#define BMP_UIC_PL_EDGE_LEFT          "bmp_component_pl_edge_left"
#define BMP_UIC_PL_EDGE_RIGHT         "bmp_component_pl_edge_right"

#define BMP_UIC_PL_CORNER_LL          "bmp_component_pl_corner_ll"
#define BMP_UIC_PL_CORNER_LR          "bmp_component_pl_corner_lr"

#define VSCROLLBAR_DEFAULT_WIDTH 14
#define PLAYLIST_INNER_XMARGIN 22
#define PLAYLIST_INNER_YMARGIN 42
#define PLAYLIST_ROOT_X 12
#define PLAYLIST_ROOT_Y 24
#define PLAYLIST_INNERPAD 2

#define REGION_L(x1,x2,y1,y2,w,h)                   \
    (event->x >= (x1) && event->x < (x2) &&     \
     event->y >= h - (y1) &&  \
     event->y < h - (y2))

#define REGION_R(x1,x2,y1,y2,w,h) \
    ((event->x >= (w-x1)) && \
     (event->x < (w-x2)) && \
     (event->y >= (h-y1)) && \
     (event->y < (h-y2)))

struct _BmpWindowPlaylistPrivate
{
    bool        dispose_has_run;
    GHashTable *ui_components;
    int         columns_visible;
    bool        focused;
    bool        edge;
    
    char      **dnd_strv;
    int         dnd_index; 
    gboolean    dnd_selection_done;
};

GtkWidget *wTracklist = 0;
GtkWidget *wTracklistScrollbar = 0;

namespace
{
  struct TBmpPlaylistComponent
  {
    char               *name;
    unsigned int        x_origin;
    unsigned int        y_origin;
    unsigned int        width;
    unsigned int        height;
    GdkPixbuf          *pixbuf;
  };

  TBmpPlaylistComponent playlist_component_data[] = 
  {
      {BMP_UIC_PL_TITLE,              26,   0, 100, 20 },
      {BMP_UIC_PL_TITLE_NOFOCUS,      26,  21, 100, 20 },

      {BMP_UIC_PL_PAD_TOP,            127,  0,  25, 20 },
      {BMP_UIC_PL_PAD_TOP_NOFOCUS,    127, 21,  25, 20 },

      {BMP_UIC_PL_PAD_BOTTOM,         179,  0,  25, 38 },

      {BMP_UIC_PL_CORNER_UL,          0,    0,  25, 20 },
      {BMP_UIC_PL_CORNER_UL_NOFOCUS,  0,   21,  25, 20 },

      {BMP_UIC_PL_CORNER_UR,          153,  0,  25, 20 },
      {BMP_UIC_PL_CORNER_UR_NOFOCUS,  153, 21,  25, 20 },

      {BMP_UIC_PL_EDGE_LEFT,          0,   42,  12, 29 },
      {BMP_UIC_PL_EDGE_RIGHT,         32,  42,  19, 29 },

      {BMP_UIC_PL_CORNER_LL,          0,   72, 125, 38 },
      {BMP_UIC_PL_CORNER_LR,          126, 72, 150, 38 },

      {BMP_UIC_PL_INFORMATION,        205,  0,  75, 32 }

  };

  void
  pledit_copy_area (GdkPixbuf    *dest,
                               unsigned int  x,
                               unsigned int  y,
                               unsigned int  width,
                               unsigned int  height)
  {
    gdk_pixbuf_copy_area (wa2_skin_bitmap[BITMAP_PLEDIT], x, y, width, height, dest, 0, 0);
  }


  GdkPixbuf*
  pledit_new_pb_simple (unsigned int width, unsigned int height)
  {
    return gdk_pixbuf_new (gdk_pixbuf_get_colorspace (wa2_skin_bitmap[BITMAP_PLEDIT]),
                           gdk_pixbuf_get_has_alpha (wa2_skin_bitmap[BITMAP_PLEDIT]),
                           gdk_pixbuf_get_bits_per_sample (wa2_skin_bitmap[BITMAP_PLEDIT]),
                           width, height);
  }

  void
  pledit_draw_pixbuf (cairo_t     *cr,
                          GdkPixbuf   *pb,
                          unsigned int x,
                          unsigned int y)
  {
      cairo_save (cr);
      gdk_cairo_set_source_pixbuf (cr, pb, x, y);
      cairo_paint (cr);
      cairo_restore (cr);
  }

  void
  set_cursors (BmpWindowPlaylist *wa2_playlist)
  {
    if (!wa2_playlist) return;
    GtkWidget *window = wa2_playlist->window;
    if (!GTK_WIDGET_REALIZED(window)) return;
    SET_CURSOR_FOR_WIDGET (CURSOR_PLAYLIST, GTK_WINDOW(window));
  }

  void
  tracklist_has_sel (BmpWindowPlaylist *wa2_playlist, gboolean active)
  {
    actions->get_action (BMP_ACTION_DEL)->set_sensitive (active);
  }

  void
  tracklist_has_items (BmpWindowPlaylist *wa2_playlist, gboolean active)
  {
    actions->get_action (BMP_ACTION_DEL_ALL)->set_sensitive (active);
    actions->get_action (BMP_ACTION_PLAY)->set_sensitive (active);
  }

  void
  tracklist_set_action_states (BmpWindowPlaylist *wa2_playlist, bool has_selection) 
  {
    if ((!has_selection) && (!chroma_list_selection_multiple_selected (CHROMA_LIST_SELECTION(wTracklist))))
      tracklist_has_sel (wa2_playlist, FALSE);
    else 
      tracklist_has_sel (wa2_playlist, TRUE);

    tracklist_has_items (wa2_playlist, (chroma_list_get_length (CHROMA_LIST(wTracklist)) == 0) ? FALSE : TRUE);
  }

  void
  destroy_value (gpointer data)
  {
    TBmpPlaylistComponent *component;
    component = (TBmpPlaylistComponent *) data;
    if (component)
    {
      if (component->pixbuf) g_object_unref (component->pixbuf);
      free(component);
    }
  }

  int
  tracklist_compare_func
            (GtkTreeModel *model,
                 GtkTreeIter  *a,
                 GtkTreeIter  *b,
                 gpointer      data)

  {
    int cmp_result;
    Bmp::Library::DatumDefine define = library->get_metadatum_define (vcolumns_get_item_nth (GPOINTER_TO_INT(data)));
    

    if (define.type == Bmp::DB::VALUE_TYPE_INT)
      {
        int val_a, val_b;
        gtk_tree_model_get (model, a, GPOINTER_TO_INT(data), &val_a, -1);
        gtk_tree_model_get (model, b, GPOINTER_TO_INT(data), &val_b, -1);
        return val_a - val_b;
      }
    else
      {
        char  *text_a,
              *text_b;

        char  *string_a,
              *string_b;

        gtk_tree_model_get (model, a, GPOINTER_TO_INT(data), &text_a, -1);
        gtk_tree_model_get (model, b, GPOINTER_TO_INT(data), &text_b, -1);

        if ( (text_a != NULL) && (text_b == NULL)) return  1;
        if ( (text_a == NULL) && (text_b == NULL)) return  0;
        if ( (text_a == NULL) && (text_b != NULL)) return -1;

        string_a = g_utf8_casefold (text_a, std::strlen(text_a));
        string_b = g_utf8_casefold (text_b, std::strlen(text_b));

        if ( (string_a != NULL) && (string_b == NULL)) return  1;
        if ( (string_a == NULL) && (string_b == NULL)) return  0;
        if ( (string_a == NULL) && (string_b != NULL)) return -1;

        cmp_result = g_utf8_collate (string_a, string_b);

        free (string_a);
        free (string_b);
        free (text_a);
        free (text_b);

        return cmp_result;
      }
  }

  void
  tracklist_cell_data_func_time   (ChromaList      *wa2_playlist,
                                   GtkTreeModel    *model,
                                   GtkTreeIter     *iter,
                                   GValue          *value,
                                   gpointer         data)
  {
    int v; 
    gtk_tree_model_get (model, iter, COLUMN_DURATION, &v, -1);

    int minutes = v / 60;
    int seconds = v % 60;

    g_value_unset (value);
    g_value_init  (value, G_TYPE_STRING);

    char *s_time = g_strdup_printf ("%d:%-2.2d", minutes, seconds);
    g_value_set_string (value, s_time);
    free (s_time);
  }

  void
  tracklist_cell_data_func_track   (ChromaList      *wa2_playlist,
                                    GtkTreeModel    *model,
                                    GtkTreeIter     *iter,
                                    GValue          *value,
                                    gpointer         data)
  {
    int v; 
    gtk_tree_model_get (model, iter, COLUMN_TRACK, &v, -1);
    if (v == 0)
      {
        g_value_unset (value);
        g_value_init (value, G_TYPE_STRING);
        g_value_set_string (value, "");
      }
  }

  void
  tracklist_cell_data_func_date    (ChromaList      *wa2_playlist,
                                    GtkTreeModel    *model,
                                    GtkTreeIter     *iter,
                                    GValue          *value,
                                    gpointer         data)
  {
    int v; 
    gtk_tree_model_get (model, iter, COLUMN_DATE, &v, -1);
    if (v == 0)
      {
        g_value_unset (value);
        g_value_init (value, G_TYPE_STRING);
        g_value_set_string (value, "");
      }
  }

  TBmpPlaylistComponent*
  playlist_get_component(BmpWindowPlaylist *wa2_playlist, const char *key)
  {
    return static_cast<TBmpPlaylistComponent*>(g_hash_table_lookup(wa2_playlist->priv->ui_components, key));
  }

  void
  set_geometry_hints(BmpWindowPlaylist *wa2_playlist)
  {
    GtkWindow *window = GTK_WINDOW(wa2_playlist->window);
    GdkGeometry geometry;
    GdkWindowHints mask;

    geometry.width_inc = 25;
    geometry.height_inc = 29;

    /* make sure the min width/height are multiple values of the resize increments */
    geometry.min_width = ((int) (BMP_WINDOW_MAIN_WIDTH * 1 / geometry.width_inc)) * geometry.width_inc;
    geometry.max_width = 0xFFFF;
    geometry.base_width = 0;

    /* make sure the min width/height are multiple values of the resize increments */
    geometry.min_height = ((int) (BMP_WINDOW_MAIN_HEIGHT * 2 / geometry.height_inc)) * geometry.height_inc;
    geometry.max_height = 0xFFFF;
    geometry.base_height = 0;

    mask = GdkWindowHints (GDK_HINT_MIN_SIZE | GDK_HINT_MAX_SIZE | GDK_HINT_RESIZE_INC | GDK_HINT_BASE_SIZE);

    gtk_window_set_geometry_hints  (GTK_WINDOW(window),
                                    GTK_WIDGET(window),
                                    &geometry,
                    mask);

  }

  gboolean
  on_window_leave  (GtkWidget           *widget,
                    GdkEventCrossing    *event,
                    BmpWindowPlaylist   *wa2_playlist)
  {
      wa2_playlist->priv->edge = false;
      set_cursors (wa2_playlist);
      return FALSE;
  }

  gboolean
  on_window_motion (GtkWidget           *widget,
                    GdkEventMotion      *event,
                    BmpWindowPlaylist   *wa2_playlist)
  {
    int x, y;
    GdkModifierType state;

    if (event->is_hint)
      {
        gdk_window_get_pointer (event->window, &x, &y, &state);
      }
    else
      {
        x = int(event->x);
        y = int(event->y);
        state = GdkModifierType(event->state);
      }

      int w = GTK_WIDGET (widget)->allocation.width;
      int h = GTK_WIDGET (widget)->allocation.height;

      if (((w-event->x) <= 24) && ((h-event->y) <=24))
    {
      if (wa2_playlist->priv->edge) return FALSE;
      wa2_playlist->priv->edge = true;
      GdkCursor *cursor = gdk_cursor_new_from_name (gdk_display_get_default (), "bottom_right_corner");
      if (!cursor) cursor = gdk_cursor_new_for_display (gdk_display_get_default (), GDK_BOTTOM_RIGHT_CORNER);
      gdk_window_set_cursor (GTK_WIDGET(widget)->window, cursor);
    }
      else
    {
      if (!wa2_playlist->priv->edge) return FALSE;
      set_cursors (wa2_playlist);
    }
      return FALSE;
  }

  gboolean
  on_window_delete_event
              (GtkWidget      *widget,
               GdkEvent       *event,
               BmpWindowPlaylist  *wa2_playlist)
  {
    Glib::RefPtr<Gtk::ToggleAction> action = Glib::RefPtr<Gtk::ToggleAction>::cast_static
                        (actions_toggle->get_action (BMP_TOGGLE_ACTION_PL));
    action->set_active (false);
    return TRUE;
  }

  gboolean
  on_window_focus_in  (GtkWidget          *widget,
                       GdkEventFocus      *event,
                       BmpWindowPlaylist  *wa2_playlist)
  {
    wa2_playlist->priv->focused = true;
    gtk_widget_queue_draw (widget);
    return FALSE;
  }

  gboolean
  on_window_focus_out
              (GtkWidget          *widget,
               GdkEventFocus      *event,
               BmpWindowPlaylist  *wa2_playlist)
  {
    wa2_playlist->priv->focused = false;
    gtk_widget_queue_draw (widget);
    return FALSE;
  }

  gboolean
  on_window_key_press (GtkWidget    *widget,
                       GdkEventKey  *event)
  {
    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    if (event->keyval != GDK_Escape) return FALSE;
    actions->get_action (BMP_ACTION_ICONIFY)->activate ();
    return TRUE;
  }

  gboolean
  on_window_expose    (GtkWidget          *widget,
                       GdkEventExpose     *event,
                       BmpWindowPlaylist  *wa2_playlist)
  {
      if (skin_changing) return FALSE;

      TBmpPlaylistComponent *component    = 0;
      cairo_t               *cr       = 0;
      unsigned int           w_width  = 0,
                             w_height = 0,
                             pad_ctr  = 0;

      bool focus  = wa2_playlist->priv->focused; 
      w_width     = widget->allocation.width;
      w_height    = widget->allocation.height;

      cr = gdk_cairo_create (widget->window);

      pad_ctr =  (w_width - 150)/25;

      component = playlist_get_component(wa2_playlist, focus ? BMP_UIC_PL_CORNER_UL : BMP_UIC_PL_CORNER_UL_NOFOCUS);
      pledit_draw_pixbuf(cr, component->pixbuf, 0, 0);

      component = playlist_get_component(wa2_playlist, focus ? BMP_UIC_PL_CORNER_UR : BMP_UIC_PL_CORNER_UR_NOFOCUS);
      pledit_draw_pixbuf(cr, component->pixbuf, (w_width-component->width), 0);

      component = playlist_get_component(wa2_playlist, focus ? BMP_UIC_PL_TITLE : BMP_UIC_PL_TITLE_NOFOCUS);
      pledit_draw_pixbuf(cr, component->pixbuf, (w_width/2)-(50), 0);

      component = playlist_get_component(wa2_playlist, BMP_UIC_PL_CORNER_LL);
      pledit_draw_pixbuf(cr, component->pixbuf, 0, (w_height-component->height));

      component = playlist_get_component(wa2_playlist, BMP_UIC_PL_CORNER_LR);
      pledit_draw_pixbuf(cr, component->pixbuf, (w_width-component->width), (w_height-component->height));

      component = playlist_get_component(wa2_playlist, focus ? BMP_UIC_PL_PAD_TOP : BMP_UIC_PL_PAD_TOP_NOFOCUS);

      for (unsigned int i = 0; i < pad_ctr / 2; ++i)
    {

        // Left of title
        pledit_draw_pixbuf(cr, component->pixbuf, 25 + i * 25, 0);

        // Right of title
        pledit_draw_pixbuf(cr, component->pixbuf, (w_width+100)/2+i*25, 0);
    }

      if (pad_ctr & 1)
    {
        /* Odd tile count, so one remaining to draw. Here we split
         * it into two and draw half on either side of the title */

        gdk_draw_pixbuf(widget->window,
                NULL,
                component->pixbuf,
                0,
                0,
                (w_width/2)+((pad_ctr/2)*25) +50,
                0,
                13,
                -1,
                GDK_RGB_DITHER_NONE,
                0,
                0);

        gdk_draw_pixbuf(widget->window,
                NULL,
                component->pixbuf,
                0,
                0,
                ((pad_ctr / 2) * 25) + 25,
                0,
                13,
                -1,
                GDK_RGB_DITHER_NONE,
                0,
                0);
      }

    // Frame sides
    for (unsigned int i = 0; i < (w_height - (20 + 38)) / 29; ++i)
      {
        component = playlist_get_component(wa2_playlist, BMP_UIC_PL_EDGE_LEFT);
        gdk_draw_pixbuf(widget->window,
                NULL,
                component->pixbuf,
                0,
                0,
                0,
                20 + i *29,
                -1,
                -1,
                GDK_RGB_DITHER_NONE,
                0,
                0);

        component = playlist_get_component(wa2_playlist, BMP_UIC_PL_EDGE_RIGHT);
        gdk_draw_pixbuf(widget->window,
                NULL,
                component->pixbuf,
                0,
                0,
                w_width-19,
                20 + i *29,
                -1,
                -1,
                GDK_RGB_DITHER_NONE,
                0,
                0);

      }

    pad_ctr =  (w_width - 275) / 25;
    component = playlist_get_component(wa2_playlist, BMP_UIC_PL_PAD_BOTTOM);

    for (unsigned int i = 0; i < pad_ctr+4; ++i)
      {
        pledit_draw_pixbuf(cr, component->pixbuf, (125+i*25), (w_height-component->height));
      }
    pledit_draw_pixbuf(cr, component->pixbuf, (125+(pad_ctr+3)*25) + 5, (w_height-component->height));
    cairo_destroy (cr);
    return FALSE;
  }

  void
  on_bmp_display_tracklist_numbers_changed
  (MCS_CB_DEFAULT_SIGNATURE, BmpWindowPlaylist *wa2_playlist)
  {
    chroma_list_set_use_numbers (CHROMA_LIST(wTracklist), boost::get<bool>(value));
  }

  void
  on_bmp_use_custom_cursors_changed
  (MCS_CB_DEFAULT_SIGNATURE, BmpWindowPlaylist *wa2_playlist)
  {
    bmp_window_playlist_configure (wa2_playlist);
  }

  void
  on_bmp_font_changed
  (MCS_CB_DEFAULT_SIGNATURE, BmpWindowPlaylist *wa2_playlist)
  {
    chroma_list_set_font (CHROMA_LIST(wTracklist), pango_font_description_from_string (boost::get<std::string>(value) . c_str ()));
  }

  gboolean
  on_window_button_press (GtkWidget         *widget,
                          GdkEventButton    *event,
                          BmpWindowPlaylist *wa2_playlist)  
  {
    int        x,
               y,
               w_width,
               w_height,
               w_xpos,
               w_ypos;

    g_return_val_if_fail (GTK_IS_WIDGET (widget), FALSE);

    x = int (event->x);
    y = int (event->y);
    w_width  = GTK_WIDGET (widget)->allocation.width;
    w_height = GTK_WIDGET (widget)->allocation.height;

    gtk_window_get_position (GTK_WINDOW (wa2_playlist->window), &w_xpos, &w_ypos);

    if ((event->button == 3) && (event->type == GDK_BUTTON_PRESS))
      {
        GtkTreePath *path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION(wTracklist));
        tracklist_set_action_states (wa2_playlist, (path != NULL));
        bool cond = ((path != NULL) && (!chroma_list_selection_multiple_selected (CHROMA_LIST_SELECTION (wTracklist))));
        gtk_tree_path_free (path);

        actions->get_action (BMP_ACTION_FILEINFO)->set_sensitive (cond);
#ifdef HAVE_GAIM
        if (gaim_dbus) actions->get_action (BMP_ACTION_SEND_TRACK)->set_sensitive (cond);
#endif
        if (cond && mcs->key_get<bool>("lastfm", "submit-enable"))
            actions->get_action (BMP_ACTION_LASTFM_RECOMMEND)->set_sensitive (true);
        else
            actions->get_action (BMP_ACTION_LASTFM_RECOMMEND)->set_sensitive (false);

        gtk_menu_popup (GTK_MENU (ui_manager_get_popup (ui_manager->gobj(), "/popup-tracklist/menu-tracklist")),
                        NULL, NULL, NULL, NULL, event->button, event->time);
        return TRUE;
      }
    else
    if ((event->button == 1) && (event->type == GDK_BUTTON_PRESS))
      {
        if (REGION_L(12, 37, 29, 11, w_width, w_height))
          {
           //ADD button
            GtkWidget *menu;
            GtkRequisition req;
            menu = ui_manager_get_popup (ui_manager->gobj(), "/popup-tracklist/menu-add");
            gtk_widget_size_request (menu, &req);
            bmp_menu_popup (GTK_MENU (menu), (w_xpos+14), (w_ypos+w_height) - 8 - req.height, event->button, event->time);
            return TRUE;
          }
        else
        if (((w_width-event->x) <= 24) && ((w_height-event->y) <=24))
          {
            gtk_window_begin_resize_drag (GTK_WINDOW (widget),
                                          GDK_WINDOW_EDGE_SOUTH_EAST,
                                          event->button,
                                          int (event->x_root),
                                          int (event->y_root),
                                          event->time);
            return TRUE;
          }
        else
        if ((y < 16) && (x < (w_width - 12)))
          {
            gdk_window_raise (widget->window);
            gtk_window_begin_move_drag (GTK_WINDOW (widget),
                                         event->button,
                                         int (event->x_root),
                                         int (event->y_root),
                                         event->time);
            return TRUE;
          }
        else
        if ((y < 16) && (x > (w_width - 12)))
          {
            Glib::RefPtr<Gtk::ToggleAction> action = 
            Glib::RefPtr<Gtk::ToggleAction>::cast_static (actions_toggle->get_action (BMP_TOGGLE_ACTION_PL));
            action->set_active (false);
            return TRUE;
          }
      }
    return FALSE;
  }

  void
  on_tracklist_cursor_position
                          (ChromaList        *list,
                           int                row,
                           BmpWindowPlaylist *wa2_playlist)
  {
    GtkTreePath *path;

    path = chroma_list_selection_get_selected (CHROMA_LIST_SELECTION(list));
    tracklist_set_action_states (wa2_playlist, (path != NULL));
    gtk_tree_path_free (path);

    if ((row == -1) || (!path)) 
        actions->get_action (BMP_ACTION_FILEINFO_CLEAR)->activate ();
    else
        actions->get_action (BMP_ACTION_FILEINFO_IF_VISIBLE)->activate ();

    bmp_system_control_update (bmp_system_control);
  }

  void
  on_tracklist_dnd      (ChromaList         *list,
                         int                 index,
                         char              **uri_strv,
                         BmpWindowPlaylist  *wa2_playlist)
  {
    if (mcs->key_get<bool>("bmp", "playlist-show-dnd-menu"))
      {
        wa2_playlist->priv->dnd_strv  = g_strdupv (uri_strv);
        wa2_playlist->priv->dnd_index = index; 
    
        GtkWidget *menu = ui_manager_get_popup (ui_manager->gobj(), "/popup-tracklist-dnd/menu-tracklist-dnd");
        gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL, 1, 0L); 
      }
    else
      {
        int out;
        bmp_system_control_add_uri_list (bmp_system_control,
                (const char**) uri_strv, 
                index, 
                FALSE,
                FALSE,
                0,
                &out,
                NULL);
      }
  }

  void
  on_tracklist_row_activated
                          (ChromaList *list,
                           int         index)
  {
    bmp_system_control_play_track (bmp_system_control, index, NULL);
  }

  void
  on_system_control_items_removed
                          (BmpSystemControl *system_control,
                           GArray           *removed_rows,
                           int               n_removed_rows,
                           gpointer          data) 
  {
    if (n_removed_rows == -1)
    {
      chroma_list_update (CHROMA_LIST (wTracklist));
      actions->get_action (BMP_ACTION_CLEAR_PLAYBACK_HISTORY)->set_sensitive (false);
    }
    actions->get_action (BMP_ACTION_FILEINFO_CLEAR)->activate ();
  }

  void
  on_system_control_track_change (BmpSystemControl  *system_control,
                                  BmpWindowPlaylist *wa2_playlist)
  {
    actions->get_action (BMP_ACTION_CLEAR_PLAYBACK_HISTORY)->set_sensitive (true);

    if (mcs->key_get<bool>("bmp", "follow-current-track"))
      {
        int position  = g_object_get_int (bmp_system_control, "current_index");
        int display   = chroma_list_get_n_item_display (CHROMA_LIST(wTracklist));
        int length    = chroma_list_get_length (CHROMA_LIST(wTracklist));
        int scroll    = position - (display/2);

        if (scroll > (length-display)) scroll = (length-display);
        if (scroll < 0) scroll = 0;

        chroma_list_scroll_to_offset (CHROMA_LIST(wTracklist), scroll);
      }

    gtk_widget_queue_draw (GTK_WIDGET(wTracklist));
  }

  void
  on_tracklist_path_removing
                      (ChromaList          *list,
                       GtkTreePath         *path,
                       BmpWindowPlaylist   *wa2_playlist)
  {
    bmp_playlist_row_removing (bmp_playlist, path);
  }

  void
  dnd_menu_cancel     (GtkWidget          *menu,
                       BmpWindowPlaylist  *wa2_playlist)
  {
    bmp_window_playlist_dnd_txn_cancel (wa2_playlist);
  }

  void
  playlist_row_swap_func (GtkTreeModel  *model,
                          int      pos_1,
                          int      pos_2)
  {
    GtkTreePath  *path_a, *path_b;
    GtkTreeIter   iter_a,  iter_b;

    path_a = gtk_tree_path_new_from_indices (pos_1, -1);
    path_b = gtk_tree_path_new_from_indices (pos_2, -1);
    gtk_tree_model_get_iter (model, &iter_a, path_a);
    gtk_tree_model_get_iter (model, &iter_b, path_b);
    gtk_tree_path_free (path_a);
    gtk_tree_path_free (path_b);

    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (model),
                      GTK_TREE_SORTABLE_UNSORTED_SORT_COLUMN_ID, 
                      GTK_SORT_ASCENDING);

    gtk_list_store_swap (GTK_LIST_STORE (model), &iter_a, &iter_b);
  }

  void
  create (BmpWindowPlaylist *wa2_playlist)
  {
    int              last_visible;
    double           sum = 0.0;

    wa2_playlist->window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_decorated (GTK_WINDOW (wa2_playlist->window), false);
    gtk_widget_set_app_paintable (wa2_playlist->window, true);
    gtk_widget_add_events (wa2_playlist->window,  GDK_FOCUS_CHANGE_MASK | GDK_BUTTON_MOTION_MASK |
                          GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK |
                          GDK_LEAVE_NOTIFY_MASK | GDK_SCROLL_MASK | GDK_VISIBILITY_NOTIFY_MASK | GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK );

    set_geometry_hints (wa2_playlist);

    gtk_window_set_title (GTK_WINDOW (wa2_playlist->window), _("Tracklist - BMP"));
    gtk_window_set_role (GTK_WINDOW (wa2_playlist->window), "bmp::window-playlist");
    bmp_window_set_icon_list (wa2_playlist->window, "playlist");

    GtkWidget *alignment = gtk_alignment_new (0.5,0.5,1.0,1.0);
    GtkWidget *box = gtk_hbox_new (false, 2);

#if 0
    GtkWidget *vbox = gtk_vbox_new (false, 2);
#endif

    gtk_alignment_set_padding (GTK_ALIGNMENT (alignment), PLAYLIST_ROOT_Y, PLAYLIST_INNER_YMARGIN,
                              PLAYLIST_ROOT_X, PLAYLIST_INNER_XMARGIN);

    wTracklist = chroma_list_new ();
    wTracklistScrollbar = chroma_vscrollbar_new ();
    g_object_set_data (G_OBJECT(wTracklistScrollbar), "list", wTracklist);

    chroma_list_set_adjustment (CHROMA_LIST (wTracklist),
                    CHROMA_VSCROLLBAR (wTracklistScrollbar));
                
    chroma_list_set_model (CHROMA_LIST (wTracklist), GTK_TREE_MODEL (bmp_playlist_get_playlist (bmp_playlist)),
                             ChromaListRowSwapFunc (playlist_row_swap_func));

    chroma_list_set_use_numbers (CHROMA_LIST (wTracklist),
                    mcs->key_get<bool>("bmp", "display-tracklist-numbers"));

    for (unsigned int n = 0; n < vcolumns_get_n_items(); n++)
      {
        Bmp::Library::DatumDefine define = library->get_metadatum_define (vcolumns_get_item_nth(n));

        chroma_list_set_column_header
                (CHROMA_LIST(wTracklist), n, _(define.title)); 

        chroma_list_set_column_show_header
                (CHROMA_LIST(wTracklist), n, TRUE);

        gtk_tree_sortable_set_sort_func
                (GTK_TREE_SORTABLE (bmp_playlist_get_playlist (bmp_playlist)),
                 n,
                 GtkTreeIterCompareFunc (tracklist_compare_func),
                 GINT_TO_POINTER (n),
                 NULL);
      }

    chroma_list_set_cell_data_func (CHROMA_LIST(wTracklist),
                                    COLUMN_TRACK,
                                    tracklist_cell_data_func_track,
                                    NULL); 
    chroma_list_set_column_xalign  (CHROMA_LIST(wTracklist),
                                    COLUMN_TRACK,
                                    1.0);

    chroma_list_set_cell_data_func (CHROMA_LIST(wTracklist),
                                    COLUMN_DURATION,
                                    tracklist_cell_data_func_time,
                                    NULL); 
    chroma_list_set_column_xalign  (CHROMA_LIST(wTracklist),
                                    COLUMN_DURATION,
                                    1.0);

    chroma_list_set_cell_data_func (CHROMA_LIST(wTracklist),
                                    COLUMN_DATE,
                                    tracklist_cell_data_func_date,
                                    NULL); 
    chroma_list_set_column_xalign  (CHROMA_LIST(wTracklist),
                                    COLUMN_DATE,
                                    1.0);

    chroma_list_set_column_visible (CHROMA_LIST(wTracklist), COLUMN_URI, FALSE);
    chroma_list_set_column_visible (CHROMA_LIST(wTracklist), COLUMN_GUID, FALSE);

    last_visible = 0;
    unsigned int n = 0;

    static boost::format f_visible ("column_%d_visible");
    static boost::format f_width ("column_%d_width");

    for (n = 0; n < vcolumns_get_n_items(); n++)
      {
        double width;
        bool visible;

        visible = mcs->key_get<bool>("tracklist", (f_visible % n).str());

        if (visible)
          {
            width = mcs->key_get<double>("tracklist", (f_width % n).str()); 
            wa2_playlist->priv->columns_visible += 1;
            sum += width;
            last_visible = n;
            chroma_list_set_column_width (CHROMA_LIST(wTracklist), n, width);
          }
        else
          {
            chroma_list_set_column_visible
            (CHROMA_LIST(wTracklist), n, FALSE);

            chroma_list_set_column_show_header
            (CHROMA_LIST(wTracklist), n, FALSE);
          }
      }

    if (sum != 1.0)
      {
        chroma_list_set_column_width (CHROMA_LIST(wTracklist), n, 1. - sum);
      }

    gtk_box_pack_start (GTK_BOX (box), GTK_WIDGET (wTracklist), TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (box), GTK_WIDGET (wTracklistScrollbar), FALSE, FALSE, 0);

#if 0
    GtkWidget *entry = chroma_entry_new ();
    gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (box), TRUE, TRUE, 0);
    gtk_box_pack_start (GTK_BOX (vbox), GTK_WIDGET (entry), FALSE, FALSE, 0);
    g_object_set_data (G_OBJECT (wa2_playlist->window), "entry", gpointer(entry));
#endif

    gtk_container_add (GTK_CONTAINER (wa2_playlist->window), alignment);
    gtk_container_add (GTK_CONTAINER (alignment), box);
#if 0
    gtk_container_add (GTK_CONTAINER (alignment), vbox);
#endif

    gtk_widget_set_size_request (GTK_WIDGET(wTracklistScrollbar), VSCROLLBAR_DEFAULT_WIDTH, -1);

    g_object_connect (G_OBJECT(alignment),
                      "swapped-signal::configure-event",
                      G_CALLBACK(gtk_widget_queue_draw),
                      wa2_playlist->window, NULL);
          
    g_object_connect (G_OBJECT(wTracklist),
                      "signal::cursor-position",
                      G_CALLBACK(on_tracklist_cursor_position),
                      wa2_playlist,
                      "signal::row-activated",
                      G_CALLBACK(on_tracklist_row_activated),
                      wa2_playlist,
                      "signal::dnd",
                      G_CALLBACK(on_tracklist_dnd),
                      wa2_playlist,
                      "signal::path-removing",
                      G_CALLBACK(on_tracklist_path_removing),
                      wa2_playlist, NULL);


    g_object_connect (G_OBJECT(wa2_playlist->window),
                      "signal::key-press-event",
                      G_CALLBACK(on_window_key_press),
                      wa2_playlist,
                      "signal::button-press-event",
                      G_CALLBACK(on_window_button_press),
                      wa2_playlist,
                      "signal::delete-event",
                      G_CALLBACK(on_window_delete_event),
                      wa2_playlist,
                      "signal::expose_event",
                      G_CALLBACK(on_window_expose),
                      wa2_playlist,
                      "signal::focus-in-event",
                      G_CALLBACK(on_window_focus_in),
                      wa2_playlist,
                      "signal::focus-out-event",
                      G_CALLBACK(on_window_focus_out),
                      wa2_playlist,
                      "signal::motion-notify-event",
                      G_CALLBACK(on_window_motion),
                      wa2_playlist,
                      "signal::leave-notify-event",
                      G_CALLBACK(on_window_leave),
                      wa2_playlist, NULL);

    Glib::RefPtr<Gtk::ToggleAction> toggle_action = 
          Glib::RefPtr<Gtk::ToggleAction>::cast_static (actions_toggle->get_action (BMP_TOGGLE_ACTION_FOLLOW_CURRENT));

    mcs_bind->bind_toggle_action (toggle_action, "bmp", "follow-current-track");
    gtk_widget_realize (GTK_WIDGET (wTracklist));
    mcs->subscribe ("Ui", "bmp", "font",
            sigc::bind (sigc::ptr_fun (&on_bmp_font_changed),
                wa2_playlist));
    mcs->subscribe ("Ui", "bmp", "use-custom-cursors",
            sigc::bind (sigc::ptr_fun(&on_bmp_use_custom_cursors_changed),
                    wa2_playlist));
    mcs->subscribe ("Ui", "bmp", "display-tracklist-numbers",
            sigc::bind (sigc::ptr_fun(&on_bmp_display_tracklist_numbers_changed),
                                wa2_playlist));

    g_object_connect (G_OBJECT(bmp_system_control),
                      "signal::track-change",
                      G_CALLBACK (on_system_control_track_change),
                      wa2_playlist,
                      "signal::tracklist-items-removed",
                      G_CALLBACK (on_system_control_items_removed),
                      wa2_playlist, NULL);

    Gtk::Window *w = Glib::wrap (GTK_WINDOW (wa2_playlist->window));
    w->add_accel_group (ui_manager->get_accel_group());

    actions->get_action (BMP_ACTION_CLEAR_PLAYBACK_HISTORY)->set_sensitive (false);
    actions->get_action (BMP_ACTION_ADD_CD)->set_sensitive (play->property_has_cdda().get_value());

    GtkWidget *menu = ui_manager_get_popup (ui_manager->gobj(), "/popup-tracklist-dnd/menu-tracklist-dnd");
    gtk_widget_realize (menu);
    g_signal_connect (G_OBJECT (menu), "cancel", G_CALLBACK (dnd_menu_cancel), wa2_playlist);
  }

  // GObject
  void
  bmp_window_playlist_init (BmpWindowPlaylist *wa2_playlist)
  {
    wa2_playlist->window = NULL;

    wa2_playlist->priv = g_new0 (BmpWindowPlaylistPrivate, 1);
    wa2_playlist->priv->dispose_has_run     = FALSE;
    wa2_playlist->priv->ui_components       = 0;
    wa2_playlist->priv->columns_visible     = 0;
    wa2_playlist->priv->focused             = FALSE;
    wa2_playlist->priv->edge                = FALSE;
    wa2_playlist->priv->dnd_strv            = 0;
    wa2_playlist->priv->dnd_index           = 0;
    wa2_playlist->priv->dnd_selection_done  = FALSE;

    create (wa2_playlist);
  }

  GObject*
  constructor (GType type,
           unsigned int n_construct_properties,
           GObjectConstructParam *construct_properties)
  {
    BmpWindowPlaylistClass *klass = BMP_WINDOW_PLAYLIST_CLASS (g_type_class_peek (BMP_TYPE_WINDOW_PLAYLIST));
    GObjectClass *parent_class = G_OBJECT_CLASS (g_type_class_peek_parent (klass));
    return parent_class->constructor (type,
                      n_construct_properties,
                      construct_properties);
  }
  void
  dispose (GObject *gobject)
  {
    BmpWindowPlaylist *wa2_playlist = (BmpWindowPlaylist *) gobject;

    if (wa2_playlist->priv->dispose_has_run) return;
    wa2_playlist->priv->dispose_has_run = TRUE;

    for (unsigned int n = 0; n < vcolumns_get_n_items(); n++)
      {
      gdouble width = chroma_list_get_column_width (CHROMA_LIST(wTracklist), n);
      char *key_name = g_strdup_printf ("column_%d_width", n);
      mcs->key_set<double>("tracklist", key_name, width);
      free (key_name);
      }

    chroma_list_set_model (CHROMA_LIST(wTracklist), NULL, NULL);
    gtk_widget_destroy (wTracklist);
    gtk_widget_destroy (wTracklistScrollbar);
    gtk_widget_destroy (wa2_playlist->window);
  }
  void
  bmp_window_playlist_class_init (BmpWindowPlaylistClass *klass)
  {
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = dispose;
    gobject_class->constructor = constructor;
  }

}

void
bmp_window_playlist_column_set_active (BmpWindowPlaylist *wa2_playlist,
                                       int                column,
                                       gboolean           visible)
{
      char *key_name;

      key_name = g_strdup_printf ("column_%d_visible", column);
      mcs->key_set<bool>("tracklist", key_name, visible);
      free (key_name);

      chroma_list_set_column_visible (CHROMA_LIST(wTracklist), column, visible);
      chroma_list_set_column_show_header (CHROMA_LIST(wTracklist), column, visible);

      if (visible)
        wa2_playlist->priv->columns_visible++;
      else
        wa2_playlist->priv->columns_visible--;

      for (unsigned int n = 0; n < vcolumns_get_n_items(); n++)
        {
            chroma_list_set_column_width (CHROMA_LIST(wTracklist), n, 1./wa2_playlist->priv->columns_visible);
            key_name = g_strdup_printf ("column_%d_width", n);
            mcs->key_set<double>("tracklist", key_name, 1./wa2_playlist->priv->columns_visible);
            free (key_name);
        }
      chroma_list_refresh (CHROMA_LIST(wTracklist));
}

void
bmp_window_playlist_show_trackinfo (BmpWindowPlaylist *wa2_playlist)
{
    GtkListStore      *list;
    GtkTreeRowReference   *reference;
    GtkTreePath       *path;

    list = bmp_playlist_get_playlist (bmp_playlist);
    reference = static_cast<GtkTreeRowReference *> (g_object_get_data (G_OBJECT(list), "current"));
    if (!reference) return;

    path = gtk_tree_row_reference_get_path (reference);
    chroma_list_set_cursor_position (CHROMA_LIST(wTracklist), gtk_tree_path_get_indices (path)[0]);
    gtk_tree_path_free (path);

    actions->get_action (BMP_ACTION_FILEINFO)->activate ();
}

void
bmp_window_playlist_configure (BmpWindowPlaylist *wa2_playlist)
{
    GtkWidget *window = wa2_playlist->window;
    TBmpPlaylistComponent *component = NULL;

    if (wa2_playlist->priv->ui_components != NULL)
        g_hash_table_destroy (wa2_playlist->priv->ui_components);

        wa2_playlist->priv->ui_components =
        g_hash_table_new_full (g_str_hash,
                               g_str_equal,
                               NULL,
                               destroy_value);

        for (unsigned int i = 0; i < G_N_ELEMENTS (playlist_component_data); i++)
        {
            component = g_new0 (TBmpPlaylistComponent, 1);

            component->name = 0;
            component->pixbuf = pledit_new_pb_simple (playlist_component_data[i].width,
                            playlist_component_data[i].height);

            pledit_copy_area (component->pixbuf,
                      playlist_component_data[i].x_origin,
                      playlist_component_data[i].y_origin,
                      playlist_component_data[i].width,
                      playlist_component_data[i].height);


            component->x_origin = playlist_component_data[i].x_origin;
            component->y_origin = playlist_component_data[i].y_origin;
            component->width = playlist_component_data[i].width;
            component->height = playlist_component_data[i].height;

            g_hash_table_insert (wa2_playlist->priv->ui_components,
                                 playlist_component_data[i].name,
                                 component);

    }

    GdkColor *bg_normal  = bmp_colors_get_color (COLOR_PL_BG_NORMAL);
    GdkColor *fg_normal  = bmp_colors_get_color (COLOR_PL_FG_NORMAL);
    GdkColor *fg_current = bmp_colors_get_color (COLOR_PL_FG_CURRENT);

#if 0
    GtkWidget *widget = GTK_WIDGET (g_object_get_data (G_OBJECT (wa2_playlist->window), "entry"));
    PangoFontDescription *pfd = pango_font_description_from_string (mcs->key_get<std::string>("bmp", "font") . c_str ());
    gtk_widget_modify_font (widget, pfd);
    pango_font_description_free (pfd);

        gtk_widget_modify_base (GTK_WIDGET (widget), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (widget), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (widget), GTK_STATE_NORMAL, fg_normal); 

        gtk_widget_modify_base (GTK_WIDGET (widget), GTK_STATE_ACTIVE, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (widget), GTK_STATE_ACTIVE, bg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (widget), GTK_STATE_ACTIVE, fg_normal); 

        gtk_widget_modify_base (GTK_WIDGET (widget), GTK_STATE_SELECTED, fg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (widget), GTK_STATE_SELECTED, fg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (widget), GTK_STATE_SELECTED, bg_normal); 

        gtk_widget_modify_text (GTK_WIDGET (widget), GTK_STATE_ACTIVE, fg_normal); 
        gtk_widget_modify_text (GTK_WIDGET (widget), GTK_STATE_SELECTED, bg_normal); 
#endif

        gtk_widget_modify_base (GTK_WIDGET (window), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (window), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (window), GTK_STATE_NORMAL, fg_normal); 

        gtk_widget_modify_base (GTK_WIDGET (wTracklist), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (wTracklist), GTK_STATE_NORMAL, bg_normal); 
        gtk_widget_modify_bg (GTK_WIDGET (wTracklistScrollbar), GTK_STATE_NORMAL, bg_normal); 

        gtk_widget_modify_fg (GTK_WIDGET (wTracklist), GTK_STATE_NORMAL, fg_normal); 
        gtk_widget_modify_fg (GTK_WIDGET (wTracklistScrollbar), GTK_STATE_NORMAL, fg_normal); 

    gtk_widget_modify_fg (GTK_WIDGET (wTracklist), GTK_STATE_PRELIGHT, fg_current); 
    gtk_widget_modify_fg (GTK_WIDGET (wTracklistScrollbar), GTK_STATE_PRELIGHT, fg_current); 

    set_cursors (wa2_playlist);
    gtk_widget_queue_draw (GTK_WIDGET (window));
}

BmpWindowPlaylist*
bmp_window_playlist_new (void)
{
  return BMP_WINDOW_PLAYLIST (g_object_new (BMP_TYPE_WINDOW_PLAYLIST, NULL)); 
}

void
bmp_window_playlist_dnd_txn_play (BmpWindowPlaylist  *wa2_playlist)
{
  if (wa2_playlist->priv->dnd_strv == NULL) return;

  int out;
  bmp_system_control_add_uri_list (bmp_system_control,
          (const char**) wa2_playlist->priv->dnd_strv,
          wa2_playlist->priv->dnd_index,
          FALSE,
          TRUE,
          -1,
          &out,
          NULL);
  g_strfreev (wa2_playlist->priv->dnd_strv);
  wa2_playlist->priv->dnd_strv  = NULL;
  wa2_playlist->priv->dnd_index = 0;
}

void
bmp_window_playlist_dnd_txn_append (BmpWindowPlaylist  *wa2_playlist)
{
  if (wa2_playlist->priv->dnd_strv == NULL) return;

  int out;
  bmp_system_control_add_uri_list (bmp_system_control,
          (const char**) wa2_playlist->priv->dnd_strv,
          wa2_playlist->priv->dnd_index,
          FALSE,
          FALSE,
          0,
          &out,
          NULL);
  g_strfreev (wa2_playlist->priv->dnd_strv);
  wa2_playlist->priv->dnd_strv = NULL;
  wa2_playlist->priv->dnd_index = 0;
}

void
bmp_window_playlist_dnd_txn_cancel (BmpWindowPlaylist  *wa2_playlist)
{
  if (wa2_playlist->priv->dnd_strv == NULL) return;
  g_strfreev (wa2_playlist->priv->dnd_strv);
  wa2_playlist->priv->dnd_strv = NULL;
  wa2_playlist->priv->dnd_index = 0;
}
