//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <gtk/gtk.h>
#include <glib/gi18n.h>

#include <cstring>
#include <iostream>
#include <sstream>

#include "widgets/bmp_button.h"
#include "widgets/bmp_button_toggle.h"
#include "widgets/bmp_slider.h"
#include "widgets/bmp_window.h"
#include "widgets/bmp_status_icon.h"
#include <chroma/chroma-list.h>
#include <goa/libgoa.h>


#include "winamp2_playlist.hpp"
#include "winamp2_main.hpp"
#include "winamp2_skin.hpp"

//Utility/auxilliary stuff
#include "ui_util.hpp"
#include "util.hpp"
#include "main.hpp"
#include "paths.hpp"
#include "debug.hpp"
#include "file_utils.hpp"

#include "x_library.hpp"
#include "x_play.hpp"
#include "x_system_control.hpp"

using namespace Bmp;

#include "gmacros.hpp"
BMP_DEFINE_TYPE (BmpWinamp2Skin, bmp_winamp2_skin, G_TYPE_OBJECT)

namespace
{
    bool dispose_has_run = false;
}

namespace
{
  GdkColor  *colors[N_COLORS];
  GdkPixbuf *cursors[N_CURSORS];
  std::vector<std::string> cursors_names;
  std::vector<std::string> color_names;
  GList *osc = 0;

  const char* skin_files[] =
  {
        "main.bmp",
        "shufrep.bmp",
        "cbuttons.bmp",
        "posbar.bmp",
        "volume.bmp",
        "titlebar.bmp",
        "balance.bmp",
        "pledit.bmp",
        "monoster.bmp",
         NULL,
        "playpaus.bmp"
  };
}

GdkPixbuf *wa2_skin_bitmap[N_BITMAPS];
GdkBitmap *mask_window_main = 0;
int        _skin_n_numbers = 0;

namespace
{
  const char* PLEDIT_TEXT_SECTION_NAME = "Text";

  std::string
  get_color (std::string const &path,
             Wa2SkinColor      color)
  {
      std::string filename, value;
      if (Util::find_file (path, BMP_FILE_PLEDIT, filename, 1))
        {
          value = Bmp::Util::read_ini_string (filename, PLEDIT_TEXT_SECTION_NAME, color_names[color]);
        }
      return value;
  }

  void
  colors_init (std::string const& archive_dirname)
  {
    gdk_color_parse (get_color (archive_dirname, COLOR_PL_BG_NORMAL).c_str(), colors[COLOR_PL_BG_NORMAL]);
    gdk_color_parse (get_color (archive_dirname, COLOR_PL_FG_CURRENT).c_str(), colors[COLOR_PL_FG_CURRENT]);
    gdk_color_parse (get_color (archive_dirname, COLOR_PL_FG_NORMAL).c_str(), colors[COLOR_PL_FG_NORMAL]);
    gdk_color_parse (get_color (archive_dirname, COLOR_PL_BG_SELECTED).c_str(), colors[COLOR_PL_BG_SELECTED]);
  }

  GdkBitmap*
  create_transparent_mask (std::string const& path,
                           std::string const& file,
                           std::string const& section,
                           int width,
                           int height)

  {
    GdkBitmap   *mask = NULL;
    GdkGC       *gc = NULL;
    GdkPoint    *points;
    GdkColor     pattern;
    GArray      *num, *point;
    bool         created_mask = false;

    std::string filename;

    if (!path.empty ())
      {
        Util::find_file (path, file, filename, 1);
      }

    if (filename.empty ())
      {
        return NULL;
      }

    if ((num   = (GArray*)Bmp::Util::read_ini_array (filename.c_str(), section.c_str(), "NumPoints")) == NULL)
      {
        return NULL;
      }

    if ((point = (GArray*)Bmp::Util::read_ini_array (filename.c_str(), section.c_str(), "PointList")) == NULL)
      {
        g_array_free (num, TRUE);
        return NULL;
      }

    mask = gdk_pixmap_new (NULL, width, height, 1);
    gc = gdk_gc_new (mask);

    pattern.pixel = 0;
    gdk_gc_set_foreground (gc, &pattern);
    gdk_draw_rectangle (mask, gc, TRUE, 0, 0, width, height);
    pattern.pixel = 1;
    gdk_gc_set_foreground (gc, &pattern);

    int k;
    int j = 0;

    for (unsigned int i = 0; i < num->len; i++)
    {
        if ((point->len - j) >= (g_array_index(num, gint, i) * 2))
        {
            points = g_new (GdkPoint, g_array_index(num, gint, i));
            for (k = 0; k < g_array_index(num, gint, i); k++)
            {
                points[k].x = g_array_index(point, gint, j + k * 2);
                points[k].y = g_array_index(point, gint, j + k * 2 + 1);
            }
            j += k * 2;
            gdk_draw_polygon (mask, gc, TRUE, points, g_array_index(num, gint, i));
            g_free (points);
        }
        created_mask = true;
    }

    g_array_free (num, TRUE);
    g_array_free (point, TRUE);

    if (!created_mask)
      {
        gdk_draw_rectangle (mask, gc, TRUE, 0, 0, width, height);
      }

    gdk_gc_destroy (gc);

    return mask;
  }

  GdkPixbuf*
  get_cursor_pixbuf (std::string const& archive_dirname,
                     std::string const& cursor_filename)
  {
      std::string filename;
      if (!Util::find_file (archive_dirname, cursor_filename, filename, 1))
            return NULL;

      GdkPixbufAnimation *cursor_animated =
            gdk_pixbuf_animation_new_from_file (filename.c_str(), NULL);

      GdkPixbuf *pb_aux =
            gdk_pixbuf_animation_get_static_image (cursor_animated);

      GdkPixbuf *pb_return =
            gdk_pixbuf_copy (pb_aux);

      g_object_unref (cursor_animated);
      return pb_return;
  }

  void
  cursor_add (std::string const& archive_dirname,
                  int                        _cursor)
  {
    cursors[_cursor] = get_cursor_pixbuf (archive_dirname.c_str(), cursors_names[_cursor].c_str());
  }

  void
  cursors_set (std::string const& archive_dirname)
  {
    GList *old_cursors = NULL;
    for (int n = 0; n < N_CURSORS; n++)
      {
         if (cursors[n] != 0)
           {
             old_cursors = g_list_append (old_cursors, gpointer(cursors[n]));
           }
         cursors[n] = 0;
      }

    cursor_add (archive_dirname, CURSOR_NORMAL);
    cursor_add (archive_dirname, CURSOR_POSBAR);
    cursor_add (archive_dirname, CURSOR_VOLBAR);
    cursor_add (archive_dirname, CURSOR_PLAYLIST);
    cursor_add (archive_dirname, CURSOR_PLAYLIST_SIZE);
    cursor_add (archive_dirname, CURSOR_MAINMENU);

    if (old_cursors) g_list_foreach (old_cursors, GFunc(g_object_unref), NULL);
  }

  void
  skin_components_free_collect (BmpWinamp2Skin *ui)
  {
      for (unsigned int n = 0; n < N_BITMAPS; n++)
        {
         if (wa2_skin_bitmap[n])
          {
            osc = g_list_append (osc, wa2_skin_bitmap[n]);
          }
        }
  }

  void
  skin_components_free_commit (BmpWinamp2Skin *ui)
  {
      for ( ; osc ; osc = osc->next )
        {
          GdkPixbuf *sc = reinterpret_cast<GdkPixbuf*>(osc->data);
          g_object_unref (sc);
        }
      osc = 0;
  }

  void
  load_skin_components (BmpWinamp2Skin        *ui,
                        std::string const           &archive_filename,
                        int                                    type)
  {
      std::string  filename;
      std::string  archive_dirname;

      debug ("winamp2-skin", "Initializing Path.");
      archive_dirname = (type == BMP_SKIN_ARCHIVE) ? Util::archive_decompress (archive_filename) : archive_filename;

      debug ("winamp2-skin", "Initializing colors.");
      colors_init (archive_dirname);
      debug ("winamp2-skin", "Loading Cursors from archive.");
      cursors_set (archive_dirname);

      debug ("winamp2-skin", "Collecting Components.");
      skin_components_free_collect (ui);

      for (unsigned int n = 0; n < G_N_ELEMENTS(skin_files); n++)
        {
          if (!skin_files[n]) continue;

          debug ("winamp2-skin", "Searching for: %s", skin_files[n]);
          if (!Util::find_file (archive_dirname, skin_files[n], filename, 1))
            {
              debug ("winamp2-skin", "%s not found", skin_files[n]);
              wa2_skin_bitmap[n] = NULL;
            }
          else
            {
              wa2_skin_bitmap[n] = gdk_pixbuf_new_from_file (filename.c_str(), NULL);
            }
        }

      debug ("winamp2-skin", "Loading numbers,");
      Util::find_file (archive_dirname, "numbers.bmp", filename, 1);
      if (filename.empty ())
        {
            Util::find_file (archive_dirname, "nums_ex.bmp", filename, 1);
        }

      if (filename.empty ())
        {
            g_warning ("%s: Numbers not found", G_STRFUNC);
        }
      else
        {
            wa2_skin_bitmap[BITMAP_NUMBERS] = gdk_pixbuf_new_from_file (filename.c_str(), NULL);
            if (gdk_pixbuf_get_width(wa2_skin_bitmap[BITMAP_NUMBERS]) >= 108)
              {
                _skin_n_numbers = 12;
              }
            else
              {
                _skin_n_numbers = 10;
              }
        }

      debug ("winamp2-skin", "Initializing main window xshape mask");
      mask_window_main = create_transparent_mask (archive_dirname,
                                                  BMP_FILE_REGION,
                                                  BMP_MASK_REGION_SECTION_MAIN_WINDOW,
                                                  275,
                                                  116);

      if (type == BMP_SKIN_ARCHIVE)
        {
          debug ("winamp2-skin", "Deleting directory.");
          Util::del_directory (archive_dirname);
        }

      debug ("winamp2-skin", "Freeing Components.");
      skin_components_free_commit (ui);
  }

  void
  bmp_winamp2_skin_dispose (GObject *obj)
  {
    BmpWinamp2Skin *ui = BMP_WINAMP2_SKIN (obj);
    skin_components_free_collect (ui);
    skin_components_free_commit (ui);
    g_message("stopped: %s", G_OBJECT_TYPE_NAME(ui));
  }

  void
  bmp_winamp2_skin_class_init (BmpWinamp2SkinClass * klass)
  {
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose  = bmp_winamp2_skin_dispose;
  }

  void
  bmp_winamp2_skin_init (BmpWinamp2Skin *ui)
  {
    for (unsigned int n = 0; n < N_BITMAPS; wa2_skin_bitmap[n++] = 0);
    for (unsigned int n = 0; n < N_COLORS; colors[n++]=0);

    mask_window_main = 0;

    const char *_cursor_names[] =
    {
     "normal.cur",
     "posbar.cur",
     "volbar.cur",
     "pnormal.cur",
     "psize.cur",
     "mainmenu.cur"
    };
    for (unsigned int i = 0; i < G_N_ELEMENTS(_cursor_names); i++)
      {
        cursors_names.push_back (_cursor_names[i]);
      }

    const char *_color_names[] =
    {
     "NormalBG",
     "Normal",
     "SelectedBG",
     "Current"
    };
    for (unsigned int i = 0; i < G_N_ELEMENTS(_color_names); i++)
      {
        color_names.push_back (_color_names[i]);
      }

    for (unsigned int i = 0; i < N_CURSORS; i++)
      {
        cursors[i] = 0;
      }
    for (unsigned int i = 0; i < N_COLORS; i++)
      {
        colors[i] = g_new0 (GdkColor, 1);
      }
  }
}

BmpWinamp2Skin*
bmp_winamp2_skin_new (void)
{
    BmpWinamp2Skin *ui;
    GtkSettings *settings;
    const char *name;
    ui = BMP_WINAMP2_SKIN (g_object_new (BMP_TYPE_WINAMP2_SKIN, NULL));

    // Hack-around for no xsettings manager available that would set the cursor theme
    settings = gtk_settings_get_for_screen (gdk_screen_get_default());
    name = g_object_get_string (settings, "gtk-cursor-theme-name");
    if (!name)
    {
      name = g_getenv("XCURSOR_THEME");
      g_object_set (settings, "gtk-cursor-theme-name", name, NULL);
    }
    return ui;
}

bool
bmp_winamp2_skin_load_skin (BmpWinamp2Skin     *ui,
                            std::string const&  filename)
{
  if (!Glib::file_test(filename, Glib::FILE_TEST_EXISTS))
    {
      return false;
    }

  if (Glib::file_test(filename, Glib::FILE_TEST_IS_DIR))
    {
        load_skin_components (ui, filename, BMP_SKIN_DIR);
    }
  else if (Glib::file_test(filename, Glib::FILE_TEST_IS_REGULAR))
    {
        load_skin_components (ui, filename, BMP_SKIN_ARCHIVE);
    }
  else
    {
        return false;
    }

  return true;
}

GdkCursor*
bmp_cursors_get_cursor (Wa2CursorId _cursor)
{
  GdkCursor * cursor;
  GdkPixbuf * pixbuf;
  const gchar *hot_x = NULL, *hot_y = NULL;
  gint hot_x_int = 0, hot_y_int = 0;

  pixbuf = cursors[_cursor];

  if (!mcs->key_get<bool>("bmp","use-custom-cursors")) return NULL;

  if (!pixbuf)
    {
      pixbuf = cursors[CURSOR_NORMAL];
      if (!pixbuf) return NULL;
    }

  hot_x = gdk_pixbuf_get_option(pixbuf,"x_hot");
  hot_y = gdk_pixbuf_get_option(pixbuf,"y_hot");

  if (hot_x) hot_x_int = std::atoi (hot_x);
  if (hot_y) hot_y_int = std::atoi (hot_x);

  cursor = gdk_cursor_new_from_pixbuf  (gdk_display_get_default(),
                                              pixbuf,
                                              hot_x_int,
                                              hot_y_int);
  return cursor;
}

GdkColor*
bmp_colors_get_color (Wa2SkinColor color)
{
    return colors[color];
}
