/*  BMPx - The Dumb Music Player
 *  Copyright (C) 2005 BMPx development team.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  $Id$
 *
 * The BMPx project hereby grant permission for non-gpl compatible GStreamer
 * plugins to be used and distributed together with GStreamer and BMPx. This
 * permission are above and beyond the permissions granted by the GPL license
 * BMPx is covered by.
 */

#include <math.h>
#include <gtk/gtk.h>
#include <widgets/bmp_slider.h>

G_DEFINE_TYPE (BmpSlider, bmp_slider, GTK_TYPE_DRAWING_AREA)

static void bmp_slider_destroy                  (GtkObject        *object);
static void bmp_slider_realize                  (GtkWidget        *widget);
static gint bmp_slider_expose                   (GtkWidget        *widget,
						GdkEventExpose   *event);
static gint bmp_slider_slider_press             (GtkWidget        *widget,
						GdkEventButton   *event);
static gint bmp_slider_slider_release           (GtkWidget        *widget,
						GdkEventButton   *event);
static gint bmp_slider_enter_notify		(GtkWidget	 *widget,
						GdkEventCrossing *event);
static gint bmp_slider_leave_notify		(GtkWidget	 *widget,
						GdkEventCrossing *event);
static gint bmp_slider_motion_notify            (GtkWidget *widget,
						GdkEventMotion *event);
static gint bmp_slider_scroll			(GtkWidget *widget,
						GdkEventScroll *event);

enum
{
  BMP_SLIDER_MOVED,
  BMP_SLIDER_MOTION_HINT,
  SLIDER_N_SIGNALS
};

static guint signals[SLIDER_N_SIGNALS] = { 0 };

static void
bmp_slider_class_init (BmpSliderClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;
  object_class->destroy = bmp_slider_destroy;
  widget_class->realize = bmp_slider_realize;
  widget_class->expose_event = bmp_slider_expose;
  widget_class->enter_notify_event = bmp_slider_enter_notify;
  widget_class->leave_notify_event = bmp_slider_leave_notify;
  widget_class->motion_notify_event = bmp_slider_motion_notify;
  widget_class->button_press_event = bmp_slider_slider_press;
  widget_class->button_release_event = bmp_slider_slider_release;
  widget_class->scroll_event = bmp_slider_scroll;

  signals[BMP_SLIDER_MOVED] =
  g_signal_new ("moved",
                  G_OBJECT_CLASS_TYPE (object_class),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (BmpSliderClass, moved),
                  NULL, NULL,
                  g_cclosure_marshal_VOID__VOID,
                  G_TYPE_NONE, 0);

  signals[BMP_SLIDER_MOTION_HINT] =
  g_signal_new ("motion_hint",
                  G_OBJECT_CLASS_TYPE (object_class),
                  G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (BmpSliderClass, motion_hint),
                  NULL, NULL,
                  g_cclosure_marshal_VOID__VOID,
                  G_TYPE_NONE, 0);
}

static void
bmp_slider_init (BmpSlider *slider)
{
	slider->width = 0;
	slider->height = 0;
	slider->pb_normal = NULL;
	slider->pb_pressed = NULL;
	slider->constructed = 0;
	slider->in_slider = 0;
	slider->state = BMP_SLIDER_STATE_IDLE;
	slider->position = 0;
	slider->button_yoffset = 0;
	slider->on_motion = FALSE;
	slider->scroll_step = 1; 
	slider->max_adjust = 0;
}


GtkWidget*
bmp_slider_new (void)
{
  BmpSlider *slider;

  slider = gtk_type_new (bmp_slider_get_type ());
  return GTK_WIDGET (slider);
}

void
bmp_slider_set_from_table (BmpSlider	   *slider,
			  const GdkPixbuf  *src,
			  guint		    width,
			  guint		    height,
			  guint		    width_button,
			  guint		    height_button,
			  guint		    x_sliderbg,
			  guint		    y_sliderbg,
			  guint		    x_normal,
			  guint		    y_normal,
			  guint		    x_pressed,
			  guint		    y_pressed,
			  gboolean	    has_button)
{
  g_return_if_fail (GDK_IS_PIXBUF (src));

  slider->pb_sliderbg = gdk_pixbuf_new (gdk_pixbuf_get_colorspace(src),
                                            gdk_pixbuf_get_has_alpha(src),
                                            gdk_pixbuf_get_bits_per_sample(src),
                                            width, height);

  gdk_pixbuf_copy_area (src, x_sliderbg, y_sliderbg, width, height,slider->pb_sliderbg, 0, 0);
                        
  if (height_button > 15)
    {
      has_button = FALSE;
    }

  if (has_button)
    {

	slider->pb_normal = gdk_pixbuf_new (gdk_pixbuf_get_colorspace(src),
						gdk_pixbuf_get_has_alpha(src),
						gdk_pixbuf_get_bits_per_sample(src),
						width_button, height_button);
	slider->pb_pressed = gdk_pixbuf_new (gdk_pixbuf_get_colorspace(src),
						gdk_pixbuf_get_has_alpha(src),
						gdk_pixbuf_get_bits_per_sample(src),
						width_button, height_button);

	if (x_normal < gdk_pixbuf_get_width (src))
	  {
	    gdk_pixbuf_copy_area (src, x_normal, y_normal, width_button, height_button,
	    			  slider->pb_normal, 0, 0);

  	    gdk_pixbuf_copy_area (src, x_pressed, y_pressed, width_button, height_button,
				  slider->pb_pressed, 0, 0);

	    slider->has_button = TRUE;
	  }
        else
	  {
	    slider->has_button = FALSE;
	  }

    }
  else
    {
	slider->has_button = FALSE;
    }

  slider->width = width;
  slider->height = height;

  slider->width_button = width_button;
  slider->height_button = height_button;

  if (slider->position < width_button/2)
      slider->position = width_button/2;

  slider->slider_max = ceilf (slider->width - ((slider->width_button)/2) - slider->max_adjust + 0.5);
  slider->slider_min = slider->width_button/2;
}

static void
bmp_slider_destroy (GtkObject *object)
{
  g_return_if_fail (BMP_IS_SLIDER (object));
  if (GTK_OBJECT_CLASS (bmp_slider_parent_class)->destroy)
    (*GTK_OBJECT_CLASS (bmp_slider_parent_class)->destroy) (object);
}

static void
bmp_slider_realize (GtkWidget *widget)
{
  GdkWindowAttr attributes;
  int attributes_mask;
  g_return_if_fail (widget != NULL);
  g_return_if_fail (BMP_IS_SLIDER (widget));
  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = gtk_widget_get_events (widget) |
    GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK |
    GDK_BUTTON_RELEASE_MASK | GDK_LEAVE_NOTIFY_MASK | GDK_ENTER_NOTIFY_MASK |
    GDK_POINTER_MOTION_MASK | GDK_BUTTON_MOTION_MASK | GDK_SCROLL_MASK;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (widget->parent->window, &attributes, attributes_mask);
  widget->style = gtk_style_attach (widget->style, widget->window);
  gdk_window_set_user_data (widget->window, widget);
  gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);
}

static gboolean
bmp_slider_expose (GtkWidget      *widget,
		   GdkEventExpose *event)
{
  g_return_val_if_fail (BMP_IS_SLIDER (widget), FALSE);
  if (event->count > 0) return FALSE;
  BmpSlider *slider = BMP_SLIDER (widget);
  int slice_offset, slice_height, pos, step;

  if (slider->sliced)
    {
          pos  =  (slider->position) - (slider->width_button/2);
          step =  slider->width / slider->num_slices;
	  slice_offset = (pos/step) * slider->slice_size;
          slice_height = BMP_SLIDER(slider)->slice_size;

	  if ( (slice_height + slice_offset) > 419) slice_height--;

	  if (slice_offset <= gdk_pixbuf_get_height(BMP_SLIDER(slider)->pb_sliderbg))
	  gdk_draw_pixbuf(GTK_WIDGET(slider)->window,
		 	NULL,	
			BMP_SLIDER(slider)->pb_sliderbg,
			0,
		        slice_offset,
			0,
			0,
			BMP_SLIDER(slider)->width,
		 	slice_height,	
			GDK_RGB_DITHER_NONE,
			0,
			0);
    }
  else
  if (GDK_IS_PIXBUF (BMP_SLIDER(slider)->pb_sliderbg))
    {
	  gdk_draw_pixbuf(GTK_WIDGET(slider)->window,
		 	NULL,	
			BMP_SLIDER(slider)->pb_sliderbg,
			0,
		        0,
			0,
			0,
			BMP_SLIDER(slider)->width,
			BMP_SLIDER(slider)->height,
			GDK_RGB_DITHER_NONE,
			0,
			0);
    }

  if (slider->has_button)
    {
	GdkPixbuf *pb = slider->pb_normal;
	if (slider->state == BMP_SLIDER_STATE_ACTIVE && slider->in_slider) pb = slider->pb_pressed;
	gdk_draw_pixbuf(GTK_WIDGET(slider)->window,
				NULL,	
				pb,
				0,
				0,
				(slider->position)-(slider->width_button/2),
				slider->button_yoffset,
				-1,
				-1,
				GDK_RGB_DITHER_NONE,
				0,
				0);
    }
  
  return FALSE;
}

static gboolean
bmp_slider_slider_press (GtkWidget      *widget,
		         GdkEventButton *event)
{
  g_return_val_if_fail (BMP_IS_SLIDER (widget), FALSE);
  if (event->type == GDK_BUTTON_PRESS && event->button == 1)
    {
	BmpSlider *slider = BMP_SLIDER (widget);
	slider->state = BMP_SLIDER_STATE_ACTIVE;

	slider->position = event->x;

	if (slider->position < slider->slider_min) slider->position = slider->slider_min;
	if (slider->position > slider->slider_max) slider->position = slider->slider_max;

	gtk_widget_queue_draw(widget);
    }

  return TRUE;
}

static gboolean
bmp_slider_scroll (GtkWidget      *widget,
	           GdkEventScroll *event)
{
  BmpSlider *slider = BMP_SLIDER(widget);
  if (event->direction == GDK_SCROLL_UP)
    {
	slider->position += slider->scroll_step;
    } 
  else if (event->direction == GDK_SCROLL_DOWN)
    {
	slider->position -= slider->scroll_step;
    }

  if (slider->position < slider->slider_min) slider->position = slider->slider_min;
  if (slider->position > slider->slider_max) slider->position = slider->slider_max;

  g_signal_emit (slider, signals[BMP_SLIDER_MOVED], 0);
  gtk_widget_queue_draw (widget);
  return TRUE;
}

static gboolean
bmp_slider_slider_release (GtkWidget      *widget,
			  GdkEventButton *event)
{
  g_return_val_if_fail (BMP_IS_SLIDER (widget), FALSE);
  BmpSlider *slider = BMP_SLIDER (widget);
  slider->state = BMP_SLIDER_STATE_IDLE;
  g_signal_emit(slider, signals[BMP_SLIDER_MOVED], 0);
  gtk_widget_queue_draw(widget);
  return TRUE;
}

static gboolean
bmp_slider_enter_notify (GtkWidget *widget, GdkEventCrossing *event)
{
  BmpSlider *slider = BMP_SLIDER(widget);
  slider->in_slider = TRUE;
  gtk_widget_queue_draw(widget);
  return TRUE;
}

static gboolean
bmp_slider_leave_notify (GtkWidget *widget, GdkEventCrossing *event)
{
  BmpSlider *slider = BMP_SLIDER(widget);
  slider->in_slider = FALSE;
  gtk_widget_queue_draw(widget);
  return TRUE;
}

static gboolean
bmp_slider_motion_notify (GtkWidget *widget, GdkEventMotion *event)
{

  BmpSlider *slider = BMP_SLIDER(widget);

  if (slider->state == BMP_SLIDER_STATE_ACTIVE)
    {
	slider->position = event->x; 

	if (slider->position < slider->slider_min) slider->position = slider->slider_min;
	if (slider->position > slider->slider_max) slider->position = slider->slider_max;

	if (slider->on_motion)
	  {
	    g_signal_emit (slider, signals[BMP_SLIDER_MOVED], 0);
	  }

	g_signal_emit (slider, signals[BMP_SLIDER_MOTION_HINT], 0);
	gtk_widget_queue_draw(widget);
    }
  return TRUE;
}

gdouble
bmp_slider_get_position (BmpSlider  *slider)
{
  g_return_val_if_fail (BMP_IS_SLIDER (slider), 0);
  return ((slider->position*1.) - (slider->width_button/2.)) / ((slider->width - slider->width_button)/100.); 
}

void
bmp_slider_set_position (BmpSlider    *slider,
			 gdouble       position)
{
  g_return_if_fail (BMP_IS_SLIDER (slider));
  double range = slider->width - slider->width_button - 0.5; 
  double percent = range / 100.;
  slider->position = (position * percent) + (slider->width_button/2.); 
  gtk_widget_queue_draw (GTK_WIDGET(slider));
}

void
bmp_slider_set_slice_data (BmpSlider	*slider,
			   gboolean	 sliced,
			   guint	 slice_size,  
			   guint	 num_slices,
			   gboolean	 has_button)
{
  g_return_if_fail (slider != NULL);
  g_return_if_fail (BMP_IS_SLIDER (slider));

  slider->sliced = sliced;
  slider->slice_size = slice_size;
  slider->num_slices = num_slices;
  slider->has_button = has_button;
}

void
bmp_slider_set_report_on_motion (BmpSlider    *slider,
			         gboolean      flag)
{
  g_return_if_fail (BMP_IS_SLIDER (slider));
  slider->on_motion = flag;
}

void
bmp_slider_set_scroll_step  (BmpSlider	  *slider,
			     guint	   step)
{
  g_return_if_fail (BMP_IS_SLIDER (slider));
  slider->scroll_step = step;
}






